/* 
DPMIINFO.C -- display information about DPMI server
Andrew Schulman, April 1994
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include "dpmish.h"

void fail(const char *s, ...) { puts(s); exit(1); }

unsigned _dpmi_flags(void)
{
	_asm mov ax, 0400h
	_asm int 31h
	_asm mov ax, bx
	// retval in AX
}

unsigned _dpmi_version(void)
{
	_asm mov ax, 0400h
	_asm int 31h
	// retval in AX
}

unsigned long _dpmi_mem(void)
{
	unsigned long buf[12], far *fp = buf;
	_asm push di
	_asm mov ax, 0500h
	_asm les di, fp
	_asm int 31h
	_asm pop di
	return buf[0];
}

static unsigned dpmi_32_flag = -1;

int real_main(int argc, char *argv[])
{
	// Should only get this flag from real/V86 mode, so call in
	// real_main, but let DPMISH tell us if DPMI is really present
	// before actually using the flag.
    _asm push si
    _asm push di
    _asm mov ax, 1687h
    _asm int 2Fh
    _asm pop di
    _asm pop si
    _asm mov dpmi_32_flag, bx
    return 0;
}

int pmode_main(int argc, char *argv[])
{
	unsigned flags = _dpmi_flags();
	unsigned vers = _dpmi_version();

	unsigned char maj = vers >> 8;
	unsigned char min = vers & 0xFF;
	if ((maj == 0) && (min == 0x90)) min = 90;	// silly Std mode bug
	printf("DPMI version %d.%02d\n", maj, min);

	printf("%s programs supported\n",
		(flags & 1) ? "32-bit" : "Only 16-bit");
	printf("%d-based DPMI host\n", 
		(flags & 1) ? 386 : 286);
	printf("Interrupts reflected to %s mode\n",
		(flags & 2) ? "real" : "V86");
	printf("%s virtual memory\n",
		(flags & 4) ? "Supports" : "No");
	printf("%lu bytes available in largest block\n",
		_dpmi_mem());
}

