/* dumpfile.c */

#if 0
/*
	12/25/92:
		Added -UNICODE switch for 16-bit chars (not finished)
		Changed -STRINGS to STRINGS n (n sets min_str; default 5)

	TO DO:
		dump as WORDS, LONGS, in addition to BYTES (default)
			
*/
#endif

#include <stdlib.h>
#ifdef WINIO
#include "windows.h"
#include "winio.h"
#else
#include <stdio.h>
#endif
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <ctype.h>

#ifndef min
#define min(a, b)	((a) < (b) ? (a) : (b))
#endif

#define BUFSIZE     4096
#define MIN_STR     (min_str)
#define MAX_STR     70
#define WIDTH       16

void dump(unsigned char far *fp, unsigned bytes, 
    char *mask, unsigned long addr, int width);		// DUMP.C
void strings(unsigned char far *fp, unsigned bytes, 
    char *mask, unsigned long addr);
unsigned long axton(char *s);
void digits(void);
void rule(void);

void fail_perror(void)  { perror("dumpfile fail"); exit(1); }
void fail(char *s)      { puts(s); exit(1); }

static int unicode = 0;
static int min_str = 5;	// default

main(int argc, char *argv[])
{
    char *filename, *buf, *mask;
    long offset = 0;
    long bytes = -1;
    int count, f, i, do_strings = 0, do_relative = 0, do_header = 0;
	int width = 16;
    
    if (argc < 2)
        fail("usage: dumpfile <filename> "
             "[-offset [0x]n] [-rel] [-bytes [0x]n] [-strings [n]] [-uni]");
         
    for (i=1; i<argc; i++)
        if (argv[i][0] == '-')
        {
            char *s = strupr(&argv[i][1]);
            if (strncmp(s, "OFFSET", 3) == 0)
                offset = axton(argv[++i]);
            else if (strncmp(s, "BYTES", 3) == 0)
                bytes = axton(argv[++i]);
			else if (strncmp(s, "UNICODESTRING", 3) == 0)
			{
				do_strings++;
				unicode++;
			}
            else if (strncmp(s, "STRINGS", 3) == 0)
			{
				int x;
                do_strings++;
				if ((x = axton(argv[i+1])) != 0)
				{
					min_str = x;
					i++;
				}
			}
            else if (strncmp(s, "RELATIVE", 3) == 0)
                do_relative++;
            else if (strncmp(s, "HEADER", 4) == 0)
                do_header++;
			else if (strncmp(s, "WIDTH", 3) == 0)
				width = axton(argv[++i]);
        }
        else
            filename = argv[i];
		
    if ((f = sopen(filename, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
        fail_perror();
    if ((buf = (char *) malloc(BUFSIZE)) == 0)
        fail("insufficient memory");
	
	if (offset < 0)
		offset += filelength(f);	// offset from end of file

    if (offset)         
        lseek(f, offset, SEEK_SET);
    if (bytes == -1)    
        bytes = filelength(f);
    
    if (do_relative)
        offset = 0;     // relative addressing; do AFTER lseek

    if (do_header)
    {
        digits();
        rule();
    }
    
    mask = (filelength(f) < 0x10000L) ? "%04lx" : "%08lx" ;
             
    for (;;)
    {
        if ((count = read(f, buf, BUFSIZE)) <= 0)
            break;
        if (do_strings)
            strings(buf, min(bytes,count), mask, offset);
        else
            dump(buf, min(bytes,count), mask, offset, width);
        offset += count;
        if ((bytes -= count) <= 0)
            break;
    }
    
    if (do_header)
    {
        rule();
        digits();
    }
             
    free(buf);
    close(f);
    return 0;
}

void rule(void)
{
    printf("\t  ------------------------------------------------- \n");
}

void digits(void)
{
    printf("\t   00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F\n");
}

void strings(unsigned char far *fp, unsigned bytes, 
    char *mask, unsigned long addr)
{
    char buf[128];
    char *str;
    unsigned i, c, loc;
    
	if (unicode)
	{
		unsigned int far *p;
		bytes >>= 1;
		for (i=0, p=(unsigned int far *) fp; i<bytes; i++, p++)
		{
			if (isalnum((unsigned char) *p))
			{
				c=0;
				str=buf;
				loc=i;
				while ((isalnum((unsigned char) *p) || 
						ispunct((unsigned char) *p) || 
						(*p == ' ')) && 
						(c < MAX_STR))
				{
					*str++ = *p++; c++; i++;
				}
				if (c >= MIN_STR)
				{
					buf[c] = '\0';
					printf(mask, addr+loc);
					printf(" %s\n", buf);
				}
			}
		}
	}
	else
	{
		unsigned char far *p;
		for (i=0, p=fp; i<bytes; i++, p++)
			if (isalnum(*p))
			{
				c=0;
				str=buf;
				loc=i;
				while ((isalnum(*p) || ispunct(*p) || (*p == ' ')) && 
					   (c < MAX_STR))
				{
					*str++ = *p++; c++; i++;
				}
				if (c >= MIN_STR)
				{
					buf[c] = '\0';
					printf(mask, addr+loc);
					printf(" %s\n", buf);
				}
			}
	}
    // BUG: not finding strings that span blocks
}

unsigned long axton(char *s)
{
    if ((s[0] == '0') && (s[1] == 'x'))
    {
        unsigned long ret;
        sscanf(&s[2], "%lx", &ret);
        return ret;
    }
    else
        return atol(s);
}
        
