/* 
EXEDUMP.C
Andrew Schulman, 1992-1994

CHANGE LOG:
-- Fixed test for no code/data; was if (! (autodata || cs_ip));
   changed to:  if (! hdr.ne_cseg); 4/17/92  (1.31)
-- Added target OS; made version third field of target, and only
   if Win app; 4/18/92 (1.32)
-- Was doing test for 'NE' wrong!  Shouldn't just check if (start_newexe==0);
   first need to check that start_reloc at 18h is >= 40h (4/19/92)
-- rewrote usage() in English; added display of modref table; broke up
   newexe() function with print_modref, print_reloctab; 4/23/92 (1.35)
-- for segment size, use max(file_size, mem_size) instead of just
   assuming file_size.  Sometimes mem_size is bigger; 6/19/92 (1.36)
-- also for 1.36, made MAX_ENTRY bigger (8000); was too small?!
-- also for 1.36, take undoc -os2dd switch for use by WINP
-- check if Visual Basic program (MODNAME(0)=="VBRUN100"); 6/20/92 (1.37)
-- also for 1.37, change ne.h nr_segno, nr_res to _unsigned_ char; was
   wrong for any segno > 128!
-- changed message from "NT portable exe" to "Win32 portable exe"
-- with compressed reloc records, still try every seg; fail at end instead
   of with first error
-- major change for 1.38 (10/10/92):  doing entry table as hash table; can
   now do any ordinal number.  previously limited to about 8,000 as highest
   ordinal number.  now sparse!
-- also for 1.38, compiling as C++ program (bcc -P)	   
-- BC3.0 inline seems to have a bug
-- for 1.39 (10/28/92), fixed possible endless loop inside do_reloc 
   (following reloc chain)
-- also for 1.39, in segtab suppress segments with no file offset: ok?
-- for 2.0 (2/3/93), check for 'SZ' header (EXPAND)
-- 2.01 (8/2/93):  check for VBRUN rather than VBRUN100!!
-- remember to run patcher always!!
-- 5/11/94:  change message from PEDUMP to W32P and W32DUMP
-- 6/5/94: fail if modnum too high; call do_selfload
	   
EXEDUMP.C

Warnings:
MUST BE COMPILED LARGE MODEL!  bcc -ml exedump.c filelib.c
ASSUMES 16-BIT CODE ("UNSIGNED" ASSUMED UNSIGNED SHORT)

Apologies:

A lot of the complexity here (and some of the bugs in earlier versions!)
comes from trying to read in large chunks at one time from disk, and then
seek in memory, rather than doing small reads and seeking on disk.  Watch
the drive light to see the difference!

The other reason some of this code is so convoluted is that the NE
format seems to have been designed to minimize storage space on disk,
so several key structures are variable length.

<<FOLLOWING IS WRONG -- made print_modref, print_reloctab>>

Basic organization:
main 
|   newexe 
|   |   open 
|   |   exe_read
|   |   do_exports
|   |   |   do_enttab
|	|	|	|	ne_read
|   |   |   do_nres (description string comes from here)
|   |   |   |   exe_read 
|   |   |   do_resnam (module name comes from here)
|   |   |   |   ne_read
|   |   |   mangled_name 
|   |   do_segtab 
|   |   |   ne_read
|	|	has_null_seg
|	|	|	exe_read
|   |   do_modref 
|   |   |   ne_read
|   |   do_impnam 
|   |   |   ne_read
|   |   |   count2null 
|   |   do_cv_sym 
|   |   |   read
|   |   |   do_subsection 
|   |   |   |   read
|   |   do_seg 
|   |   |   exe_read
|	|	|	do_reloc
|   |   |   |	print_reloc 
|   |   |   |	|	in_entry_segofs 
|   |   |   |   |	print_reloc (recursive)
|   |   |   |   |	print_reloc_2 
|	|	|	do_gptab
|   |   close
*/

#include <stdlib.h>
#ifdef WINDOWS
#include "windows.h"
#include "winio.h"
#else
#include <stdio.h>
#endif
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include "ne.h"
#include "filelib.h"

#ifdef __cplusplus
// missing from \borlandc\include\stdlib.h
#define max(a,b)    (((a) > (b)) ? (a) : (b))
#endif

#if (sizeof(void*) != 4)
#error Must be compiled large model!
#endif

typedef int BOOL;
typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef void far *FP;

typedef struct entry_segofs {
	struct entry_segofs *next;
	WORD ord;
	WORD seg, ofs;
	char *name;
	} ENTRY_SEGOFS;

#define IMPNAME(x)	(&impnam[(x)+1])
#define MODNAME(x)	(strupr(IMPNAME(modref[x])))
#define FUNCNAME(x)	(get_funcname(x))
	
void failret(int ret, char *s)
{
	extern int export_only;
	extern int do_desc_only;
	if (! (export_only || do_desc_only))
		printf("exedump fail: %s\n", s);
	exit(ret); 
}

void fail(char *s) { failret(1, s); }

void fail_perror(void)	{ perror("exedump fail"); exit(1); }

void banner(void)
{
	fputs("Windows Segmented-Executable (NE) Dumper version 2.03\n", stderr);
	fputs("from \"Unauthorized Windows\" by Andrew Schulman "
		  "(IDG Books, 1994)\n", stderr);
	fputs("Copyright (c) 1992-94 Andrew Schulman.  All rights reserved.\n\n", 
		stderr);
}
	
void usage()
{
	banner();
	puts("EXEDUMP displays information about Windows and OS/2 1.x executables");
	puts("(EXEs, DLLs, DRVs, etc.).  It shows the entry table (exports),");
	puts("relocation table (imports), segment table, NE header, module-");
	puts("reference table, plus CodeView (R) information, if present.");
	puts("");
	puts("To display information about a Windows or OS/2 NE executable:");
	puts("         EXEDUMP exe_file");
	puts("example: exedump \\windows\\system\\krnl386.exe");
	puts("");
	puts("To generate a .DAT file listing of exports from an executable:");
	puts("         EXEDUMP -EXPORTS exe_file");
	puts("example: for %f in (\\windows\\system\\*.*) do exedump -exports %f >> winfunc.dat");
	puts("");
	puts("To display only the executable-file type (MZ NE LE W3 LX PE): "
		 "-MAGIC");
	puts("To display only the description string: -DESC");
	puts("To suppress the display of relocation information (imports): "
		 "-NORELOC");
	puts("To show blank and/or mangled CodeView symbols, "
		 "normally suppressed: -CVBLANKS");
	exit(1);
}
	
void newexe(char *name);
void print_reloc(WORD seg, WORD ofs, WORD type,
	WORD p1, WORD p2, RELOC_ITEM *r, BYTE *code);
void print_reloc_2(WORD ofs, RELOC_ITEM *r, BYTE *code);
void do_seg(int segnum, SEG *seg);
void do_reloc(int segnum, BYTE *code, RELOC_ITEM *rel,
	WORD nr_nreloc);
void do_gptab(BYTE *code);
void do_segtab(void);
void do_os2dd(void);
void do_exports(void);
void do_enttab(void);
void do_modref(void);
void do_impnam(void);
void do_cv_sym(void);
void do_nres(void);
void do_resnam(void);
int has_null_seg(void);
void dump(BYTE *p, WORD bytes, char *mask, DWORD addr);
void print_modref(void);
void print_reloctab(void);
char *get_magic_str(WORD magic);
void print_target_os(void);
void verify_if_ne(void);
BYTE *read_seg(int segnum);
void init_entry(void);
ENTRY_SEGOFS *get_entry(WORD ord);
void set_entry_segofs(WORD ord, WORD seg, WORD ofs);
void set_entry_name(WORD ord, char *name);
char *get_funcname(WORD ord);
void print_entrytab(void);
void do_selfload(void);

/* key global structures and variables */
DWORD start_newexe = 0;
static NEWEXE hdr = {0} ;
static SEG *segtab;
static WORD *modref;
static BYTE *impnam;
static char *this_modname;
static char *filename;
static WORD max_ord;
static DWORD align;
static int f;
static WORD gptab_seg = 0;	// seg:ofs of __GP table
static WORD gptab_ofs = 0;
static int first_code = 0;		// first code segment
static int first_data = 0;		// first data segment
static int comp_reloc_fail = 0;	// was there a compressed reloc failure?

/* command-line options */
static int verbose = 0;
static int export_only = 0;
static int no_reloc = 1;		/// turned around!! ///
static int do_all = 1;
static int no_banner = 0;
static int do_cv_blanks = 0;
static int do_type_only = 0;
static int do_desc_only = 0;
static int assume_ne = 0;
static int os2dd = 0;

/* arbitrary limits */
#define MAX_IMPNAME		8092

main(int argc, char *argv[])
{
	char *arg;
	int i;
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			if (argv[i][1] == '?')
				usage();
			
			arg = strupr(&argv[i][1]);
			
			if (strncmp(arg, "VERBOSE", 4) == 0)
				verbose = 1;
			else if (strncmp(arg, "EXPORTS", 3) == 0)
			{
				export_only = 1;
				do_all = 0;
			}
			else if (strncmp(arg, "RELOC", 3) == 0)
				no_reloc = 0;
			else if (strncmp(arg, "NOBANNER", 3) == 0)
				no_banner = 1;
			else if (strncmp(arg, "CVBLANKS", 3) == 0)
				do_cv_blanks = 1;
			else if (strncmp(arg, "MAGIC", 3) == 0)
			{
				no_banner = 1;
				do_type_only = 1;
			}
			else if (strncmp(arg, "DESCRIPTION", 4) == 0)
			{
				no_banner = 1;
				do_desc_only = 1;
			}
			else if (strncmp(arg, "ASSUME", 3) == 0)
				assume_ne = 1;
			else if (strncmp(arg, "OS2DD", 5) == 0)
				os2dd = 1;
			else
				usage();
		}
		else
		{
			char buf[128];
			if (! no_banner)
				banner();
			strcpy(buf, argv[i]);
			if (! strchr(buf, '.'))		// sort-of check for .ext
				strcat(buf, ".exe");
			newexe(buf);
			exit(0);
		}
	return 0;
}

void newexe(char *name)
{
    WORD w;
	WORD i;
	int call_do_selfload = 0;
	
	filename = name;
	
	if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
		fail_perror();
	if (! assume_ne)
	{
		exe_read(f, &w, 0, 2);	// read first two bytes of file
		if (w != EMAGIC)
		{
			if (w == SZMAGIC)
				fail("compressed file (SZ) -- use EXPAND.EXE");
			else
				fail("not an executable");
		}
		exe_read(f, &w, 0x18, 2);
		if (w < 0x40)
		{
			if (do_type_only)
			{
				puts("MZ -- not a new executable");
				exit(0);
			}
			else
				fail("not a segmented (new) executable");
		}
		exe_read(f, &start_newexe, ENEWHDR, 4);
	}
	else
		start_newexe = 0;	// some kind of extracted file with NE at start
			
	ne_read(f, &hdr, 0, sizeof(NEWEXE));
	
	if (do_type_only)
	{
		puts(get_magic_str(hdr.ne_magic));
		exit(0);
	}

	if (! assume_ne)
		verify_if_ne();
	
	align = 1 << hdr.ne_align;
	
	if (do_all)
	{
		if (! do_desc_only)
		{
			printf("name %s\n", name);

			printf("%s\n", hdr.ne_flags & DLL_FLAG ? "library" : "program");

			if (hdr.ne_flags & APPLOAD)
			{
				puts("Self-loading");
				call_do_selfload++;
			}

			print_target_os();

			// not reliable -- OS/2 only????
			printf("80%d instructions",
				(hdr.ne_flags & NEI386) ? 386 :
				(hdr.ne_flags & NEI286) ? 286 :
				(hdr.ne_flags & NEI086) ? 86 :
				/* default */			  286);

			if (hdr.ne_flags & NEMATH)
				printf(" (MATH)");		// appears to never be used!
			printf("\n");

			printf("dgroup %u\n", hdr.ne_autodata);

			printf("start_csip %u:%04x\n", 
				FP_SEG(hdr.ne_csip),
				FP_OFF(hdr.ne_csip));
			
			printf("start_newexe %lx\n", start_newexe);
		}

		do_exports();
		do_modref();
		do_impnam();
		
		if (call_do_selfload)
			do_selfload();
		
		// see if Visual Basic .EXE
		if (strstr(strupr(MODNAME(0)), "VBRUN") != 0)	// VBRUN100, VBRUN200
			puts("Visual Basic program -- code kept as USERDATA resource\n");
		
		do_segtab();
		
		if (os2dd)		// is_os2dd not good yet
			do_os2dd();
		
		// must do test after set up segtab
		if (hdr.ne_autodata)		// if has DGROUP
			if (! has_null_seg())	// but doesn't have NULL segment
				puts("No NULL segment\n");

		do_cv_sym();
	}
	else if (export_only)
	{
		if (! (hdr.ne_flags & DLL_FLAG))
			exit(1);
		printf(";\n; %s\n", name);
		do_exports();
		exit(0);
	}
	
	print_modref();
	
	if (! no_reloc)
		print_reloctab();
	
	close(f);
}

void verify_if_ne(void)
{
	switch (hdr.ne_magic)
	{
		case LEMAGIC:
			// can't change this message, because WINP relies on it!!
			failret(2, "Use LEDUMP to examine this file");
			break;
		case LXMAGIC:
			fail("sorry, this is an OS/2 2.0 linear executable, not "
				 "a segmented executable");
			break;
		case W3MAGIC:
			// can't change this message, because WINP relies on it!!
			failret(3, "Use W3MAP to examine this file");
			break;
		case PEMAGIC:
			fail("use W32P or W32DUMP to examine this Win32 Portable Executable");
			break;
		case NEMAGIC:
			// this is one we want, except if no CODE or DATA
			if (! hdr.ne_cseg)
				failret(4, "No CODE/DATA; use RESDUMP to examine this file");
			break;
		default:
			fail("not a segmented (new) executable");
	}
}

void print_target_os(void)
{
	if (hdr.ne_exetyp < 5)
	{
		static char *target_os[] = {
			"UNKNOWN", 
			"OS2", 
			"WINDOWS", 
			"DOS4 (or 286|DOS-Extender real-mode DLL)", 
			"WIN386"
			} ;
		printf("target %s", target_os[hdr.ne_exetyp]);
		if (hdr.ne_exetyp == 2)		// Windows
		{
			printf(" %d.%02d",
				hdr.ne_expwinvermaj, hdr.ne_expwinvermin);
		}
		printf("\n");
	}
	else
		printf("target ???\n");
}

char *get_magic_str(WORD magic)
{
	switch (magic)
	{
		case LEMAGIC: return "LE -- Linear Executable (Windows VxD)";
		case LXMAGIC: return "LX -- OS/2 2.x Linear Executable";
		case W3MAGIC: return "W3 -- WIN386 (collection of LE files)";
		case PEMAGIC: return "PE -- Win32 Portable Executable";
		case NEMAGIC: return "NE -- New (Segmented) Executable";
		default:      return "MZ -- not a new executable";
	}
}
		
void do_modref(void)
{
	if (! hdr.ne_cmod) return;
	modref = (WORD *) allocate(hdr.ne_cmod, sizeof(WORD));
	ne_read(f, modref, hdr.ne_modref, hdr.ne_cmod * sizeof(WORD));
}

void do_impnam(void)
{
	impnam = (BYTE *) allocate(MAX_IMPNAME, 1);	
	ne_read(f, impnam, hdr.ne_impname, MAX_IMPNAME);
	count2null(impnam, MAX_IMPNAME);
}
		
void do_segtab(void)
{
	SEG *seg;
	WORD i;
	
	if (! hdr.ne_cseg) return;
	segtab = (SEG *) allocate(hdr.ne_cseg, sizeof(SEG));
	ne_read(f, segtab, hdr.ne_segtab, hdr.ne_cseg * sizeof(SEG));
	
	printf("begin segtab\n");
	for (i=0, seg=segtab; i<hdr.ne_cseg; i++, seg++)
	{
		if ((seg->ns_flags & NSTYPE) == NSCODE)
		{
			if (! first_code)
				first_code = i+1;
		}
		else
		{
			if (! first_data)
				first_data = i+1;
		}

		if (seg->ns_sector)		// sometimes 0!
		printf("\t%u\t%lx\t%04x\t%s\t%s\n",
			i+1,
			align * seg->ns_sector,
			max(seg->ns_cbseg, seg->ns_minalloc),
			((seg->ns_flags & NSTYPE) == NSCODE) ? "CODE" : "DATA",
			(seg->ns_flags & NS32BIT) ? "USE32" : "USE16");
	}
	putchar('\n');
}

int is_os2dd(void)	// not a very good test yet!
{
	if ((hdr.ne_exetyp == 1) &&			// os2 file
		(first_data != 0) &&			// has a data segment
		(first_code != 0))				// has a code segment
	{
		// it could be an OS/2 device driver -- now look for dev chain
		SEG *seg = &segtab[first_data-1];
		BYTE *data;
		WORD *p = (WORD *) data;
		int isdd = 1;
		int numdev = 1;
#if 1
		data = read_seg(first_data);
#else
		// this thing needs a read_seg() primitive!
		data = allocate(seg->ns_cbseg, 1);
		exe_read(f, data, align * seg->ns_sector, seg->ns_cbseg);
#endif		
		while (*p && (*p != 0xffff))
		{
			numdev++;
			if (*p < seg->ns_cbseg)
				p = (WORD *) &data[*p];		// follow possible chain
			else
			{
				isdd = 0;
				break;
			}
		}
		free(data);
		if (numdev > 5)		// probably not devs
			return 0;
		return isdd;
	}
	else
		return 0;
}

void do_os2dd(void)
{
	WORD dataseg;
	SEG *seg;
	BYTE *data;
	WORD *p;
	WORD next, addr;		// ignore seg part of next?
	int devnum;
	
	dataseg = (hdr.ne_autodata) ? hdr.ne_autodata : first_data ;
	if (! dataseg)
		return;

	seg = &segtab[dataseg-1];
#if 1
	data = read_seg(dataseg);
#else
	data = allocate(seg->ns_cbseg, 1);
	exe_read(f, data, align * seg->ns_sector, seg->ns_cbseg);
#endif

	puts("begin os2dd");
	next = 0;
	devnum = 0;
	addr = 0;
	do {
		p = (WORD *) &data[next];
		addr = next;	// before change next
		next = *((WORD *) &p[0]);
		if ((next == 0) || (next == addr) ||
			((next != -1) && (next > seg->ns_cbseg)))
				break;	// can't be valid
		devnum++;
		printf("\t%d\t%04X\t%04X (%s)\t%04X",
			devnum,
			addr,
			p[2],		//dev attr
			(p[2] & 0x8000) ? "CHAR" : "BLOCK",
			p[3]);		//Strategy
		if (p[2] & 0x4000)	// IDC bit
			printf("\t%04X", p[4]);	// inter-device communication
		else
			printf("\t----");
		if (p[2] & 0x8000)	// char devs have names
			printf("\t\'%.08s\'", (char *) &p[5]);
		putchar('\n');
	} while (next != -1);
	putchar('\n');
	free(data);
}

/**********************************************************************/
/* from Microsoft C Developer's Toolkit Reference, Chapter 3, 
   "Extended .EXE Format for Debug Information" */

typedef struct {
	BYTE sign[4];
	DWORD offset;
	} CV_HDR;
	
typedef struct {
	WORD SubSection;
	WORD iMod;
	signed long lfo;
	WORD cb;
	} DNT;
	
typedef struct {
	WORD cbDirHeader;
	WORD cbDirEntry;
	DWORD cDir;
	DWORD lfoNextDir;
	DWORD flags;
	} DNT7_HDR;
	
typedef struct {
	WORD SubSection;
	WORD iMod;
	signed long lfo;
	DWORD cb;
	} DNT7;
	
typedef struct {
	WORD seg;
	WORD ofs;
	WORD len;
	WORD overlay;
	WORD index;
	BYTE  cseg;
	BYTE  reserved;
	BYTE  str[1];
	} MOD;
	
typedef struct {
	WORD seg, pad;
	DWORD offset, cbSeg;
	} SEGINFO;	// C7
	
typedef struct {
	WORD ovlNumber;
	WORD iLib;
	WORD cSeg;
	WORD Style;
	// SEGINFO seginfo[cSeg]
	// BYTE name[1];
	} MOD_C7;
	
typedef struct {
	WORD ofs;
	WORD seg;
	WORD typ;
	BYTE str[1];
	} PUB;
	
typedef struct {
	WORD unknown1;
	WORD unknown2;
	WORD ofs;
	WORD seg;		// often out of range?!?!
	WORD typ;
	BYTE str[1];
	// plus pad to align next on long
	} PUB_C7;
	
typedef struct {
	WORD symhash, addrhash;
	DWORD cbSymbol, cbSymHash, cbAddrHash;
	} GLOBPUB;	// C7
	
#define SSTMODULES		0x101
#define SSTPUBLIC		0x102
#define SSTTYPES		0x103
#define SSTSYMBOLS		0x104
#define SSTSRCLINES		0x105
#define SSTLIBRARIES	0x106
#define SSTCOMPACTED	0x107
#define SSTSRCLNSEG		0x108

// C7 types
#define SSTMODULE_C7	0x120
#define SSTTYPES_C7		0x121
#define SSTPUBLIC_C7	0x122
#define SSTPUBLICSYM_C7 0x123
#define SSTSYBOLS_C7	0x124
#define SSTALIGNSYM_C7	0x125
#define SSTRCLNSEG_C7	0x126
#define SSTSRCMODULE_C7 0x127
#define SSTLIBRARIES_C7 0x128
#define SSTGLOBALSYM_C7	0x129
#define SSTGLOBALPUB_C7 0x12A
#define SSTGLOBALTYPES_C7 0x12b
#define SSTMPC_C7		0x12c
// have seen 0x12d ?!

void do_subsection(DNT *pdnt);

static DWORD base;

void do_cv_sym(void)		// CodeView symbols
{
	CV_HDR cv_hdr;
	DWORD filesize;
	DWORD lfoDIR;
	
	// go to last eight bytes of file, look for NB0 signature
	exe_read(f, &cv_hdr, filelength(f)-8, sizeof(CV_HDR));
	if (memcmp(cv_hdr.sign, "NB0", 3) != 0)
		return;	// no CodeView debug information
	
	// go to CV info base address, look for NB0 signature again
	base = filelength(f) - cv_hdr.offset;
	exe_read(f, &cv_hdr, base, sizeof(CV_HDR));
	if (memcmp(cv_hdr.sign, "NB0", 3) != 0)
		return;	// error finding second CV debug signature
	if (cv_hdr.offset == -1L)
		return;	// not CodeView:  Turbo Debug?
	
	printf("begin CodeView\n");

	lfoDIR = base + cv_hdr.offset;
	
	if (memcmp(cv_hdr.sign, "NB07", 4) == 0)	// C7
	{	// PE use NB08???
		DNT7_HDR *pdnt7hdr;
		DNT7 *pdnt7, *p7;
		DWORD ofs;
		int i;
		
		pdnt7hdr = (DNT7_HDR *) allocate(1, sizeof(DNT7_HDR));
		
		while (lfoDIR)
		{
			exe_read(f, pdnt7hdr, lfoDIR, sizeof(DNT7_HDR));
			ofs = lfoDIR + pdnt7hdr->cbDirHeader;	// skip past header
			if (pdnt7hdr->cbDirEntry < sizeof(DNT7))
			{
				printf("\t; C7 CodeView size inconsistency?\n\n");
				return;
			}
			pdnt7 = (DNT7 *) allocate(1, pdnt7hdr->cbDirEntry);
			for (i=0; i<pdnt7hdr->cDir; i++)
			{
				exe_read(f, pdnt7, ofs, pdnt7hdr->cbDirEntry);
				do_subsection((DNT *) pdnt7);
				ofs += pdnt7hdr->cbDirEntry;
			}
			free(pdnt7);
			lfoDIR = pdnt7hdr->lfoNextDir;	// 0 for last one
		}
	
		free(pdnt7hdr);
	}
	else
	{
		DNT *pdnt, *p;
		WORD cdnt;
		int i;
		
		// find the SubSection directory
		exe_read(f, &cdnt, lfoDIR, sizeof(cdnt));	// number of dir. entries

		// read the dnt
		pdnt = (DNT *) allocate(cdnt, sizeof(DNT));
		read(f, pdnt, cdnt * sizeof(DNT));
		for (i=0, p=pdnt; i<cdnt; i++, p++)
			do_subsection(p);
		printf("\n");
		free(pdnt);
	}
	
	printf("\n");
}

int allspace(char *s)
{
	int saw_space = 0;
	while (*s)
		if (isspace(*s))
		{
			s++;
			saw_space = 1;
		}
		else if (saw_space && isdigit(*s))	/* e.g., "       193" */
			return 1;
		else
			return 0;
	/* still here */
	return 1;
}

char *get_typ_name(WORD type)
{
	static char typbuf[16];
	char *typ;
	if ((type & 0x80) == 0)	// special type
	{
		if (type == 1)
			return "(TYP ABS)";	// T_ABS
		else
			return "";	// T_NOTYPE
	}
	else
	switch (type)
	{
		case 0x80:	typ = "(CHAR)"; break;
		case 0x81:	typ = "(SHORT)"; break;
		case 0x82:	typ = "(LONG)"; break;
		case 0x84:	typ = "(UCHAR)"; break;
		case 0x85:	typ = "(USHORT)"; break;
		case 0x86:	typ = "(ULONG)"; break;
		case 0x88:	typ = "(FLOAT)"; break;
		case 0x89:	typ = "(DOUBLE)"; break;
		case 0x90:	typ = "(BOOL)"; break;
		case 0x91:	typ = "(BOOL16)"; break;
		case 0x92:	typ = "(BOOL32)"; break;
		case 0x94:	typ = "(ASCII)"; break;
		case 0x95:	typ = "(ASCII16)"; break;
		case 0x96:	typ = "(ASCII32)"; break;
		case 0x97:	typ = "(STRING)"; break;

		case 0xa0:	typ = "(CHAR NEAR *)"; break;
		case 0xa1:	typ = "(SHORT NEAR *)"; break;
		case 0xa2:	typ = "(LONG NEAR *)"; break;
		case 0xa4:	typ = "(UCHAR NEAR *)"; break;
		case 0xa5:	typ = "(USHORT NEAR *)"; break;
		case 0xa6:	typ = "(ULONG NEAR *)"; break;
		case 0xa8:	typ = "(FLOAT NEAR *)"; break;
		case 0xa9:	typ = "(DOUBLE NEAR *)"; break;
		case 0xb0:	typ = "(BOOL NEAR *)"; break;
		case 0xb1:	typ = "(BOOL16 NEAR *)"; break;
		case 0xb2:	typ = "(BOOL32 NEAR *)"; break;
		case 0xb4:	typ = "(ASCII NEAR *)"; break;
		case 0xb5:	typ = "(ASCII16 NEAR *)"; break;
		case 0xb6:	typ = "(ASCII32 NEAR *)"; break;
		case 0xb7:	typ = "(STRING NEAR *)"; break;

		case 0xc0:	typ = "(CHAR FAR *)"; break;
		case 0xc1:	typ = "(SHORT FAR *)"; break;
		case 0xc2:	typ = "(LONG FAR *)"; break;
		case 0xc4:	typ = "(UCHAR FAR *)"; break;
		case 0xc5:	typ = "(USHORT FAR *)"; break;
		case 0xc6:	typ = "(ULONG FAR *)"; break;
		case 0xc8:	typ = "(FLOAT FAR *)"; break;
		case 0xc9:	typ = "(DOUBLE FAR *)"; break;
		case 0xd0:	typ = "(BOOL FAR *)"; break;
		case 0xd1:	typ = "(BOOL16 FAR *)"; break;
		case 0xd2:	typ = "(BOOL32 FAR *)"; break;
		case 0xd4:	typ = "(ASCII FAR *)"; break;
		case 0xd5:	typ = "(ASCII16 FAR *)"; break;
		case 0xd6:	typ = "(ASCII32 FAR *)"; break;
		case 0xd7:	typ = "(STRING FAR *)"; break;

		default:
			sprintf(typbuf, "(TYP %04Xh)", type);
			typ = typbuf;
	}
	return typ;
}

void do_subsection(DNT *pdnt)
{
	MOD *pmod;
	GLOBPUB *pglobpub;
	PUB *ppub, *psav;
	PUB_C7 *ppubc7;
	char *typ;
	char name[128];
	WORD offset;
	
#if 0	
	printf("\t; SS=%04x cb=%04x lfo=%08lx\n", 
		pdnt->SubSection, pdnt->cb, base+pdnt->lfo);
#endif	
	
	if (pdnt->cb == 0)
		return;
	
	switch (pdnt->SubSection)
	{
		case SSTMODULE_C7:
			{
				MOD_C7 *pmodc7;
				DWORD ofs;
				char *name;
				WORD len = 0;

				pmodc7 = (MOD_C7 *) allocate(1, pdnt->cb);
				ofs = base + pdnt->lfo;
				exe_read(f, pmodc7, ofs, pdnt->cb);
				ofs += sizeof(MOD_C7) + (pmodc7->cSeg * sizeof(SEGINFO));
				exe_read(f, &len, ofs, 1);	// get string length
				name = (char *) allocate(1, len+1);
				ofs++;
				exe_read(f, name, ofs, len);	// get name
				name[len] = '\0';
				if (pmodc7->cSeg)
				{
					// now read in SEGINFO blocks
					SEGINFO si;
					int i;
					ofs = base + pdnt->lfo + sizeof(MOD_C7);
					for (i=0; i<pmodc7->cSeg; i++)
					{
						exe_read(f, &si, ofs, sizeof(SEGINFO));
						ofs += sizeof(SEGINFO);
						printf("\t%d\t%04lx\t%04lx\t%s (MODULE)\n", 
							si.seg, si.offset, si.cbSeg, name);
					}
				}
				else
					;	// printf("\t%s\n", name); -- not useful (API names)
				free(name);
				free(pmodc7);
			}
			break;
		case SSTMODULES:
			{
				pmod = (MOD *) allocate(1, pdnt->cb);
				exe_read(f, pmod, base + pdnt->lfo, pdnt->cb);
				if (pmod->cseg > 1)
					;	// ok?
				strncpy(name, &pmod->str[1], pmod->str[0]);
				name[pmod->str[0]] = '\0';
				printf("\t%d\t%04x\t%04x\t%s (MODULE)\n", 
					pmod->seg, pmod->ofs, pmod->len, name);
				free(pmod);
			}
			break;
		case SSTGLOBALPUB_C7:
			{
				pglobpub = (GLOBPUB *) allocate(1, sizeof(GLOBPUB));
				exe_read(f, pglobpub, base + pdnt->lfo, sizeof(GLOBPUB));
				if (pglobpub->cbSymbol > 0xFF00UL)
				{
					// just for now!
					printf("\t; >64k symbols in C7 CodeView -- overflow\n");
					pglobpub->cbSymbol = 0xFF00UL;
				}
				ppub = (PUB *) allocate(1, pglobpub->cbSymbol);
				exe_read(f, ppub, base + pdnt->lfo + sizeof(GLOBPUB),
					pglobpub->cbSymbol);
#if 1
				ppubc7 = (PUB_C7 *) ppub;
				offset = 0;
				while (offset < pglobpub->cbSymbol)
				{
					WORD incr;
					strncpy(name, &ppubc7->str[1], ppubc7->str[0]);
					name[ppubc7->str[0]] = '\0';
					printf("\t%d\t%04x\t\t%s %s\n",
						ppubc7->seg, ppubc7->ofs, name, 
						get_typ_name(ppubc7->typ));
					incr = sizeof(PUB_C7) + ppubc7->str[0];
					while (incr % 4) incr++;	// pad out
					offset += incr;
					((BYTE *) ppubc7) += incr;
				}
#else
				dump(ppub, pglobpub->cbSymbol, "%08lx", 
					base + pdnt->lfo + sizeof(GLOBPUB));
#endif				
				free(ppub);
				free(pglobpub);
			}
			break;
		case SSTPUBLIC:
		case SSTPUBLIC_C7:
		case SSTPUBLICSYM_C7:
			{
				WORD incr;
				ppub = psav = (PUB *) allocate(1, pdnt->cb);
				offset = 0;
				exe_read(f, ppub, base + pdnt->lfo, pdnt->cb);
				while (offset < pdnt->cb)
				{
					strncpy(name, &ppub->str[1], ppub->str[0]);
					name[ppub->str[0]] = '\0';

					typ = get_typ_name(ppub->typ);

					if (! allspace(name))	// don't print blank symbols!
						printf("\t%d\t%04x\t\t%s %s\n",
							ppub->seg, ppub->ofs, name, typ);
					else if (do_cv_blanks)	// show blank symbols
						printf("\t\t\t\t[%d spaces]\n",
							strlen(name));

					incr = sizeof(PUB) + ppub->str[0];
					offset += incr;
					((BYTE *) ppub) += incr;
				}
				free(psav);
			}
			break;
		case SSTMPC_C7:
			printf("\t; MPC -- Executable contains Pcode\n");
			break;
	}
}

/**********************************************************************/
	
int mangled_name(char *s)	// look for mangled names like "mmmmmm"
{
	int c = *s;
	char *s2 = s;
	int len = strlen(s);
	if (len < 3)
		return 0;
	while (*s2 == c)
		s2++;
	return (s + len == s2);
}

void print_modref(void)
{
	int i;
	printf("begin modref\n");
	for (i=0; i<hdr.ne_cmod; i++)
		printf("\t%s\n", MODNAME(i));
	printf("\n");
}

void do_exports(void)
{
	ENTRY_SEGOFS *eso;
	WORD i;
	
	init_entry();
	
	// walk the entry table (bundles)
	do_enttab();

	// walk non-resident name table; module description from first entry
	do_nres();
	
	// walk resident names table; module name from first entry
	do_resnam();
	
	print_entrytab();
		
	// we've got everything:  now print it out
	if (do_all)
	{
		printf("modname %s\n\n", this_modname);
		
		printf("begin entry\n");
		for (i=1, eso=get_entry(i); i<=max_ord; i++, eso=get_entry(i))
			if (eso && eso->seg)
			{
				printf("\t%u\t%04x\t%u", 
					eso->seg, eso->ofs, i);
				if (eso->name)
					printf("\t%-30s", eso->name);
				if (strcmp(eso->name, "__GP") == 0)
				{
					gptab_seg = eso->seg;
					gptab_ofs = eso->ofs;
				}
				putchar('\n');
			}
		putchar('\n');
	}
	else if (export_only)
	{
		for (i=1, eso=get_entry(i); i<=max_ord; i++, eso=get_entry(i))
			if (eso && eso->seg && eso->name)
			{
#ifdef WINDOWS
				printf("%s . %u\t%s\n", this_modname, i, eso->name);
#else
				if (mangled_name(eso->name))
					fprintf(stderr, "Warning: mangled name? "
									"%s.%u: \"%s\"\n",
								this_modname, i, eso->name);
				else
					printf("%s . %u\t%s\n", this_modname, i, eso->name);
#endif				
			}
	}
}
					
void do_enttab(void)
{
	ENTRY_BUNDLE *enttab, *ep;
	WORD eb;
	WORD ord = 0;
	
	if (! hdr.ne_cbenttab) return;
		/* number BYTES in entry table */ 
	enttab = (ENTRY_BUNDLE *) allocate(hdr.ne_cbenttab, 1);
	ne_read(f, enttab, hdr.ne_enttab, hdr.ne_cbenttab);
	for (eb=0, ep=enttab; ; )
	{
		BUNDLE_HEADER bhdr = ep->eb_hdr;
		if (bhdr.eb_nentry == 0)		// end of table
			break;
		else if (bhdr.eb_segnum == 0)	// null entry
		{
			((BYTE *) ep) += sizeof(BUNDLE_HEADER);
			ord += bhdr.eb_nentry;		// for ordinal numbers
			continue;
		}
		eb++;
		((BYTE *) ep) += sizeof(BUNDLE_HEADER);
		if (bhdr.eb_segnum == MOVEABLE)
		{
			WORD i;
			ENTRY_MOVEABLE *em;
			for (i=0, em=(ENTRY_MOVEABLE *) ep; i<bhdr.eb_nentry; i++, em++)
			{
				ord++;
				set_entry_segofs(ord, em->ep_segnum, em->ep_offset);
			}
			((BYTE *) ep) += 
				(bhdr.eb_nentry * sizeof(ENTRY_MOVEABLE));
		}
		else							// fixed
		{
			WORD i;
			ENTRY_FIXED *ef;
			for (i=0, ef= (ENTRY_FIXED *) ep; i<bhdr.eb_nentry; i++, ef++)
			{
				ord++;
				set_entry_segofs(ord, bhdr.eb_segnum, ef->ep_offset);
			}
			((BYTE *) ep) += 
				(bhdr.eb_nentry * sizeof(ENTRY_FIXED));
		}
	}
	max_ord = ord;
	free(enttab);	// temp:  we don't need it
}

void do_nres(void)
{
	BYTE *nres = (BYTE *) allocate(hdr.ne_cbnrestab, 1);
	BYTE *nr;
	WORD i;
	BYTE c;
	WORD ndx;
	
	if (! hdr.ne_cbnrestab) return;
	exe_read(f, nres, hdr.ne_nonrestab, hdr.ne_cbnrestab);
	
	for (i=0, nr=nres; i<hdr.ne_cbnrestab; )
	{
		if (! (c = nr[0]))
			break;
		
		ndx = *((WORD *) &nr[c+1]);		// ordinal number
		nr[c+1] = '\0';		// terminate string, replacing ord#
			
		if ((ndx == 0) && nr[1])
		{
			if (do_desc_only)
			{
				puts(filename);
				// puts(this_modname);    don't have it yet!
				puts(&nr[1]);
				puts("");
				exit(0);
			}
			else if (do_all)
				printf("description \"%s\"\n", &nr[1]);
			else if (export_only)
				printf("; \"%s\"\n;\n", &nr[1]);	// show desc as comment
		}
		else
		{
			set_entry_name(ndx, &nr[1]);
		}
		
		nr += (c+3);
		i += (c+3);
	}
}

void do_resnam(void)
{
	WORD res;
	BYTE c;
	BYTE *s;
	WORD ndx;
	
	for (res=hdr.ne_restab;;)
	{
		ne_read(f, &c, res, 1);
		if (c == 0)
			break;
		s = (BYTE *) allocate(c+2, 1);
		ne_read(f, s, ++res, c+2);
		ndx = *((WORD *) &s[c]);		// WORD, not a BYTE!
		s[c] = '\0';			// now turn ord# into NULL terminate
		if (ndx == 0)			// module name
			this_modname = s;
		else
		{
			set_entry_name(ndx, s);
		}
		res += (c+2);
	}
}

/**********************************************************************/

BYTE *read_seg(int segnum)
{
	SEG *seg;
	BYTE *p;
	segnum--;
	if (segnum > hdr.ne_cseg)
		fail("internal error: segnum too big");
	seg = &segtab[segnum];
	p = (BYTE *) allocate(1, seg->ns_cbseg);
	exe_read(f, p, align * seg->ns_sector, seg->ns_cbseg);
	return p;		// caller must use free()
}

void do_seg(int segnum, SEG *seg)
{
	/* multiply seg->ns_sector by hdr.ne_align to get _file_ offset; 
	   _not_ NE offset */
	DWORD reloc = (align * seg->ns_sector) + seg->ns_cbseg;
	BYTE *code;
	RELOC_ITEM *rel;
	WORD nr_nreloc;
	WORD i;

	exe_read(f, &nr_nreloc, reloc, sizeof(WORD));
	if (nr_nreloc == 0)
		return;
	if (nr_nreloc > (0xFFFFU / sizeof(RELOC_ITEM)))
	{
		comp_reloc_fail++;
		return;
	}
	
#if 1	
	code = read_seg(segnum);
#else
	code = allocate(seg->ns_cbseg, 1);
	exe_read(f, code, align * seg->ns_sector, seg->ns_cbseg);
#endif	
	rel = (RELOC_ITEM *) allocate(nr_nreloc, sizeof(RELOC_ITEM));
	exe_read(f, rel, reloc+2, nr_nreloc * sizeof(RELOC_ITEM));
	do_reloc(segnum, code, rel, nr_nreloc);
	if (segnum == gptab_seg)
		do_gptab(code);
	free(code);
	free(rel);
}

int has_null_seg(void)
{
	WORD w = 0xFFFF;
	SEG *seg;
	if (! hdr.ne_autodata) 
		return 0;	// no DGROUP, so no NULL seg
	seg = &segtab[hdr.ne_autodata - 1];
	exe_read(f, &w, align * seg->ns_sector, 2);
	return (w == 0);
}

typedef struct {
	WORD seg, low, high, func;
	} GPTAB;

void do_gptab(BYTE *code)
{
	GPTAB *__GP;
	GPTAB *p;
	int i;

	__GP = (GPTAB *) &code[gptab_ofs];
	for (p=__GP, i=gptab_ofs; p->seg; p++, i += sizeof(GPTAB))
		printf("\t%d\t%04x\t[g] %04x %04x\n",
			p->seg, p->func, p->low, p->high);
}

void print_reloctab(void)
{
	SEG *seg;
	int i;
	
	printf("begin reloc\n");

	printf("\t%u\t%04x\t[e] start\n",	// program entry point
		FP_SEG(hdr.ne_csip),
		FP_OFF(hdr.ne_csip));

	for (i=0, seg=segtab; i<hdr.ne_cseg; i++, seg++)
		if (seg->ns_flags & NSRELOC)
			do_seg(i+1, seg);

	putchar('\n');

	if (comp_reloc_fail)
		fail("compressed relocation records?");
}
	
void do_reloc(int segnum, BYTE *code, RELOC_ITEM *rel,
	WORD nr_nreloc)
{
	int i;
	for (i=0; i<nr_nreloc; i++)
	{
		RELOC_ITEM *r = &rel[i];
		WORD next;
		int ordnum, segno;

		WORD tmp = *((WORD *) &code[r->nr_soff]);
		
		/*
			relocation prefix:
				[i]		module.funcname
				[m] 	module.ordinal	
				[n]		funcname // thismod.ord
				[s] 	segment:offset
		*/

		switch (r->nr_flags & 0x03)
		{
			case NRRINT:
				ordnum = r->nr_union.nr_intref.nr_entry;
				segno = r->nr_union.nr_intref.nr_segno;
				if (segno == 0xFF)	// MOVEABLE
				{
					if (FUNCNAME(ordnum))
						print_reloc(segnum, r->nr_soff, 
							'n', 0, ordnum,	r, code);
					else
						print_reloc(segnum, r->nr_soff, 
							'm', -1	/* this_modname */, ordnum, r, code);
				}
				else	// fixed
				{
					print_reloc(segnum, r->nr_soff, 
						's', segno, ordnum, r, code);
				}
				break;
			case NRRORD:
				print_reloc(segnum, r->nr_soff, 
					'm', 
					r->nr_union.nr_import.nr_mod,
					r->nr_union.nr_import.nr_proc,
					r, code);
				break;
			case NRRNAM:
				print_reloc(segnum, r->nr_soff, 
					'i', 
					r->nr_union.nr_import.nr_mod,
					r->nr_union.nr_import.nr_proc,
					r, code);
				break;
			case NRROSF:
				/* do nothing: ok as is */
				break;
		}

		/* follow possible relocation chain */
#if 1
		if (! (r->nr_flags & NRADD))
		{
			next = tmp;
			while (next != 0xFFFF)
			{
				tmp = *((WORD *) &code[next]);
				if (tmp == next)	// hack for endless cycle!!
					break;
				print_reloc(segnum, next, -1, 0, 0, 0, code);
				if ((next = tmp) == 0)
					break;
			}
		}
#else
		if (! (r->nr_flags & NRADD))
		{
			next = *((WORD *) &code[r->nr_soff]);
			while (next != 0xFFFF)
			{
				print_reloc(segnum, next, -1, 0, 0, 0, code);
				next = *((WORD *) &code[next]);
			}
		}
#endif		
	}
}

WORD in_entry_segofs(WORD seg, WORD ofs)
{
	WORD i;
	ENTRY_SEGOFS *eso;
//	printout helps establish that only called when necessary
//	printf("in_entry_so %04x:%04x\n", seg, ofs);
	for (i=1, eso=get_entry(i); i<=max_ord; i++, eso=get_entry(i))
		if (eso && (seg == eso->seg) && (ofs == eso->ofs))
			return i;
	/* still here */
	return 0;
}
		
void print_reloc(WORD seg, WORD ofs, WORD type,
	WORD p1, WORD p2, RELOC_ITEM *r, BYTE *code)
{
	static WORD s_type, s_p1, s_p2;
	static RELOC_ITEM *s_r;
	WORD ord;
	
	if (type == (WORD) -1)
	{	/* use same as last time, except for seg and ofs */
		type=s_type; p1=s_p1; p2=s_p2; r=s_r;
	}
	else
	{	/* new relocation chain	*/  
		s_type=type; s_p1=p1; s_p2=p2; s_r=r;
	}
	
	if (type == 's')
	{
		/*
			Following looks wrong.  Just to keep compiler happy 
			for now.  Casting NRSPTR to RELOC_ITEM* ??
		*/
		if ((r == (RELOC_ITEM *) NRSPTR) &&		// only valid for s:o 
			(ord = in_entry_segofs(p1, p2)))
		{
			// turning [s] into [n]
			print_reloc(seg, ofs, 'n', 0, ord, r, code);
			return;		// NB! don't print reloc type again!
		}
		else
			printf("\t%u\t%04x\t[%c] %u %04x", seg, ofs, type, p1, p2);
	}
	else
		printf("\t%u\t%04x\t[%c] ", seg, ofs, type);

	if (type == 'i')
	{
		if (p1 > hdr.ne_cmod) 
			fail("Corrupt or compressed file: module number too high");
		printf("%s %s", MODNAME(p1-1), IMPNAME(p2));
	}
	else if (type == 'm')
	{
		if (p1 > hdr.ne_cmod) 
			fail("Corrupt or compressed file: module number too high");
		printf("%s %u", 
			(p1 == 0xFFFF) ? this_modname : MODNAME(p1-1), p2);
	}
	else if (type == 'n')
	{
		char *funcname = FUNCNAME(p2);
		if (funcname)
			printf("%s /* %s %u */", funcname, this_modname, p2);
		else
			printf("%s_%u", this_modname, p2);
	}
	
	putchar('\t');
	print_reloc_2(ofs, r, code);
	putchar('\n');
}

void print_reloc_2(WORD ofs, RELOC_ITEM *r, BYTE *code)
{
	char *typ;
	
	// display relocation type -- crucial for WINP!
	switch (r->nr_stype)
	{
		case NRSPTR:	typ = "PTR";	break;
		case NRSBYT:	typ = "BYT";	break;
		case NRSSEG:	typ = "SEG";			// BASE
			// actually apply SEG reloc, for use in do_gptab!
			*((WORD *) &code[ofs]) = 
				r->nr_union.nr_intref.nr_segno;
			break;
		case NRSOFF:	typ = "OFF";	break;
		case NRSPTR48:	typ = "P48";	break;
		case NRSOFF32:	typ = "O32";	break;
		default:		typ = "???";	break;
	}
	printf("%s", typ);
	
	if (r->nr_flags & NRADD)
	{
		void *cod = &code[ofs];
		switch (r->nr_stype)
		{
			case NRSPTR:	
			case NRSOFF32:
				printf(" ADD+%08lx", (*(DWORD *) cod)); 
				break;
			case NRSBYT:	
				printf(" ADD+%02x", (*(BYTE *) cod)); 
				break;
			case NRSSEG:
			case NRSOFF:
				if (*((WORD *) cod) != 0)	// don't print if ADD+0000
					printf(" ADD+%04x", (*(WORD *) cod)); // signed????
				break;
			case NRSPTR48:
				printf(" ADD+%04x:", (*(WORD *) cod));
				printf("%08lx", (*(DWORD *) cod+2)); 
				break;
			default:
				printf(" ADD????"); 
				break;
		}
	}
}

/* entry table hash table */

typedef ENTRY_SEGOFS NODE;

/* linked list package */

NODE *mk_list(void)
{
	return (NODE *) 0;
}
	
void insert(NODE **list, unsigned ord, unsigned seg, unsigned ofs)
{
	NODE *node = (NODE *) allocate(1, sizeof(NODE));
	node->next = *list;
	*list = node;
	node->ord = ord;
	node->seg = seg;
	node->ofs = ofs;
}

NODE *find(NODE *list, unsigned ord)
{
	while (list)
	{
		if (list->ord == ord)
			return list;
		else
			list = list->next;
	}
	return 0;	// didn't find
}

void print_list(NODE *list)
{
	while (list)
	{
		printf("   [%u %04x:%04x %s]\n",
			list->ord, list->seg, list->ord, list->name);
		list = list->next;
	}
}

/* hash table package */

#define TAB_SIZE		1024		// prime not needed for this 

#define HASH(x)			((x) % TAB_SIZE)

NODE **mk_hashtab(void)
{
	return (NODE **) allocate(TAB_SIZE, sizeof(NODE *));
}

void set(NODE **hashtab, unsigned ord, unsigned seg, unsigned ofs)
{
	insert(&hashtab[HASH(ord)], ord, seg, ofs);
}

NODE *get(NODE **hashtab, unsigned ord)
{
	return find(hashtab[HASH(ord)], ord);
}

void set_name(NODE **hashtab, unsigned ord, char *name)
{
	NODE *node = get(hashtab, ord);
	node->name = name;
}

void print_hashtab(NODE **hashtab)
{
#if 0	
	int i;
	for (i=0; i<TAB_SIZE; i++)
		if (hashtab[i])
		{
			printf("----- %u -----\n", i);
			print_list(hashtab[i]);
		}
#endif		
}

/* entry table hash table */

static NODE **entrytab = (NODE **) 0;

void init_entry(void)	{ entrytab = mk_hashtab(); }

ENTRY_SEGOFS *get_entry(WORD ord)	{ return get(entrytab, ord); }

void set_entry_segofs(WORD ord, WORD seg, WORD ofs)
{
	set(entrytab, ord, seg, ofs);
}

void set_entry_name(WORD ord, char *name)
{
	set_name(entrytab, ord, name);
}

char *get_funcname(WORD ord)
{
	NODE *entry = get_entry(ord);
	return (entry) ? entry->name : NULL;
}

void print_entrytab(void)	{ print_hashtab(entrytab); }

/* land of dead code */

#if 0
#define WIDTH 16

void dump(BYTE *p, WORD bytes, char *mask, DWORD addr)
{
	BYTE *p2;
	WORD i, j, c;
	
    for (i=0; i<bytes; i += WIDTH)
    {
		c = ((bytes-i) > WIDTH) ? WIDTH : bytes-i;
		printf(mask, addr+i);
		printf(" | ");
		for (j=c, p2=p+i; j--; p2++)
			printf("%02X ", *p2);
		for (j=WIDTH-c; j--; )	// pad out on last line
			printf("   ");
		printf("| ");
        for (j=c, p2=p+i; j--; p2++)
			putchar( isprint(*p2) ? *p2 : '.' );
		putchar('\n');
    }
}
#endif

void do_selfload(void)
{
	// handle self-loading Windows apps
	if (hdr.ne_flags & APPLOAD)
	{
		BYTE tbl[0x28];
		SEG *seg;
		int i;

		puts("\nSelfload");

		if (first_code)
		{
			seg = &segtab[first_code];
			exe_read(f, tbl, align * seg->ns_sector, 0x28);
			printf("\tVersion %04x\n", *((WORD *) &tbl[0]));
			printf("\tBoot %08lx\n", *((DWORD *) &tbl[4]));
			printf("\tReload %08lx\n", *((DWORD *) &tbl[8]));
			printf("\tAlloc %08lx\n", 
				*((DWORD *) &tbl[0x10]));
			printf("\tEntry %08lx\n", 
				*((DWORD *) &tbl[0x14]));
			printf("\tExit %08lx\n", *((DWORD *) &tbl[0x18]));
			printf("\tSetOwner %08lx\n", 
				*((DWORD *) &tbl[0x24]));
		}
		else
			puts("\tCouldn't locate first CODE segment!");
		putchar('\n');
	}
}


