/*
FAKEEMM.C
Global EMM Import spec

With thanks to John Spinks and taQ
*/

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef void far *FP;

// in FAKEWIN.C
extern void call_v86_switch(int flag);
#define V86_TO_REAL		0
#define REAL_TO_V86		1

extern void fail(const char *s, ...);

// in DUMP.C
extern void dump(unsigned char far *fp, unsigned bytes, 
    char *mask, unsigned long addr, int width);

void block_move(DWORD source, FP destin, WORD size);

#pragma pack(1)
typedef struct {
	DWORD addr;
	BYTE maj, min;
	} EMM_IMPORT;	// see DDJ, July 1994, p. 128
		
typedef struct {
	BYTE type;
	BYTE handle;
	WORD logpage;
	BYTE physpage;
	BYTE flag;
	} FRAME_STRUCT; 
		
typedef struct {
	BYTE flags;
	BYTE reserved1;
	WORD size;
	WORD version;
	DWORD reserved2;
	FRAME_STRUCT frame[64];
	BYTE reserved3;
	} EMM_HDR;
#pragma pack()

void check_emm_import(void)
{
	static EMM_IMPORT emm_import;
	FRAME_STRUCT far *frame;
	EMM_HDR far *emm_hdr;
	WORD size;
	int i;
	
	int emm = open("EMMXXXX0", O_RDWR | O_BINARY);
	if (emm == -1) emm = open("$MMXXXX0", O_RDWR | O_BINARY);
	if (emm == -1) emm = open("EMMQXXX0", O_RDWR | O_BINARY);
	if (emm == -1) return;	// no EMM
		
	emm_import.addr = 1;	// set first byte to 1
	emm_import.maj = emm_import.min = 0;
	#define IOCTLREAD 2
	if (ioctl(emm, IOCTLREAD, &emm_import, sizeof(emm_import)) != 0)
	{
		printf("\nGlobal EMM Import @ phys %08lXh (version %d.%02d)\n", 
			emm_import.addr, emm_import.maj, emm_import.min);

#if 0		
		// turn V86 mode OFF and ON again to force the IMPORT data to be built
		// according to taQ, 386MAX requires doing the move in here!!

#if 1
		#define IMPORTSIZE 0x1000
		if (! (emm_hdr = (EMM_HDR far *) _fmalloc(IMPORTSIZE)))
			goto done;
		
		call_v86_switch(V86_TO_REAL);
		block_move(emm_import.addr, emm_hdr, IMPORTSIZE);
		call_v86_switch(REAL_TO_V86);
#else
		disable();           // disable interrupts while switching modes.
		call_v86_switch(V86_TO_REAL);
		call_v86_switch(REAL_TO_V86);
		enable();
		
		#define IMPORTSIZE 0x1000
		if (! (emm_hdr = (EMM_HDR far *) _fmalloc(IMPORTSIZE)))
			fail("Insufficient memory");
		
		block_move(emm_import.addr, emm_hdr, IMPORTSIZE);
#endif		
		
		// dump((BYTE far *) emm_hdr, emm_hdr->size, "%08lX", 
		//		emm_import.addr, 16);
		
		for (i=0, frame=emm_hdr->frame; i<64; i++, frame++)
		{
			printf("%04Xh: ", i * 0x400);
			if (! (frame->type & 1)) 
			{
				if (frame->type & 0x3c) 
				{
					int j;
					printf("UMB");
					for (j = 0; j < 4; j++) 
					{
						if (frame->type & (1 << (j + 2))) 
							printf("/%04x",	(i * 4 + j) * 0x100);
					}
					printf(" (umbinfo %d)", frame->handle);
				}
			} 
			else 
			{
				printf("%sEMS (phys page %02d)",
						(frame->type & 2) ? "" : "large ",
						frame->physpage);
				if (frame->handle && frame->handle != 0xff)
					printf(" mapped to handle %d page %d",
						frame->handle, frame->logpage);
			}
			printf("\n");
		}
		
		_ffree(emm_hdr);
#endif		
	}
done:		
	close(emm);
}

void block_move(DWORD source, FP destin, WORD size)
{
    struct {
     WORD limit0, base0;
     BYTE base16, access;
     WORD reserved;
    } gdt[] = {
     { 0, 0, 0, 0, 0 },
     { 0, 0, 0, 0, 0 },
     { 0, 0, 0, 0x93, 0 },
     { 0, 0, 0, 0x93, 0 },
     { 0, 0, 0, 0, 0 },
     { 0, 0, 0, 0, 0 }
    };

	#define  SELSOURCE  0x02
	#define  SELDESTIN  0x03

    union REGS registers;
    struct SREGS segregs;
    DWORD destin24;

    // calculate the physical address of the destination.
    destin24 = ((DWORD) FP_SEG(destin) << 4) + (DWORD) FP_OFF(destin);

    // setup the selector to the source address (24bit physical address).
    gdt[SELSOURCE].limit0 = size;
    gdt[SELSOURCE].base0  = (WORD) source;
    gdt[SELSOURCE].base16 = (BYTE) (source >> 16);

    // setup the selector to the destination address (segment:offset).
    gdt[SELDESTIN].limit0 = size;
    gdt[SELDESTIN].base0  = (WORD) destin24;
    gdt[SELDESTIN].base16 = (BYTE) (destin24 >> 16);

	// call INT 15h function 87h
    registers.h.ah = 0x87;
    registers.x.cx = (size >> 1);
    registers.x.si = FP_OFF((void far *) &gdt);
    segregs.es     = FP_SEG((void far *) &gdt);
    int86x(0x15, &registers, &registers, &segregs);
}

