/*
HOOKINT.C -- See which BIOS INTs called
from Schulman, Unauthorized Windows, 1994

hookint -win win
While Windows running:							Hooked by:
INT 10h     8               Video				VMPoll, VDDVGA
INT 11h     3               Equipment List
INT 14h						Communications		VMPoll, VCD
INT 13h     861             Disk				VFD, IOS/BLOCKDEV/VHD
INT 15h     200             System Services		VKD, V86MMGR, IOS
INT 16h     82              Keyboard			VMPoll, VDDVGA, VKD
INT 17h						Printer				IFSMgr, VMPoll, VPD
INT 1Ah						Time				VTD
INT 25h						DOS Read			VFAT
INT	26h						DOS Write			VFAT
INT 28h     6380            Idle
INT 33h     4               Mouse				VMD

10/00: 2    10/0F: 1    10/10: 3    10/11: 1    10/12: 1    
13/02: 742  13/03: 119  
15/53: 199  15/C0: 1
16/01: 80   16/11: 2    
33/00: 4    

< While Windows running (32BDA)
! While Windows running /D:C (where significantly different from 32DFA)
@ Ran DOS box instead of WinWord (where different from WinWord)
> includes before Windows init, after Windows exit

< INT 10h       8               Video
> INT 10h       124             Video

< INT 11h       3               Equipment List
> INT 11h       6               Equipment List

> INT 12h       1               Memory Size

< INT 13h       861             Disk
! INT 13h       2150            Disk
> INT 13h       1625            Disk

< INT 15h       200             System Services
> INT 15h       10463           System Services

< INT 16h       82              Keyboard
@ INT 16h       2314            Keyboard
> INT 16h       110             Keyboard

> INT 1Ah       10              Time

> INT 25h       4               DOS Disk Read

< INT 28h       6380            Idle
> INT 28h       6303            Idle

< INT 33h       4               Mouse
> INT 33h       5               Mouse

< 10/00: 2  10/0F: 1                            10/10: 3    
> 10/00: 7  10/01: 1    10/0E: 100  10/0F: 4    10/10: 4    

< 10/11: 1  10/12: 1    
> 10/11: 3  10/12: 3    10/6F: 2    

< 13/02: 742    13/03: 119
! 13/02: 1929   13/03: 221  
> 13/02: 1475   13/03: 147  13/08: 3    

< 15/53: 199                                        15/C0: 1
> 15/53: 154    15/87: 2052 15/90: 4118 15/91: 4135 15/C0: 3    15/C1: 1    

< 16/01: 80 16/11: 2    
@ 16/01: 12 16/10: 9    16/11: 2293 
> 16/01: 29 16/11: 81   

> 1A/00: 9  1A/02: 1

< 33/00: 4
> 33/00: 5
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <process.h>
#include <dos.h>
#include <bios.h>

typedef unsigned short WORD;
typedef unsigned long DWORD;

#pragma pack(1)

typedef struct {
#ifdef __TURBOC__
    WORD bp,di,si,ds,es,dx,cx,bx,ax;
#else
    WORD es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    WORD ip,cs,flags;
    } REG_PARAMS;
    
typedef void (interrupt far *FUNCPTR)(REG_PARAMS);

typedef struct {
    FUNCPTR old;
    FUNCPTR new;
    DWORD count;
    DWORD *func;
    int intno;
    int has_func;
    char *str;
    } INT_DATA;
    
extern INT_DATA int_data[];
    
int check_windows = 0;
int in_windows = 0;
int int_ndx[0x100] = {0} ;
    
#define INT_HANDLER(rax, intno) { \
    INT_DATA *pid = &int_data[int_ndx[intno]]; \
    if (in_windows || (! check_windows)) { \
        if (pid->has_func) \
            pid->func[rax >> 8]++; \
        pid->count++; \
        } \
    _chain_intr(pid->old); \
    }

void interrupt far int10(REG_PARAMS r) { INT_HANDLER(r.ax,0x10); }
void interrupt far int11(REG_PARAMS r) { INT_HANDLER(r.ax,0x11); }
void interrupt far int12(REG_PARAMS r) { INT_HANDLER(r.ax,0x12); }
void interrupt far int13(REG_PARAMS r) { INT_HANDLER(r.ax,0x13); }
void interrupt far int14(REG_PARAMS r) { INT_HANDLER(r.ax,0x14); }
void interrupt far int15(REG_PARAMS r) { INT_HANDLER(r.ax,0x15); }
void interrupt far int16(REG_PARAMS r) { INT_HANDLER(r.ax,0x16); }
void interrupt far int17(REG_PARAMS r) { INT_HANDLER(r.ax,0x17); }
void interrupt far int1A(REG_PARAMS r) { INT_HANDLER(r.ax,0x1A); }
void interrupt far int25(REG_PARAMS r) { INT_HANDLER(r.ax,0x25); }
void interrupt far int26(REG_PARAMS r) { INT_HANDLER(r.ax,0x26); }
void interrupt far int28(REG_PARAMS r) { INT_HANDLER(r.ax,0x28); }
void interrupt far int33(REG_PARAMS r) { INT_HANDLER(r.ax,0x33); }
    
INT_DATA int_data[] = {
//  old,    new,    count, func,    intno,  has_func,   str
    0,      int10,  0, 0,           0x10,   1,          "Video",
    0,      int11,  0, 0,           0x11,   0,          "Equipment List",
    0,      int12,  0, 0,           0x12,   0,          "Memory Size",
    0,      int13,  0, 0,           0x13,   1,          "Disk",
    0,      int14,  0, 0,           0x14,   1,          "Serial",
    0,      int15,  0, 0,           0x15,   1,          "System Services",
    0,      int16,  0, 0,           0x16,   1,          "Keyboard",
    0,      int17,  0, 0,           0x17,   1,          "Printer",
    0,      int1A,  0, 0,           0x1A,   1,          "Time",
    0,      int25,  0, 0,           0x25,   0,          "DOS Disk Read",
    0,      int26,  0, 0,           0x26,   0,          "DOS Disk Write",
    0,      int28,  0, 0,           0x28,   0,          "Idle",
    0,      int33,  0, 0,           0x33,   1,          "Mouse",
    } ;

#define NUM_INT     (sizeof(int_data) / sizeof(INT_DATA))

FUNCPTR GetSetInt(int intno, FUNCPTR new)
{
    FUNCPTR old = _dos_getvect(intno);
    _dos_setvect(intno, new);
    return old;
}

FUNCPTR prev_2f = (FUNCPTR) 0;

#define WIN_INIT_COMPLETE   0x1608
#define WIN_EXIT_BEGIN      0x1609

void interrupt far int2f(REG_PARAMS r)
{
    if (r.ax == WIN_INIT_COMPLETE)
        in_windows++;
    else if (r.ax == WIN_EXIT_BEGIN)
        in_windows--;
    _chain_intr(prev_2f);
}

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    int i, j;
    INT_DATA *pint;
    FILE *f;
    
    if (strncmp(strupr(argv[1]), "-WIN", 3) == 0)
    {
        check_windows++;
        argv++; argc--;
    }
    
    /* set up func counters */
    for (i=0, pint=int_data; i<NUM_INT; i++, pint++)
        if (pint->has_func)
            if ((pint->func = (DWORD *) calloc(0x100, sizeof(DWORD))) == 0)
                fail("Insufficient memory");
    
    /* initialize int_ndx */
    for (i=0, pint=int_data; i<NUM_INT; i++, pint++)
        int_ndx[pint->intno] = i;
    
    /* install handlers */
    for (i=0, pint=int_data; i<NUM_INT; i++, pint++)
        pint->old = GetSetInt(pint->intno, pint->new);
    if (check_windows)
        prev_2f = GetSetInt(0x2f, int2f);

    /* run the command */
    if (argc < 2)
        system(getenv("COMSPEC"));
    else
        spawnvp(P_WAIT, argv[1], &argv[1]);
    
    /* restore previous handlers */
    if (check_windows)
        (void) GetSetInt(0x2f, prev_2f);
    for (i=0, pint=int_data; i<NUM_INT; i++, pint++)
        (void) GetSetInt(pint->intno, pint->old);
    
    /* display results */
    if (f = fopen("HOOKINT.LOG", "w"))
        printf("Creating HOOKINT.LOG\n");
    else
        f = stdout;
    if (check_windows)
        fprintf(f, "While Windows running:\n");
    for (i=0, pint=int_data; i<NUM_INT; i++, pint++)
        if (pint->count)
            fprintf(f, "INT %02Xh\t\t%-8lu\t\t%s\n", 
                pint->intno, pint->count, pint->str);
    fprintf(f,"\n");
    for (i=0, pint=int_data; i<NUM_INT; i++, pint++)
        if (pint->has_func && pint->count)
        {
            for (j=0; j<0x100; j++)
                if (pint->func[j])
                    fprintf(f,"%02X/%02X: %lu\t",
                        pint->intno, j, pint->func[j]);
            fprintf(f,"\n");
        }
    fclose(f);
            
    return 0;
}

