/*
INTVECT.C
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

bcc intvect.c map.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

#define MK_LIN(fp)  ((((DWORD) FP_SEG(fp)) << 4) + FP_OFF(fp))

extern char *find_owner(DWORD lin_addr);    // in map.c

#define ARPL    0x63
#define IRET    0xCF
#define JMPF    0xEA
#define JMP8    0xEB
#define JMP16   0xE9

BYTE far *get_vect(int intno)   // call INT 21h AH=35h
{
    _asm push es
    _asm mov al, byte ptr intno
    _asm mov ah, 35h
    _asm int 21h
    _asm mov dx, es
    _asm mov ax, bx
    _asm pop es
    // return value in DX:AX
}

void print_vect(int intno)
{
    char *s;
    BYTE far *fp = get_vect(intno);
    printf("INT %02Xh   %Fp   ", intno, fp);
    if (fp == 0)
    {
        printf("unused\n");
        return;
    }
    s = find_owner(MK_LIN(fp));
    printf("%-08s   ", s? s: " ");

    switch (*fp)    // see if first instruction of interrupt handler
    {               // is anything really obvious
        case ARPL:  printf("arpl -- Windows V86 breakpoint"); break;
        case IRET:  printf("iret -- NOP function"); break;
        case JMP8:  printf("jmp %Fp", 
            ((BYTE far *) fp) + fp[1] + 2); break;
        case JMP16: printf("jmp %Fp",
            ((BYTE far *) fp) + *((WORD far *) &fp[1]) + 3); break;
        case JMPF:  printf("jmp %Fp", 
            *((void far * far *) &fp[1])); break;
    }
    printf("\n");
}

main(int argc, char *argv[])
{
    char *end;
    int intno, i;

    if (argc < 2)
        for (intno=0; intno<256; intno++)
            print_vect(intno);
    else for (i=1; i<argc; i++)
        print_vect((int) strtoul(argv[i], &end, 16));

    return 0;
}
