/*
ISWIN.C
Detecting Windows mode, version from DOS
Andrew Schulman, February 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

bcc -DSTANDALONE iswin.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

#define REAL_MODE           1
#define STANDARD_MODE       2
#define ENHANCED_MODE       3

#define SYSTEM_VM           1

int detect_switcher(void)
{
    int retval = 1;
    _asm push di
    _asm push es
    _asm xor bx, bx
    _asm mov di, bx
    _asm mov es, bx
    _asm mov ax, 4b02h
    _asm int 2fh
    _asm mov cx, es
    _asm or cx, di
    _asm je no_switcher
done:       
    _asm pop es
    _asm pop di
    return retval;
no_switcher:
    retval = 0;
    goto done;
}

int is_win(int *pmaj, int *pmin, int *pmode)
{
    unsigned short retval;
    int maj=0, min=0, mode=0;
    
    /* make sure someone, anyone has INT 2Fh */
    if (_dos_getvect(0x2F) == 0)
        return 0;
    
    /* call 2F/160A to see if Windows 3.1+ */
    _asm mov ax, 160ah
    _asm int 2fh
    _asm mov retval, ax
    if (retval == 0)                /* AX=0 if Windows running */
    {
        _asm mov mode, cx           /* CX=2 means Std; CX=3 means Enh */
        _asm mov byte ptr maj, bh   /* BX = major/minor (e.g., 030Ah) */
        _asm mov byte ptr min, bl
        *pmaj = maj;
        *pmin = min;
        *pmode = mode;
        return 1;
    }

    /* call 2F/1600 to see if Windows 3.0 Enhanced mode or Windows/386 */
    _asm mov ax, 1600h
    _asm int 2fh
    _asm mov byte ptr maj, al
    _asm mov byte ptr min, ah
    if ((maj == 1) || (maj == 0xFF))    /* Windows/386 2.x is running */
    {
        *pmaj = 2;              /* Windows/386 2.x */
        *pmin = 1;              /* don't know; assume 2.1? */
        *pmode = ENHANCED_MODE; /* Windows/386 sort of like Enhanced */
        return 1;
    }
    else if (! ((maj == 0) || (maj == 0x80))) /* AL=0 or 80h if no Windows */
    {   /* must be Windows 3.0 Enhanced mode */
        *pmaj = maj;
        *pmin = min;
        *pmode = ENHANCED_MODE;
        return 1;
    }
    
    /* call 2F/4680 to see if Windows 3.0 Standard or Real mode; but, 
       this could be a "3.0 derivative" such as DOSSHELL task switcher! */
    _asm mov ax, 4680h
    _asm int 2fh
    _asm mov retval, ax
    if (retval == 0)            /* AX=0 if 2F/4680 handled */
    {
        /* make sure it isn't DOSSHELL task switcher */
        if (detect_switcher())
            return 0;
        *pmaj = 3;
        *pmin = 0;
        
        /* either have Windows Standard mode or Real mode; to 
           distinguish, have to do fake Windows broadcasts with
           2F/1605.  Yuk!  We'll avoid that here by assuming
           3.0 Standard mode.  If you really want to distinguish
           3.0 Standard mode and Real mode, see _MSJ_, March 1991,
           p. 113; and MS KB articles Q75943 and Q75338 */
        *pmode = STANDARD_MODE;
        return 1;
    }
        
    /* still here -- must not be running Windows */
    return 0;
}

#ifdef STANDALONE
main()
{
    int maj, min, mode=0;
    if (! is_win(&maj, &min, &mode))
        printf("Windows is not running\n");
    else if (maj == 2)
        printf("Running Windows/386 2.x\n");
    else 
        printf("Running Windows %u.%02u (or higher) %s mode\n",
            maj,
            min,
            (mode == REAL_MODE) ?     "Real" :
            (mode == STANDARD_MODE) ? "Standard" :
            (mode == ENHANCED_MODE) ? "Enhanced" :
            /* don't know */ "???");

    if (mode == ENHANCED_MODE)
    {
        unsigned short vm;
        /* call 2F/1683 to see if DOS app running in System VM; if so, 
           this must be some hacked version of Windows like MSDPMI,
           or we must be running inside WINSTART.BAT */
        _asm mov ax, 1683h
        _asm int 2fh
        _asm mov vm, bx
        if (vm == SYSTEM_VM)
            printf("Running DOS app in System VM: "
                   "Must be WINSTART.BAT or hacked Windows!\n");
        else
            printf("VM #%u\n", vm);
    }
    
    /* could also call 2F/160C to check for Windows in ROM */
            
    return mode;    /* can be tested with ERRORLEVEL */
}
#endif
