/* 
le2.c -- used by ledump.c and w3map.c

changes:
	8/12/92 -- for Service_Table, changed DD, R ??  to DD, OSN, R ??
	10/10/92 -- only do service table if numbers "reasonable"
	1/16/93 -- print out Init Order too
	4/20/93 -- put macros back in; how get lost?
	5/3/94 -- added ddb_ to names of DDB fields (e.g., ddb_control_proc)
	5/6/94 -- first attempt to handle multiple objects
	5/9/94 -- finally, handling VxD start (real-mode init!)
*/		

#define CHICAGO

#include <stdlib.h>
#include <stdio.h>
#include "filelib.h"

extern int f;

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
#include "exe386.h"

#include "vxdfunc.h"

#pragma pack(1)

typedef struct {
	BYTE cnt, bundle;
	WORD obj;		/// <------ !!!!
	// above is bundle; below is one entry
	BYTE flags;
	DWORD ent_offset;
	WORD ord;
	} ENTRY;	// hack for only one entry!
				// maybe this is the problem with WfW 3.11 VxDs!
					
typedef struct {
	DWORD DDB_Next;
	WORD DDB_SDK_Version;
	WORD DDB_Req_Device_Number;
	BYTE DDB_Dev_Major_Version;
	BYTE DDB_Dev_Minor_Version;
	WORD DDB_Flags;
	BYTE DDB_Name[8];
	DWORD DDB_Init_Order;
	DWORD DDB_Control_Proc;
	DWORD DDB_V86_API_Proc;
	DWORD DDB_PM_API_Proc;
	DWORD DDB_V86_API_CSIP;
	DWORD DDB_PM_API_CSIP;
	DWORD DDB_Reference_Data;
	DWORD DDB_Service_Table_Ptr;
	DWORD DDB_Service_Table_Size;
#ifdef CHICAGO	
	DWORD DDB_Win32_Service_Table_Ptr;
#endif
	} DDB;		// from ddk include/vmm.inc

static struct o32_obj *objtab = (struct o32_obj *) 0;
static struct o32_map *objmap = (struct o32_map *) 0;
static DWORD objcnt = 0L;

void init_objmap(DWORD le_offset, struct e32_exe *phdr, int display)
{
	struct o32_obj *o;
	int i;

	if (display)
	{
		printf("Object table @ %08lx (%lx objects)\n", 
			phdr->e32_objtab, phdr->e32_objcnt);
		printf("Object page map @ %08lx\n",	phdr->e32_objmap);
	}

	if (! (objcnt = phdr->e32_objcnt))
		fail("No objects");
	objtab = (struct o32_obj *) allocate(objcnt, sizeof(struct o32_obj));
	objmap = (struct o32_map *) allocate(objcnt, sizeof(struct o32_map));
	exe_read(f, objtab, le_offset + phdr->e32_objtab, 
		objcnt * sizeof(struct o32_obj));
	exe_read(f, objmap, le_offset + phdr->e32_objmap,
		objcnt * sizeof(struct o32_map));
	
	if (display)
	{
		for (i=0, o=objtab; i<objcnt; i++, o++)
		{
			printf("%d\t%08lx\t%08lx\t%08lx\t",
				i, o->o32_size, o->o32_base, o->o32_pagemap);
			printf("%08lx\n", GETPAGEIDX(objmap[i]));
		}
		printf("\n");
	}
}

#define GET_FOFS(x)		get_file_offset(x)
#define GET_FOFS_SR(x)	get_obj_offset(x)
#define GET_OBJ(x)	('a' + get_obj(x))

WORD get_obj(DWORD virt_addr)
{
	struct o32_obj *o;
	int i;
	
	if (! objcnt) 
		fail("Internal error: no objects");
	
	for (i=0, o=objtab; i<objcnt; i++, o++)
	{
		if ((virt_addr >= o->o32_base) &&
			(virt_addr < o->o32_base + o->o32_size))
			return i;
	}
	// still here
	return 0;
}

DWORD get_file_offset(DWORD virt_addr)
{
	struct o32_obj *o;
	int i;
	
	if (! objcnt) 
		fail("Internal error: no objects");
	
	for (i=0, o=objtab; i<objcnt; i++, o++)
	{
		if ((virt_addr >= o->o32_base) &&
			(virt_addr < o->o32_base + o->o32_size))
			return ((virt_addr - o->o32_base) +
				    ((GETPAGEIDX(objmap[i]) - 1) * 0x1000L));
	}
	/* still here */
//	printf("Warning: Couldn't map virtual address %08lXh\n", virt_addr);
	return (virt_addr + 0x1000L);
}

DWORD get_obj_offset(DWORD virt_addr)
{
	struct o32_obj *o;
	int i;
	
	if (! objcnt) 
		fail("Internal error: no objects");
	
	for (i=0, o=objtab; i<objcnt; i++, o++)
	{
		if ((virt_addr >= o->o32_base) &&
			(virt_addr < o->o32_base + o->o32_size))
			return (virt_addr - o->o32_base);
	}
	/* still here */
	return virt_addr;
}

void le2(DWORD le_offset)
{
	struct e32_exe hdr;
	ENTRY entry;
	DDB ddb;
	char *modname;
	char **devtbl;
	unsigned devtbl_size;
	DWORD inito;
#ifdef DO_NRESTAB	
	BYTE *nrestab;
	char *ddb_name;
	char *desc;
#endif	
	BYTE len;
	
	exe_read(f, &hdr, le_offset, sizeof(hdr));
	if (! (hdr.e32_magic[0] == 'L') && (hdr.e32_magic[1] == 'E'))
		fail("not an 'LE' (linear executable)");
	
	if (hdr.e32_os != 4)	// Windows
		fail("not a Windows VxD");
	
	init_objmap(le_offset, &hdr, 1);

	/* get module name from resident names table */
	exe_read(f, &len, le_offset + hdr.e32_restab, 1);
	modname = (char *) allocate(len+1, 1);
	exe_read(f, modname, le_offset + hdr.e32_restab+1, len);
	modname[len] = '\0';
	printf("Module name: %s\n", modname);

#ifdef DO_NRESTAB	
	/* get nonresident names table */
	nrestab = (BYTE *) allocate(hdr.e32_cbnrestab, 1);
	exe_read(f, nrestab, hdr.e32_nrestab, hdr.e32_cbnrestab);
	
	/* get description string */
	len = nrestab[0];
	desc = (char *) allocate(1, len+1);
	strncpy(desc, &nrestab[1], len);
	printf("Description: \"%s\"\n", desc);
	
	/* print DDB name */
	ddb_name = (char *) allocate(1, nrestab[len+3]+1);
	strncpy(ddb_name, &nrestab[len+4], nrestab[len+3]);
#endif	
	
	/* get DDB offset from entry table */
	exe_read(f, &entry, le_offset + hdr.e32_enttab, sizeof(entry));
	entry.ent_offset = GET_FOFS(entry.ent_offset);
#ifdef DO_NRESTAB
	printf("%s @ %04x:%08lx\n", ddb_name, entry.obj, entry.ent_offset);
#else
	printf("%s_DDB @ %04X:%08lx\n", modname, entry.obj, entry.ent_offset);
#endif

	printf("Real-mode Init @ %04lX:%08lX\n", hdr.e32_startobj, hdr.e32_eip);
	
	// hdr.e32_datapage is offset for load image
		
#if 0		
	printf("datapage: %08lx\n", hdr.e32_datapage);
	printf("entoffs:  %08lx\n", entry.ent_offset);
#endif	
	
	/* get Device Descriptor Block (DDB) */
	///// have to look at entry.obj !!!!
	if (hdr.e32_datapage == 0x1000)
		exe_read(f, &ddb, entry.ent_offset, sizeof(ddb));
	else
		exe_read(f, &ddb, hdr.e32_datapage + entry.ent_offset, sizeof(ddb));

	/* get the device ID# and see if there's a function name tbl */
	devtbl = (char **) 0;
	devtbl_size = 0;
	if (ddb.DDB_Req_Device_Number)
	{
		WORD devnum = ddb.DDB_Req_Device_Number;
		printf("Device # %04x\n", devnum);
		if (devnum < NUM_DEV)
		{
			// NB!!  For dev 0x10 (BLOCKDEV or VHD), have to
			// decide between 3.0 and 3.10!!
			if ((devnum == 0x10) &&
				(ddb.DDB_Dev_Major_Version == 3) &&
				(ddb.DDB_Dev_Minor_Version == 0))
			{
				devtbl = VHD_Tbl;
				devtbl_size = TBL_SIZE(VHD_Tbl);
				printf("Virtual Hard Disk Device\n");
			}
			else
			{
				devtbl = (char **) VxD_Tbl[devnum].tbl;   // may be zero
				devtbl_size = VxD_Tbl[devnum].size;	// ditto
				if (VxD_Tbl[devnum].desc)
					printf("%s\n", VxD_Tbl[devnum].desc);
			}
		}
		else
		{
			int i;
			for (i=0; i<NUM2_DEV; i++)
				if (devnum == VxD_Tbl2[i].devnum)
				{
					devtbl = (char **) VxD_Tbl2[i].tbl;   // may be zero
					devtbl_size = VxD_Tbl2[i].size;	// ditto
					if (VxD_Tbl2[i].desc)
						printf("%s\n", VxD_Tbl2[i].desc);
					break;
				}
		}
	}
	
	printf("Version %u.%02u\n", 
		ddb.DDB_Dev_Major_Version, ddb.DDB_Dev_Minor_Version);
	
	inito = ddb.DDB_Init_Order;
	printf("Init order: %08lX ", inito);
	if (inito == 0x80000000L)
		printf("(Undefined)\n");
	else if (inito == 0)
		printf("(Earliest -- same as VMM)\n");
	else if (inito < 0x1000000L)
		printf("(Early -- before WIN386 built-in VxDs)\n");
	else if (inito > 0xB0000000L)
		printf("(Late -- after all WIN386 built-in VxDs)\n");
	else if (inito == 0xa8000000L)
		printf("(Touch first meg at crit init -- after V86MMGR)\n");
	else
		printf("\n");
	
	// PROBLEM!  Control_Proc could be at 0!
	// There is always a Control_Proc!
	
	printf("DDB_Control_Proc @ %08lx\n", GET_FOFS(ddb.DDB_Control_Proc));

	// If V86 or PM API, can't be located at offset 0????

	if (ddb.DDB_V86_API_Proc)
		printf("DDB_V86_API_Proc @ %08lx\n", GET_FOFS(ddb.DDB_V86_API_Proc));
	if (ddb.DDB_PM_API_Proc)
		printf("DDB_PM_API_Proc  @ %08lx\n", GET_FOFS(ddb.DDB_PM_API_Proc));
	if (ddb.DDB_V86_API_CSIP)
		printf("DDB_V86_API_CSIP @ %08lx\n", ddb.DDB_V86_API_CSIP);
	if (ddb.DDB_PM_API_CSIP)
		printf("DDB_PM_API_CSIP  @ %08lx\n", ddb.DDB_PM_API_CSIP);
	/* be prepared for total garbage in here? */
	if (ddb.DDB_Service_Table_Ptr && 
		(ddb.DDB_Service_Table_Ptr < 0x2000000L))
	{
		DWORD *servtbl;
		DWORD *serv;
		int i;
		DWORD size = ddb.DDB_Service_Table_Size;
		WORD devnum = ddb.DDB_Req_Device_Number;
		ddb.DDB_Service_Table_Ptr = GET_FOFS(ddb.DDB_Service_Table_Ptr);
		printf("DDB_Service_Table @ %08lx (%0lx services)\n", 
			ddb.DDB_Service_Table_Ptr, size);
		if (size > 0x1000)
			printf("Service table size looks wrong!\n");
		else
		{
			servtbl = (DWORD *) allocate(size, sizeof(long));
			exe_read(f, servtbl, hdr.e32_datapage + ddb.DDB_Service_Table_Ptr,
				size * sizeof(long));
			for (i=0, serv=servtbl; i<size; i++, serv++)
			{
				printf("\t%02x%04x @ %08lx", devnum, i, GET_FOFS(*serv));
				if (devtbl && (i < devtbl_size))
				{
					if (*devtbl[i] == '~')
					{
						if (devtbl[i][1] == '~')
							printf("\t%s_Get_Version", modname);
						else
							printf("\t%s_%s", modname, devtbl[i] + 1);
					}
					else
						printf("\t%s", devtbl[i]);
				}
				putchar('\n');
			}
			free(servtbl);
		}
	}

#if 0	
	// Win32 service table stuff
#if 0
VMM       4.00   0001h   8001DBE4   80003F55   80004F56   80004F56   366 ! 34
VCOMM     1.00   002Bh   80069F38   80069CF4   80245D88   80245D88   32 ! 26
VNETBIOS  3.00   0014h   80045640   80044704                         4 ! 2
VWIN32    1.02   002Ah   800677F8   80066B74              80243B70   13 ! 42
VCOND     1.00   0038h   80082ADC   80082AB0   80248DB4   80248E6B   2 ! 46
#endif
	if (((ddb.DDB_Req_Device_Number == 1) && (ddb.DDB_Major_Version >= 4)) ||
		(ddb.DDB_Req_Device_Number == 0x2b) ||
		((ddb.DDB_Req_Device_Number == 0x14)&&(ddb.DDB_Major_Version >= 3)) ||
		(ddb.DDB_Req_Device_Number == 0x2a) ||
		(ddb.DDB_Req_Device_Number == 0x38))
	{
	}
#endif	

#if 0	
	if (ddb.DDB_Next != 0)
		printf("Next = %08lX\n", ddb.DDB_Next);
#endif		

#ifdef DO_NRESTAB		
	free(ddb_name);
	free(desc);
	free(nrestab);
#endif	
	free(modname);
}

