/* 
LEDUMP.C -- linear executable dumper

bcc -I..\include -L..\lib ledump.c unauthwr.lib 

CHANGES:
	1.04 (7/27/92):  
		fixed le2.c so that Control_Proc recognized even when addr==0

	1.05 (8/12/92):
		changed Service_Table from dd, r ?? to dd, osn, r ?? (le2.c)
			
	1.06 (10/19/92):
		changed service table stuff in le2.c for service table size
		and ptrs which are garbage (msdnlib\vmpages\vmpages.386 weird!)
			
	1.07 (11/3/92):
		changed le2.c more to check for garbage ptrs -- check that
		file offset < filelength(f)
			
	1.08 (1/16/93):
		printing out init order in le2.c
			
	2.00 (4/20/93):
		put macros back in le2.c -- somehow got lost?!
			
	2.02 (5/4/94):
		Fixed DDB printout in le2.c
		Display MAYBE_API if 0
		Added new VMM services
		Added new VxDs for WfW 3.11
			
	2.10 (5/6/94):
		in le2.c, first attempt to handle multiple objects
		finally, doing real_mode_init in le2.c
			
	2.20 (8/1/94):
		improved multiple-object support
		.def file has ClientXXX
			
TODO:
	Need SYMDUMP program to do .sym files!!  
*/		


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include "filelib.h"

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC		0x454C			/* linear executable */
#define LXMAGIC		0x584C			/* IBM OS/2 2.0 linear executable */
#define PEMAGIC		0x4550			/* NT portable executable */
#define W3MAGIC		0x3357			/* WIN386.EXE signature */
#define SZMAGIC		0x5A53			/* SZ -- compressed file */

void fail(char *s) { fputs("ledump fail: ", stdout); puts(s); exit(1); }

void fail_perror(void)	{ perror("ledump fail"); exit(1); }

void usage()
{
	puts("LEDUMP displays information about a Windows 3.x Enhanced mode");
	puts("Virtual Device Driver (VxD), in a linear-executable (LE) file.");
	puts("");
	puts("To display information about a VxD:");
	puts("         LEDUMP vxd_file");
	puts("example: ledump \\win31\\system\\vtdapi.386");
	exit(1);
}
	
void le(char *name);
void le2(unsigned long offset);

static unsigned long align;
int f;
unsigned long start_newexe;

main(int argc, char *argv[])
{
	int i;

	fputs("Linear Executable Dumper version 2.20\n", stderr);
	fputs("from \"Unauthorized Windows\" by Andrew Schulman "
		  "(IDG Books, 1994)\n", stderr);
	fputs("Copyright (c) 1992-94 Andrew Schulman.  All rights reserved.\n\n", 
		stderr);
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			if (argv[i][1] == '?')
				usage();
		}
		else
		{
			char buf[128];
			strcpy(buf, argv[i]);
			if (! strchr(buf, '.'))
				strcat(buf, ".386");
			le(buf);
		}
	return 0;
}

void le(char *name)
{
	unsigned long pos;
    unsigned short w;
	unsigned i;
	
	if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
		fail_perror();
	exe_read(f, &w, 0, 2);	// read first two bytes of file
    if (w != EMAGIC)
	{
		if (w == SZMAGIC)
			fail("compressed file (SZ) -- use EXPAND.EXE");
		else
			fail("not an executable");
	}
	exe_read(f, &w, 0x18, 2);
	if (w < 0x40)
		fail("not an LE (linear executable)");
	exe_read(f, &start_newexe, ENEWHDR, 4);
	exe_read(f, &w, start_newexe, 2);
	if (w == W3MAGIC)
		fail("use W3MAP to examine this WIN386 file");
	else if (w != LEMAGIC)
		fail("not an LE (linear executable)");
	le2(start_newexe);
	close(f);
}

/**********************************************************************/

#define DO_NRESTAB
#include "le2.c"

/**********************************************************************/


