/* LFN.C -- test presence of long filename functions */

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

#define GET_VOLUME_INFORMATION			0x71A0
#define OLD_GET_VOLUME_INFORMATION		0x4302

#define FS_CASE_SENSITIVE				1
#define FS_CASE_IS_PRESERVED			2
#define FS_UNICODE_ON_DISK				4
#define FS_LFN_APIS						0x4000
#define FS_VOLUME_COMPRESSED			0x8000

int GetVolumeInformation(char far *RootName, 
	char far *Buffer, unsigned BufSize,
	unsigned *pFlags, unsigned *pMaxFilename, unsigned *pMaxPath)
{
	unsigned rbx, rcx, rdx;
	_asm push ds
	_asm push di
	_asm les di, dword ptr Buffer
	_asm lds dx, dword ptr RootName
	_asm mov cx, BufSize
	_asm mov ax, GET_VOLUME_INFORMATION
	_asm int 21h
	_asm pop di
	_asm pop ds
	_asm jc no_71A0
	_asm mov rbx, bx
	_asm mov rcx, cx
	_asm mov rdx, dx
	*pFlags = rbx;
	*pMaxFilename = rcx;
	*pMaxPath = rdx;
	return 0;
no_71A0:
	_asm xor ah, ah
	// error code in AL
}

main(int argc, char *argv[])
{
	char *RootName = (argc < 2) ? "C:\\" : argv[1];
	char Buffer[128];
	unsigned Flags, MaxFilename, MaxPath;
	int ret;
	if ((ret = GetVolumeInformation(RootName, 
		Buffer, sizeof(Buffer),
		&Flags, &MaxFilename, &MaxPath)) != 0)
	{
		printf("%s -- No long filename (LFN) support\n", RootName);
		printf("error code: %d (%02Xh)\n", ret, ret);
		return 1;
	}
	else
	{
		printf("%s -- Long filename (LFN) support\n", RootName);
		printf("File system name: \"%s\"\n", Buffer);
		printf("MaxFilename: %d\n", MaxFilename);
		printf("MaxPath: %d\n", MaxPath);
		
		#define PRINT_FLAG(fl, s1, s2) \
			printf("%s\n", (Flags & (fl)) ? (s1) : (s2))
				
		PRINT_FLAG(FS_CASE_SENSITIVE,
			"Searches are case sensitive",
			"Searches are NOT case sensitive");
		PRINT_FLAG(FS_CASE_IS_PRESERVED,
			"Preserves case in directory entries",
			"Does NOT preserve case in directory entries");
		PRINT_FLAG(FS_LFN_APIS,
			"Supports DOS long filename functions",
			"Does NOT support DOS long filename functions");
		PRINT_FLAG(FS_VOLUME_COMPRESSED,
			"Volume is compressed",
			"Volume is NOT compressed");
		
		// output from following one looks wrong!!
		PRINT_FLAG(FS_UNICODE_ON_DISK,
			"Uses Unicode characters in file and directory names",
			"Does NOT use Unicode for file/directory names");

		return 0;
	}
}


