/* 
MAP.C
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

bcc intvect.c map.c
bcc intchain.c map.c
bcc -DTESTING map.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef void far *FP;

#ifndef MK_FP
#define MK_FP(s,o)      ((((DWORD) s) << 16) + (o))
#endif

#pragma pack(1)

typedef struct {
    DWORD start, end;
    char name[9];
    } BLOCK;
    
static BLOCK *map;
static int num_block = 0;

int cmp_func(const void *b1, const void *b2)    
{
    if (((BLOCK *) b1)->start < ((BLOCK *) b2)->start)       return -1;
    else if (((BLOCK *) b1)->start > ((BLOCK *) b2)->start)  return 1;
    else                                                     return 0;
}

typedef struct {
    BYTE type;          /* 'M'=in chain; 'Z'=at end */
    WORD owner;         /* PSP of the owner */
    WORD size;          /* in 16-byte paragraphs */
    BYTE unused[3];
    BYTE name[8];       /* in DOS 4+ */
    } MCB;

#define IS_PSP(mcb)     (FP_SEG(mcb) + 1 == (mcb)->owner)

WORD get_first_mcb(void)
{
    _asm mov ah, 52h
    _asm int 21h
    _asm mov ax, es:[bx-2]
    // retval in AX
}

typedef struct DEV {
    struct DEV far *next;
    WORD attr, strategy, intr;
    union {
        BYTE name[8], blk_cnt;
        } u;
    } DEV;

#define IS_CHAR_DEV(dev)    ((dev)->attr & (1 << 15))

DEV far *get_nul_dev(void)
{
    _asm mov ah, 52h
    _asm int 21h
    _asm mov dx, es
    _asm lea ax, [bx+22h]
    // retval in DX:AX
}

int get_num_block_dev(DEV far *dev)
{
	// can't rely on # block devices at SysVars+20?
	int num_blk = 0;
	do {
		if (! IS_CHAR_DEV(dev))
			num_blk += dev->u.blk_cnt;
		dev = dev->next;
	} while(FP_OFF(dev->next) != (WORD) -1);
	return num_blk;
}

WORD get_umb_link(void)
{
    _asm mov ax, 5802h
    _asm int 21h
    _asm xor ah, ah
    // return value in AX
}

WORD set_umb_link(WORD flag)
{
    _asm mov ax, 5803h
    _asm mov bx, flag
    _asm int 21h
    _asm jc error
    _asm xor ax, ax
error:;
    // return 0 or error code in AX
}

WORD get_dos_ds(void)
{
    _asm push ds
    _asm mov ax, 1203h
    _asm int 2fh
    _asm mov ax, ds
    _asm pop ds
    // retval in AX
}

/* find IO.SYS segment with built-in drivers */
WORD get_io_seg(void)
{
    WORD io_seg = 0;
    DEV far *dev = get_nul_dev();
    do {
        if (IS_CHAR_DEV(dev))
            if (_fstrncmp(dev->u.name, "CON     ", 8) == 0)
                io_seg = FP_SEG(dev);   // we'll take the last one
        dev = dev->next;
    } while(FP_OFF(dev->next) != (WORD) -1);
    return io_seg;
}

static int did_init = 0;

void do_init(void)
{
    MCB far *mcb;
    DEV far *dev;
    WORD dos_ds, io_seg, mcb_seg, next_seg, save_link;
    BLOCK *block;
    int blk, i;
    
    map = (BLOCK *) calloc(100, sizeof(BLOCK));
    block = map;
    
    io_seg = get_io_seg();
    block->start = io_seg << 4; block->end = (DWORD) -1;
    strcpy(block->name, "IO");
    block++;
    
    dos_ds = get_dos_ds();
    block->start = dos_ds << 4; block->end = (DWORD) -1;
    strcpy(block->name, "DOS");
    block++;
    
	// should really check if there IS an HMA!
    block->start = 0x100000L;   block->end =  0x10FFEEL;
    strcpy(block->name, "HMA");
    block++;
    num_block = 3;
    
    /* walk MCB chain, looking for PSPs, interrupt owners */
    if (_osmajor >= 4)
    {
        mcb_seg = get_first_mcb();
        mcb = (MCB far *) MK_FP(mcb_seg, 0);
        
        if (_osmajor >= 5)  // be lazy; see ch. 7 for DOS < 5
        {
            save_link = get_umb_link();
            set_umb_link(1);    // access UMBs too
        }

        for (;;)
        {
            next_seg = mcb_seg + mcb->size + 1;
            if (IS_PSP(mcb))
            {
                block->start = ((DWORD) mcb_seg) << 4;
                block->end = ((DWORD) next_seg) << 4;
                _fstrncpy(block->name, mcb->name, 8);
                block->name[8] = '\0';
                block++;
                num_block++;
            }
            mcb_seg = next_seg;

            if (mcb->type == 'M')
                mcb = (MCB far *) MK_FP(next_seg, 0);
            else
                break;
        }
    }
    
    /* walk device chain looking for non-builtin drivers */
    dev = get_nul_dev();
    blk = get_num_block_dev(dev);
    do {
        MCB far *dev_mcb;
        if ((FP_SEG(dev) != dos_ds) && (FP_SEG(dev) != io_seg))
        {
            block->start = (((DWORD) FP_SEG(dev)) << 4) + FP_OFF(dev);
            dev_mcb = (MCB far *) MK_FP(FP_SEG(dev)-1,0);
            if (dev_mcb->owner == 8)
            {
                dev = dev->next;
                continue;
            }
            if (dev_mcb->type == 'M')
                block->end = block->start + ((DWORD) dev_mcb->size << 4);
            else
                block->end = (DWORD) -1;
            if (IS_CHAR_DEV(dev))
            {
                _fstrncpy(block->name, dev->u.name, 8);
                block->name[8] = '\0';
            }
            else
            {
				// to determine drive letter, have to work backwards
				// from number of block devices
                blk -= dev->u.blk_cnt;
                block->name[0] = blk + 'A';
                block->name[1] = ':';
                block->name[2] = '\0';
            }
            block++;
            num_block++;
        }
        dev = dev->next;
    } while(FP_OFF(dev->next) != (WORD) -1);
    
    if (_osmajor >= 5)
        set_umb_link(save_link);
    
    qsort(map, num_block, sizeof(BLOCK), cmp_func);
    
    for (i=0, block=map; i<num_block-1; i++, block++)
        if (block->end == (DWORD) -1)
            block->end = map[i+1].start;
    if (block->end == (DWORD) -1)   // last one
        block->end = 0xFFFFFL;
    
    did_init = 1;
}

char *find_owner(DWORD lin_addr)
{
    BLOCK *block;
    int i;
    
    if (! did_init) 
        do_init();
    
    for (i=0, block=map; i<num_block; i++, block++)
        if ((lin_addr >= block->start) &&
            (lin_addr <= block->end))
            return block->name;
        
    /* still here */
    return (char *) 0;
}

#ifdef TESTING
main()
{
    BLOCK *block;
    int i;
    do_init();
    for (i=0, block=map; i<num_block; i++, block++)
        printf("%08lX   %08lX   %s\n",
            block->start, block->end, block->name);
}
#endif
