/* 
MEMLOOP.C
How much memory can be allocated through DOS (INT 21h AH=48h)?
Can be compiled for real mode or protected mode (DPMI)

real mode: bcc memloop.c
prot mode: bcc -2 -DDPMI_APP memloop.c dpmish.c ctrl_c.asm

On machine with 12 megabytes:
Real mode:                   536 Kb
DPMI app, under 386MAX     10764 Kb (18916 Kb with DPMI swap file)
DPMI app, under DOSX       10894 Kb (18916 Kb with 386MAX DPMI swap file)
DPMI app, under WIN386     38096 Kb
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#ifdef DPMI_APP
#include "dpmish.h"		// includes _dos_allocmem redefinition
#endif

void mem_loop(void)
{
    unsigned long kb = 0;
	unsigned blocks = 0;
    unsigned blocksize = 2048;
    unsigned kbytes = 32;
    unsigned segsel;
    while (kbytes)
    {
        while (_dos_allocmem(blocksize, &segsel) == 0)  // INT 21h AH=48h
        {
			// touch every byte!
			// _fmemset(MK_FP(segsel, 0), 'x', blocksize * 16);
			kb += kbytes;
			blocks++;
			printf("%04Xh\t%d\t%lu\t\t\r", segsel, blocks, kb);
			#ifdef DPMI_APP
				if (ctrl_c_hit)
					fail("\nCtrl-C detected");
			#endif
        }
        blocksize >>= 1;
        kbytes >>= 1;
    }
    printf("\nAllocated %lu Kb in %u blocks\n", kb, blocks);
}

#ifdef DPMI_APP

void fail(const char *s, ...) { puts(s); _dos_exit(1); }

int real_main(int argc, char *argv[])
{
    return 0;   // okay to switch to protected mode
}

int pmode_main(int argc, char *argv[])
#else
int main()
#endif
{
    mem_loop();
    return 0;
}

