/* 
MYDIR.C -- Compare old INT 21h with LFN
bcc mydir.c
bcc -DLFN -elfndir.exe mydir.c

from Schulman, Unauthorized Windows, 1994

Find First calls: 494
Find Next calls:  11311 - 11317

					Clocks			Calls/Clock
					------			-----------
DOS					 659				17
#2					 654				18

SmartDrv			 548				21
#2					 513				23

WfW 3.11 32BFA		 538				21
#2					  82			   144

Chicago				 953			    12
#2					 409				28

Chicago LFNDIR		 699				16
#2					 216				54
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include <time.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef unsigned short HANDLE;
typedef int BOOL;

#ifdef LFN
#define MAX_PATH		260

typedef struct {
    DWORD dwLowDateTime, dwHighDateTime;
} FILETIME;

typedef struct {
    DWORD 	 dwFileAttributes;
    FILETIME ftCreationTime, ftLastAccessTime, ftLastWriteTime;
    DWORD 	 nFileSizeHigh, nFileSizeLow;
    DWORD 	 dwReserved0, dwReserved1;
    BYTE 	 cFileName[MAX_PATH], cAlternateFileName[14];
} WIN32_FIND_DATA;

#define DATETIME_64				  0
#define DATETIME_DOS			  1

HANDLE FindFirstFile(char far *Name, WIN32_FIND_DATA far *Result, 
	WORD Attributes, WORD DateTimeFormat)
{
	HANDLE Handle;
	WORD ConversionCode;
	_asm {
		push si
		push di
		push ds
		mov ax, 714Eh             // Find First File
		mov cx, Attributes        // file attributes
		lds dx, dword ptr Name    // address of name for search
		les di, dword ptr Result  // address of results buffer
		mov si, DateTimeFormat    // format for returned date and time
		int 21h
		pop ds
		pop di
		pop si
		jc  error
		mov [Handle], ax          // search handle
		mov [ConversionCode], cx  // UNICODE to OEM/ANSI conversion ok?
	}
	return Handle;
error:
	return 0;
}

BOOL FindNextFile(HANDLE Handle, WIN32_FIND_DATA far *Result, 
	WORD DateTimeFormat)
{
	WORD ConversionCode;
	_asm {
		push si
		push di
		push ds
		mov ax, 714Fh             // Find Next File
		mov bx, Handle            // search handle from Find First File
		les di, dword ptr Result  // address of results buffer
		mov si, DateTimeFormat    // format for returned date and time
		int 21h
		pop ds
		pop di
		pop si
		jc error
		mov [ConversionCode], cx  // UNICODE to OEM/ANSI conversion ok?
		}
	return 1;
error:
	return 0;
}

BOOL FindClose(HANDLE Handle)
{
	_asm {
		mov ax, 71A1h   // Find Close
		mov bx, Handle  // search handle from Find First File
		int 21h
		jc  error
		}
	return 1;
error:
	return 0;
}
#endif

void fail(const char *s) { puts(s); exit(1); }

#ifdef LFN
#define FILEINFO	WIN32_FIND_DATA
#else
#define FILEINFO	struct find_t
#endif

void do_filename(int dirpos, const char *dir, FILEINFO *pinfo);
void do_directory(int dirpos, const char *s);

static char dirname[2048] = {0};

static unsigned long ffirst = 0, fnext = 0;
static int print_it = 0;

main(int argc, char *argv[])
{
	clock_t c1, c2;
	
	c1 = clock();
	do_directory(0, (argc < 2) ? "C:" : argv[1]);
	c2 = clock();
	
	printf("Find First calls: %lu\n", ffirst);
	printf("Find Next calls:  %lu\n", fnext);
	printf("Clocks:           %lu\n", c2 - c1);
	printf("Calls/Clock:      %lu\n", (ffirst + fnext) / (c2 - c1));
	return 0;
}

void do_directory(int dirpos, const char *s)
{
    struct find_t info;
	int prev_dirpos;
	char *s2 = (char *) malloc(2048);
	if (! s2) fail("Insufficient memory");

	if ((prev_dirpos = dirpos) == 0)
		strcpy(dirname, s);
	else
	{
		strcat(dirname, "\\");
		strcat(dirname, s);
	}
	dirpos = strlen(dirname);
	
	strcpy(s2, dirname);
	strcat(s2, "\\*.*");
	
    #define ATTRIB (_A_NORMAL | _A_SUBDIR | _A_RDONLY)
		
#ifdef LFN
{
	HANDLE h;
	WIN32_FIND_DATA info;
	ffirst++;
	if ((h = FindFirstFile(s2, &info, ATTRIB, DATETIME_DOS)) == 0)
		return;
    do_filename(dirpos, dirname, &info);
	while (FindNextFile(h, &info, DATETIME_DOS))
	{
		fnext++;
        do_filename(dirpos, dirname, &info);
	}
	FindClose(h);
}
#else
	ffirst++;
    if (_dos_findfirst(s2, ATTRIB, &info) != 0)
		return;
    do_filename(dirpos, dirname, &info);
    while (_dos_findnext(&info) == 0)
	{
		fnext++;
        do_filename(dirpos, dirname, &info);
	}
#endif	
	
	dirname[prev_dirpos] = '\0';
	free(s2);
}

#ifdef LFN
void do_filename(int dirpos, const char *dir, WIN32_FIND_DATA *pinfo)
{
	if (pinfo->dwFileAttributes & FA_DIREC)
	{
		if (pinfo->cFileName[0] != '.')
			do_directory(dirpos, pinfo->cFileName);
	}
	else if (print_it)
		printf("%s\\%s\n", dir, pinfo->cFileName);
}
#else
void do_filename(int dirpos, const char *dir, struct find_t *pinfo)
{
	if (pinfo->attrib & FA_DIREC)
	{
		if (pinfo->name[0] != '.')
			do_directory(dirpos, pinfo->name);
	}
	else if (print_it)
		printf("%s\\%s\n", dir, pinfo->name);
}
#endif
