/* NE.H -- New Executable (segmented executable) format */

#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC		0x454C			/* linear executable */
#define LXMAGIC		0x584C			/* IBM OS/2 2.0 linear executable */
#define PEMAGIC		0x4550			/* NT portable executable */
#define W3MAGIC		0x3357			/* WIN386.EXE signature */
#define SZMAGIC		0x5A53			/* SZ -- compressed file */

#define DLL_FLAG    0x8000
#define APPLOAD		0x0800			/* self-loading Win app */
#define NEMATH		0x0080
#define NEI386		0x0040
#define NEI286		0x0020
#define NEI086		0x0010

#define NSTYPE		0x0007
#define NSCODE		0x0000
#define NSDATA		0x0001
#define NSRELOC		0x0100
#define NS32BIT		0x2000

#pragma pack(1)

typedef struct
{
    unsigned    ne_magic;
    char        ne_ver;			/* linker version # */
    char        ne_rev;
    unsigned    ne_enttab;		/* NE offset of entry table */	
    unsigned    ne_cbenttab;	/* #bytes in entry table */
    long        ne_crc;
    unsigned    ne_flags;		/* DLL vs. program, APPLOAD, etc. */
    unsigned    ne_autodata;	/* logical segment# (LSN) of DGROUP */
    unsigned    ne_heap;
    unsigned    ne_stack;
    void far *  ne_csip;		/* initial entry point (LSN:ofs) */
    void far *  ne_sssp;
    unsigned    ne_cseg;		/* number of segments */
    unsigned    ne_cmod;		/* number of WORDs in module ref tbl */
    unsigned    ne_cbnrestab;   /* #bytes in nonres name table */
    unsigned    ne_segtab;		/* NE offset of segment table */
    unsigned    ne_rsrctab;     /* *aligned* NE offset of resource table */
    unsigned    ne_restab;      /* NE offset of resident name table */
    unsigned    ne_modref;		/* NE offset of module reference table */
    unsigned    ne_impname;		/* NE offset of imported names table */
    long        ne_nonrestab;	/* file (NOT NE) offset of nonres name tbl */
	unsigned	ne_cmovent;
	unsigned	ne_align;		/* segment align shift count: segs, rsrc */
	unsigned	ne_cres;
	unsigned char	ne_exetyp;	/* target OS; 1=OS/2 2=Windows */
	unsigned char ne_flagsothers;
	unsigned	ne_fastload;
	unsigned	ne_lenfastload;
	unsigned	ne_reserved;
	unsigned char	ne_expwinvermin;	/* expected Windows version */
	unsigned char	ne_expwinvermaj;
} NEWEXE;

typedef struct {
	unsigned short ns_sector;
	unsigned short ns_cbseg;
	unsigned short ns_flags;
	unsigned short ns_minalloc;
	} SEG; 
	
typedef struct {
	unsigned char ep_flags;
	unsigned short ep_int_3f;
	unsigned char ep_segnum;
	unsigned short ep_offset;
	} ENTRY_MOVEABLE;

typedef	struct {
	unsigned char ep_flags;
	unsigned short ep_offset;
	} ENTRY_FIXED;
			
typedef struct {
	union {
		ENTRY_MOVEABLE ep_moveable;
		ENTRY_FIXED ep_fixed;
		} ep_union;
	} ENTRY_POINT;
	
typedef struct {
	unsigned char eb_nentry;
	unsigned char eb_segnum;
	} BUNDLE_HEADER;
	
typedef struct {
	BUNDLE_HEADER eb_hdr;
	ENTRY_POINT eb_epoint [1] ;
	} ENTRY_BUNDLE;

typedef struct {
	char nr_stype;
	char nr_flags;
	unsigned short nr_soff;
	union {
		struct {
			unsigned char nr_segno;
			unsigned char nr_res;
			unsigned short nr_entry;
			} nr_intref;
		struct {
			unsigned short nr_mod;
			unsigned short nr_proc;
			} nr_import;
		struct {
			unsigned short nr_ostype;
			unsigned short nr_osres;
			} nr_osfix;
		} nr_union;
	} RELOC_ITEM;
	
#define NRSTYP		0x0f
#define NRSBYT		0x00
#define NRSSEG		0x02
#define NRSPTR		0x03
#define NRSOFF		0x05
#define NRSPTR48	0x0b
#define NRSOFF32	0x0d

#define NRADD		0x04
#define NRRTYP		0x03
#define NRRINT		0x00
#define NRRORD		0x01
#define NRRNAM		0x02
#define NRROSF		0x03

#define MOVEABLE	0xFF

