/* 
PROTTAB.C
Protected mode version of FTAB from UndocDOS 2e
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include "dpmish.h"
#include "prot.h"
#include "vxdcalls.h"

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;

void fail(const char *s, ...)           { puts(s); _dos_exit(1); }

int real_main(int argc, char *argv[])   { return 0; }

int pmode_main(int argc, char *argv[])
{
	char *prefix;
    void far *tab;
	DWORD lin;
	DWORD filter;
	char *name;
	DWORD addr = 0;
	int no_filter = 1;
    WORD seg, ofs;
    int num, size, i;

    if (argc < 3)
        fail("usage: prottab <linear> <num> [size] [prefix] [filter] "
			 "[name=baseaddr]");

	sscanf(argv[1], "%08lX", &lin);
	printf("; Table at %08lXh\n", lin);
    
    sscanf(argv[2], "%04X", &num);
    
    size = (argc > 3) ? atoi(argv[3]) : 4;  /* default to DWORD table */
	
    if (! (tab = map_linear(lin, (size * num))))
		fail("Can't map memory");
	
	prefix = (argc > 4) ? argv[4] : "x";
	
	if (argc > 5)
	{
		no_filter = 0;
		sscanf(argv[5], "%08lX", &filter);
		printf("; Filter all out entries = %lXh\n", filter);
	}
	
	if (argc > 6)
	{
		char *s = strchr(argv[6], '=');
		if (! s) fail("reladdr usage: name=addr");
		sscanf(s+1, "%08lX", &addr);
		*s = '\0';
		name = argv[6];
		printf("; %s=%08lX\n", name, addr);
	}
	
	#define PRINT_TABLE(TYPE, ptr, mask) { \
		TYPE far *ptr; \
		for (i=0, ptr=(TYPE far *)tab; i<num; i++, ptr++) \
			if (no_filter || (*ptr != filter)) { \
				printf(mask ## "\t%s_%04X", *ptr, prefix, i); \
				if (addr) \
					printf("\t; %s+%0lX", name, *ptr - addr); \
				printf("\n"); \
				} \
		}

    switch (size)
    {
        case 1: PRINT_TABLE(BYTE, btab, "%02X"); break;
        case 2: PRINT_TABLE(WORD, wtab, "%04X"); break;
        case 4: PRINT_TABLE(DWORD, dtab, "%08lX"); break;
        default: fail("size only 1 (byte), 2 (word), 4 (dword)");
    }

	free_mapped_linear(tab);
    return 0;
}
