/* RESDUMP.C */

/*
	CHANGES:
		1.11 (4/19/92):  beginning to do VERSIONINFO
		1.12 (5/29/92):  beginning to do cursors, icons, fonts, bitmaps;
						 if dialog has font, print name;
						 allow UNKNOWN_OS since \winword\prev.fon uses it
		1.13 (6/21/92):  beginning to do Visual Basic; check if USERDATA
						 is Visual Basic .FRM file
			 (7/2/92):   added -win10 switch (see do_menu10)
						 added -type x switch (only show resources of 
						 type x; only works for RT_X < 32)
					     fixed bug in argv handling!
		1.14 (7/14/92):	 now does .RES files (function resfile)
					     currently can't verify if .RES file or not
						 added NAMETAB support
						 menu display now shows SEPARATOR properly
						 better handling for named resources (using idname)
						 added pt. size to FONT display for dialogs
		1.15 (9/11/92):	 Added Ansi to OEM character conversion; wasn't
						 displaying German, etc. resources properly!!
						 Thanks to Haselier & Fahnenstein for code
		1.16 (11/17/92): Fixed bug: couldn't read SETUP.EXE as NE file;
						 had start_newexe as WORD rather than DWORD?!?!
						 (Found by Geoff C., of course!)
		2.00 (2/3/93):	 Converted to C++
						 Backed out VB stuff
						 Added -DBCS switch to disable ANSI to OEM conversion
		2.01 (3/11/93):	 Check size of rsrctab; if 0, no resources
		2.10 (3/14/93):	 Adding Win32 PE support; RESDMP32.C
		2.11 (5/10/93):  Added support for Kanji menu items in do_menu()
		2.12 (5/21/93):	 Check for Win32 .RES files!
		2.13 (6/18/93):  Kanji 0x1e/0x1f support not just for menus!!
			
	TO DO:
		Japan stuff for MENUs needs to be for all text, in parentheses
*/

/*
	Try to handle VB programs
		USERDATA 01 -- 02 10 8B
		USERDATA 02 or 03 -- 23 CC 24 00 -- .FRM files!!
		USERDATA 02 or 03+ -- FC 01 
			
	Should show module name, description, name tables, etc. (e.g. *.FOT)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef signed long LONG;
typedef unsigned long DWORD;

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC		0x454C			/* linear executable */
#define LXMAGIC		0x584C			/* IBM OS/2 2.0 linear executable */
#define W3MAGIC		0x3357			/* WIN386.EXE signature */
#define PEMAGIC		0x4550			/* Win32 portable executable */
#define SZMAGIC		0x5A53			/* SZ -- compressed file */

#define DLL_FLAG    0x8000

#define UNKNOWN_OS	0
#define WIND_FLAG	2

#define	NERESBYTES	9

/* lifted from WINDOWS.H */
/* Predefined Resource Types */
#define MAKEINTRESOURCE(x)	((x) + 0x8000)		/* for file only!!!! */
#define RT_CURSOR           MAKEINTRESOURCE(1)
#define RT_BITMAP           MAKEINTRESOURCE(2)
#define RT_ICON             MAKEINTRESOURCE(3)
#define RT_MENU             MAKEINTRESOURCE(4)
#define RT_DIALOG           MAKEINTRESOURCE(5)
#define RT_STRING           MAKEINTRESOURCE(6)
#define RT_FONTDIR          MAKEINTRESOURCE(7)
#define RT_FONT             MAKEINTRESOURCE(8)
#define RT_ACCELERATOR      MAKEINTRESOURCE(9)
#define RT_RCDATA           MAKEINTRESOURCE(10)
#define RT_ERRORTABLE		MAKEINTRESOURCE(11)	/* from Duncan Murdoch */
#define RT_GROUPCURSOR		MAKEINTRESOURCE(12)	/* windows.inc */
#define RT_GROUPICON		MAKEINTRESOURCE(14)	/* windows.inc */
#define RT_NAMETABLE		MAKEINTRESOURCE(15)	/* from Open Tools */
#define RT_VERSIONINFO		MAKEINTRESOURCE(16) /* version info */
#define RT_TRUETYPEFONT		MAKEINTRESOURCE(0xCC) /* from Matt P. */

char *restypename[] = {
	"",
	"CURSOR",
	"BITMAP",
	"ICON",
	"MENU",
	"DIALOG",
	"STRING",
	"FONTDIR",
	"FONT",
	"ACCELERATOR",
	"RCDATA",
	"ERRORTABLE",
	"GROUPCURSOR",
	"",
	"GROUPICON",
	"NAMETABLE",
	"VERSIONINFO"
	} ; 

/* #204 has something to do with TrueType */
/* have also seen #400 used */

#define DS_SETFONT          0x40L   /* User specified font for Dlg controls */

#define MF_POPUP           0x0010
#define MF_END             0x0080
#define MF_STRING          0x0000
#define MF_BITMAP          0x0004
#define MF_OWNERDRAW       0x0100
#define MF_SEPARATOR       0x0800

#define BS_PUSHBUTTON      0x00L
#define BS_DEFPUSHBUTTON   0x01L
#define BS_CHECKBOX        0x02L
#define BS_AUTOCHECKBOX    0x03L
#define BS_RADIOBUTTON     0x04L
#define BS_3STATE          0x05L
#define BS_AUTO3STATE      0x06L
#define BS_GROUPBOX        0x07L
#define BS_USERBUTTON      0x08L
#define BS_AUTORADIOBUTTON 0x09L
#define BS_PUSHBOX         0x0AL
#define BS_OWNERDRAW       0x0BL
#define BS_LEFTTEXT        0x20L

#define SS_ICON            0x03L

#define ES_MULTILINE       0x0004L

#pragma pack(1)

typedef struct
{
    WORD    ne_magic;
    char        ne_ver;
    char        ne_rev;
    WORD    ne_enttab;
    WORD    ne_cbenttab;
    long        ne_crc;
    WORD    ne_flags;
    WORD    ne_autodata;			/* dgroup */
    WORD    ne_heap;
    WORD    ne_stack;
    void far *  ne_csip;
    void far *  ne_sssp;
    WORD    ne_cseg;
    WORD    ne_cmod;
    WORD    ne_cbnrestab;           /* size of nonres name table */
    WORD    ne_segtab;
    WORD    ne_rsrctab;             /* offset of resource table */
    WORD    ne_restab;              /* offset of resident name table */
    WORD    ne_modref;
    WORD    ne_impname;
    long        ne_nonrestab;           /* offset of nres table from begin */
	WORD	ne_cmovent;
	WORD	ne_align;				/* segment align shift count */
	WORD	ne_cres;		/* argh!  always 0! */
	BYTE	ne_exetyp;
	char		ne_res[NERESBYTES];		/* pad struct to 64 bytes */
	/* more stuff, but we don't care */
} NEWEXE;

/* following from MSJ, September 1991, p. 65 */

typedef struct {
	WORD rt_type;
	WORD rt_count;
	DWORD rt_proc;
} RESTYPEINFO;

typedef struct {
	WORD rn_offset;
	WORD rn_length;
	WORD rn_flags;
	WORD rn_id;
	WORD rn_handle;
	WORD rn_usage;
} RESNAMEINFO;

typedef RESNAMEINFO *PRESNAMEINFO;

typedef struct {
	RESTYPEINFO rt;
	RESNAMEINFO rn[1];
} RESTAB;

typedef RESTAB *PRESTAB;

void fail(char *s) { fputs("resdump fail: ", stdout); puts(s); exit(1); }

void fail_perror(void)	{ perror("resdump fail"); exit(1); }

void usage()
{
    puts("RESDUMP displays information about resources in a Windows .RES");
    puts("file or executable (EXE, DLL, DRV, etc.). Detailed information");
    puts("is provided for dialog boxes, controls, menus, string tables,");
    puts("accelerator tables, and version resources.");
    puts("");
    puts("To display resources in a Windows .RES or executable:");
    puts("         RESDUMP [options] res_or_exe_file");
    puts("example: resdump \\windows\\winfile.exe");
    puts("");
    puts("To display resources only of a given type:");
    puts("         RESDUMP -TYPE [type] res_or_exe_file");
    puts("example: resdump -type menu \\windows\\winfile.exe");
    puts("         resdump -type menu -type dialog -hex \\foo\\bar.exe"); 
    puts("types:   CURSOR BITMAP ICON MENU DIALOG STRINGTAB FONTDIR FONT");
    puts("         ACCEL RCDATA ERRORTAB CURSDIR ICONDIR NAMETAB VERSION");
    puts("");
    puts("To also display (x,y) locations for dialog items: -VERBOSE");
    puts("To also dump bytes (hex) for each resource:       -HEX");
    puts("For Windows 1.0 programs:                         -WIN10");
	puts("To disable ANSI to OEM conversion (Japan):        -DBCS");
	puts("");
	puts("Also works with Win32 (NT) portable executable (PE) files");
	exit(1);
}

extern void pe2(DWORD start);	// for Win32 PE files -- in RESDMP32.C
void resfile(char *name);
void newexe(char *name);
int ne_read(int f, void *p, DWORD ne_offset, WORD bytes);
int exe_read(int f, void *p, DWORD exe_offset, WORD bytes);
void *allocate(size_t x, size_t y);
void count2null(char *s, int len);
void dump(BYTE far *fp, WORD bytes, char *mask, DWORD addr);
BYTE *AnsiToOemBuf(BYTE *buf);

void do_restype_switch(WORD type, DWORD offset, DWORD length, WORD id);
void do_restab(WORD offset);
void do_dialog(DWORD offset, DWORD length, WORD id);
void do_menu(DWORD offset, DWORD length, WORD id);
void do_menu10(DWORD offset, DWORD length, WORD id);
void do_string(DWORD offset, DWORD length, WORD id);
void do_version(DWORD offset, DWORD length, WORD id);
void do_accel(DWORD offset, DWORD length, WORD id);
void do_cursdir(DWORD offset, DWORD length, WORD id);
void do_icondir(DWORD offset, DWORD length, WORD id);
void do_fontdir(DWORD offset, DWORD length, WORD id);
void do_bitmap(DWORD offset, DWORD length, WORD id);
void do_userdata(DWORD offset, DWORD length, WORD id);
void do_nametab(DWORD offset, DWORD length, WORD id);
void do_vb(BYTE *data, DWORD length);
void dump_generic(DWORD offset, DWORD length);

static DWORD start_newexe = 0;
static NEWEXE hdr = {0} ;
static DWORD align;
int f;
static int verbose = 0;
static int do_dump = 0;
static int do_win10 = 0;
static int no_vers = 0;
static int no_convert = 0;
static unsigned long do_type = 0;	// all types
static char idname[128] = {0} ;		// used by resfile - hack!

#define CMP(s1, s2)		(strcmp((s1), (s2)) == 0)

main(int argc, char *argv[])
{
	char *s;
	int i;
	
	fputs("Windows Resource Dumper version 2.13\n", stderr);
	fputs("from \"Unauthorized Windows\" by Andrew Schulman "
		  "(IDG Books, 1994)\n", stderr);
	fputs("Copyright (c) 1992-94 Andrew Schulman.  All rights reserved.\n\n",
		stderr);
	
	if (argc < 2)
		usage();
	
	for (i=1; i<argc; i++)
		if ((argv[i][0] == '-') || (argv[i][0] == '/'))
		{
			s = strupr(&argv[i][1]);
			if (*s == '?')
				usage();
			else if (CMP(s, "VERBOSE"))
				verbose = 1;
			else if (CMP(s, "HEX"))
				do_dump = 1;
			else if (CMP(s, "NOVERS"))
				no_vers = 1;
			else if (CMP(s, "WIN10"))
				do_win10 = 1;	// ignore OS type
			else if (CMP(s, "DBCS"))
				no_convert = 1;
			else if (CMP(s, "TYPE"))
			{
				char *type = strupr(argv[++i]);
				if (! type) usage();
				if (CMP(type, "CURSOR")) do_type |= (1L << RT_CURSOR);
				else if (CMP(type, "BITMAP")) do_type |= (1L << RT_BITMAP);
				else if (CMP(type, "ICON")) do_type |= (1L << RT_ICON);
				else if (CMP(type, "MENU")) do_type |= (1L << RT_MENU);
				else if (CMP(type, "DIALOG")) do_type |= (1L << RT_DIALOG);
				else if (CMP(type, "STRINGTAB")) do_type |= (1L << RT_STRING);
				else if (CMP(type, "FONTDIR")) do_type |= (1L << RT_FONTDIR);
				else if (CMP(type, "FONT")) do_type |= (1L << RT_FONT);
				else if (CMP(type, "ACCEL")) do_type |= (1L<<RT_ACCELERATOR);
				else if (CMP(type, "RCDATA")) do_type |= (1L<<RT_RCDATA);
				else if (CMP(type,"ERRORTAB")) do_type |= (1L<<RT_ERRORTABLE);
				else if (CMP(type,"CURSDIR")) do_type |= (1L<<RT_GROUPCURSOR);
				else if (CMP(type,"ICONDIR")) do_type |= (1L<<RT_GROUPICON);
				else if (CMP(type,"NAMETAB")) do_type |= (1L<<RT_NAMETABLE);
				else if (CMP(type,"VERSION")) do_type |= (1L<<RT_VERSIONINFO);
				else usage();
			}
			else
				usage();
		}
		else	
		{
			char buf[128];
			strcpy(buf, argv[i]);
			if (! strchr(buf, '.'))	// sort-of check for .ext
				strcat(buf, ".exe");
			if (strstr(strupr(buf), ".RES"))
				resfile(buf);
			else if (strstr(strupr(buf), ".FRM"))
				;
			else
				newexe(buf);
		}
	return 0;
}

#define WANT_TYPE(x) ((do_type == 0) || (do_type & (1L << (x))))

// information from "Microsoft Windows 3.0 Internal Resource Formats,"
// Microsoft doc, 11/12/90 
// is there any way to verify if it's really a .RES file??
void resfile(char *name)
{
	WORD type=0, id, flags;
	BYTE b;
	DWORD dw, len;
	BYTE *typename=0;
	DWORD offset = 0;
	DWORD fsize;
	if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
		fail_perror();
	fsize = filelength(f);
	while (offset < fsize)
	{
		/* get type */
		exe_read(f, &b, offset, 1);
		offset++;
		if (b == 0xFF)	// standard type
		{
			exe_read(f, &type, offset, 2);
			offset += 2;
		}
		else
		{
			typename = (BYTE *) allocate(1, 128);
			exe_read(f, typename, offset, 127);	// named type
			if (typename[0] == 0)	// illegal res for Win32 .RES files!
				fail("Sorry, can't do Win32 .RES files yet "
					 "(can do EXEs and DLLs though)");
			offset += (strlen(typename) + 1);
		}
		
		/* get ID */
		exe_read(f, &b, offset, 1);
		if (b == 0xFF)	// ordinal ID
		{
			offset++;
			exe_read(f, &id, offset, 2);
			offset += 2;
		}
		else
		{
			id = -1;
			exe_read(f, idname, offset, 127);	// named resource
			// not sure about this; just for TYPE or for NAME too???
			if (idname[0] == 0)	// illegal res for Win32 .RES files!
				fail("Sorry, can't handle Win32 .RES files yet");
			offset += (strlen(idname) + 1);
		}
		exe_read(f, &flags, offset, 2);
		offset += 2;
		exe_read(f, &len, offset, 4);
		offset += 4;
		if (! WANT_TYPE(type | 0x8000))
		{
			offset += len;
			continue;
		}
		switch (type)
		{
			case 0:
				// named resource type -- do later???
				break;
			default:
				do_restype_switch(type | 0x8000, offset, len, id);
				break;
		}
		if (do_dump)
			dump_generic(offset, len);
		putchar('\n');
		offset += len;
	}
	
	if (typename) free(typename);
	close(f);
}

void newexe(char *name)
{
    WORD w;
	WORD i;
	
	puts(name);
	
	if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
		fail_perror();
	exe_read(f, &w, 0, 2);	// read first two bytes of file
    if (w != EMAGIC)
	{
		if (w == SZMAGIC)
			fail("compressed file (SZ) -- use EXPAND.EXE");
		else
			fail("not an executable");
	}
	exe_read(f, &w, 0x18, 2);
	if (w < 0x40)
		fail("not a segmented (new) executable");
	exe_read(f, &start_newexe, ENEWHDR, 4);
	exe_read(f, &hdr, start_newexe, sizeof(NEWEXE));
	
	switch (hdr.ne_magic)
	{
		case LEMAGIC:
		case LXMAGIC:
		case W3MAGIC:
			fail("this is a linear executable, not a segmented executable");
			break;
		case PEMAGIC:
			printf("Win32 portable executable (PE) file:\n");
			pe2(start_newexe);	// also accept PE files
			goto done;
		case NEMAGIC:
			break;	// okay
		default:
			fail("not a segmented (NE) executable");
	}
	
	if (! do_win10)
	if ((hdr.ne_exetyp != WIND_FLAG) &&	/* NOT a bit flag */
		(hdr.ne_exetyp != UNKNOWN_OS))
		fail("not a Windows program");
	
	align = 1 << hdr.ne_align;
	
#if 0	
	// NO!  hdr.ne_cres always 0 in Win app (used for OS/2 resources???)
	if (hdr.ne_cres == 0)
		fail("no resources");
#endif	
	
	// both tdump and exehdr use this rule to determine size of rsrc table!!
	if ((hdr.ne_restab - hdr.ne_rsrctab) > 0)	
		do_restab(hdr.ne_rsrctab);
	else
		fail("no resources");
	
done:	
	close(f);
}

void notimp(char *s, unsigned id, DWORD res_offset, DWORD res_length)
{
	printf("%s %u (%02Xh) ", s, id, id);
	if (verbose)
		printf("(ofs=%lu, len=%lu)", res_offset, res_length);
	printf("\n");
}

#define NOTIMP(s, id) { notimp(s, id, res_offset, res_length); }

char *type_name(WORD offset)
{
	static char buf[128];
	ne_read(f, buf, offset, 128);
	return &buf[6];		// name table?
}

#define WANT_TYPE(x) ((do_type == 0) || (do_type & (1L << (x))))

void do_restype_switch(WORD type, DWORD res_offset, DWORD res_length, WORD id)
{
	void (*do_func)(DWORD offset, DWORD length, WORD id) = 0;
	
	switch (type)
	{
		case RT_CURSOR:		NOTIMP("CURSOR", id);	break;
		case RT_BITMAP:		do_func = do_bitmap;	break;
		case RT_ICON:		NOTIMP("ICON", id);		break;
		case RT_MENU:		do_func = do_win10? do_menu10 : do_menu; break;
		case RT_DIALOG:		do_func = do_dialog;	break;
		case RT_STRING:		do_func = do_string;	break;
		case RT_FONTDIR:	do_func = do_fontdir;	break;
		case RT_FONT:		NOTIMP("FONT", id);		break;
		case RT_ACCELERATOR:do_func = do_accel;		break;
		case RT_RCDATA:		do_func = do_userdata;	break;
		case RT_ERRORTABLE: NOTIMP("ERRORTABLE", id); break;
		case RT_GROUPCURSOR:do_func = do_cursdir;	break;
		case RT_GROUPICON:	do_func = do_icondir;	break;
		case RT_NAMETABLE:	do_func = do_nametab;	break;
		case RT_VERSIONINFO:do_func = do_version;	break;
		default:
			if (type & 0x8000)
				printf("#%u", type & ~0x8000);
			else
				printf("#%u", type); 
			NOTIMP("", id); 
			break;
	}
	
	if (do_func != 0)
		(*do_func)(res_offset, res_length, id);
}
			
void do_restab(WORD offset)
{
	RESTYPEINFO *rt = (RESTYPEINFO *) allocate(1, sizeof(RESTYPEINFO));
	RESNAMEINFO *rn, *rn2;
	DWORD res_offset, res_length;
	WORD align_shift;
	WORD id;
	int i;
	
	/* printf("offset = %04x\n", offset); */
	
	ne_read(f, &align_shift, offset, sizeof(short));
	offset += sizeof(short);
	
	for (;;)
	{
		ne_read(f, rt, offset, sizeof(RESTYPEINFO));
		if (rt->rt_type == 0)
			break;
		if (rt->rt_count == 0)
			fail("no resources?");
		if (((long) rt->rt_count * sizeof(RESNAMEINFO)) > 0xFFFFUL)
			fail("no resources??");
		rn = (RESNAMEINFO *) allocate(rt->rt_count, sizeof(RESNAMEINFO));
		read(f, rn, sizeof(RESNAMEINFO) * rt->rt_count);
		for (i=rt->rt_count, rn2=rn; i--; rn2++)
		{
			res_offset = (DWORD) rn2->rn_offset << align_shift;
			res_length = (DWORD) rn2->rn_length << align_shift;
			if (rn2->rn_id & 0x8000)
				id = rn2->rn_id & ~0x8000;		// for now!!!

			if (! WANT_TYPE(rt->rt_type))
				continue;
		
			do_restype_switch(rt->rt_type, res_offset, res_length, id);
			
			if (do_dump)
				dump_generic(res_offset, res_length);
			putchar('\n');
		}
end:
		offset += (sizeof(RESTYPEINFO) + 
				  (rt->rt_count * sizeof(RESNAMEINFO)));
		free(rn);
	}
	
	free(rt);
}

/**********************************************************************/

typedef struct {
	DWORD dlg_style;
	BYTE num_items;
	WORD x, y, cx, cy;
	// rest is variable length!
} DLGHEADER;

typedef struct {
	WORD x, y, cx, cy, id;
	DWORD ctrl_style;
	// rest is variable length!
} CTRLHEADER;

void do_dialog(DWORD offset, DWORD length, WORD id)
{
	DWORD style;
	BYTE *dlg = (BYTE *) allocate(1, length);
	DLGHEADER *dlghdr = (DLGHEADER *) dlg;
	BYTE *control;
	BYTE *ctrl;
	BYTE *text;
	BYTE *menu_name = 0;
	BYTE *class_name = 0;
	BYTE *class_type;
	int num_controls;
	int control_type;
	int typ_len, text_len;
	int len;
	int i;
	
	exe_read(f, dlg, offset, length);
	num_controls = dlghdr->num_items;
	text = dlg + sizeof(DLGHEADER);
	if (*text == 0) 
		text++;
	else if (*text == 0xFF) 
	{
		menu_name = text;
		text += 3;
	}
	else
	{
		menu_name = text;
		text += (strlen(text) + 1);
	}
	if (*text)
		class_name = text;
	text += (strlen(text) + 1);
	AnsiToOemBuf(text);
	if (id == 0xFFFF)
		printf("DIALOG \"%s\" \"%s\"", idname, text);
	else
		printf("DIALOG %u (%02Xh), \"%s\"", id, id, text);
	if (verbose)
		printf(" [%u,%u,%u,%u]", 
			dlghdr->x, dlghdr->y, dlghdr->cx, dlghdr->cy);
	putchar('\n');
	if (menu_name)
	{
		if (*menu_name == 0xFF)
		{
			WORD ord = *((WORD *) &menu_name[1]);
			printf("MENU #%u (%02Xh)\n", ord);
		}
		else
			printf("MENU \"%s\"\n", menu_name);
	}
	if (class_name)
		printf("CLASS \"%s\"\n", class_name);
	ctrl = text + strlen(text) + 1;
	if (dlghdr->dlg_style & DS_SETFONT)
	{
		WORD ptsize = *((WORD *) ctrl);
		ctrl += 2;
		printf("FONT \"%s\" (%u pt.)\n", ctrl, ptsize);
		ctrl += (strlen(ctrl) + 1);
	}
	for (i=num_controls; i--; )
	{
		// this code is ridiculous -- fix it!
		CTRLHEADER *ctrlhdr = (CTRLHEADER *) ctrl;
		control = ctrl;
		id = ctrlhdr->id;
		style = ctrlhdr->ctrl_style;
		if (control[sizeof(CTRLHEADER)] & 0x80)
		{
			switch (control[sizeof(CTRLHEADER)])
			{
				case 0x80: // BUTTON
					switch ((WORD) style)	// not bit mask???
					{
						case BS_PUSHBUTTON:
							class_type = "PUSHBUTTON     "; break;
						case BS_DEFPUSHBUTTON:
							class_type = "DEFPUSHBUTTON  "; break;
						case BS_CHECKBOX:
							class_type = "CHECKBOX       "; break;
						case BS_AUTOCHECKBOX:
							class_type = "AUTOCHECKBOX   "; break;
						case BS_RADIOBUTTON:
							class_type = "RADIOBUTTON    "; break;
						case BS_3STATE:
							class_type = "3STATE         "; break;
						case BS_GROUPBOX:
							class_type = "GROUPBOX       "; break;
						case BS_USERBUTTON:
							class_type = "USERBUTTON     "; break;
						case BS_AUTORADIOBUTTON:
							class_type = "AUTORADIOBUTTON"; break;
						case BS_PUSHBOX:
							class_type = "PUSHBOX        "; break;
						case BS_OWNERDRAW:
							class_type = "OWNERDRAW      "; break;
						case BS_LEFTTEXT:
							class_type = "LEFTTEXT       "; break;
						default:
							class_type = "BUTTON         "; break;
					}
					break;
				case 0x81: 
					// should do other ES_ styles
					if (style & ES_MULTILINE)
						class_type =     "EDITMULTILINE  ";
					else
						class_type =     "EDIT           "; 
					break;
				case 0x82:
					// should do other SS_ styles
					class_type =         "STATIC         "; 
					break;
				case 0x83: 
					// should do LBS_ styles
					class_type =         "LISTBOX        "; 
					break;
				case 0x84: 
					// should do SBS_ styles
					class_type =         "SCROLLBAR      "; 
					break;
				case 0x85: 
					// should do CBS_ styles
					class_type =         "COMBOBOX      "; 
					break;
				default:   
					class_type =         "????          "; 
					break;
			}
			ctrl++;
			text = &control[sizeof(CTRLHEADER)+1];
			printf("    %s %u (%02Xh), \"%s\"",
				class_type, id, id, AnsiToOemBuf(text));
		}
		else
		{
			class_type = &control[sizeof(CTRLHEADER)];
			len = strlen(class_type);
			ctrl += (len + 1);
			text = &control[sizeof(CTRLHEADER)] + (len + 1);
			printf("    \"%s\" %u (%02Xh), \"%s\"",
				class_type, id, id, text);
		}
		if (verbose)
			printf(" [%u,%u,%u,%u]", 
				ctrlhdr->x, ctrlhdr->y, ctrlhdr->cx, ctrlhdr->cy);
		putchar('\n');
		ctrl += (sizeof(CTRLHEADER) + strlen(text) + 2);
	}
	free(dlg);
}

/**********************************************************************/

void do_menu(DWORD offset, DWORD length, WORD id)
{
	BYTE *m, *p;
	BYTE *menu = (BYTE *) allocate(1, length);
	WORD flags;
	int saw_end = 0;
	exe_read(f, menu, offset, length);
	if (id == 0xFFFF)
	{
#if 1
/* following needed for all strings, not just menus!!! */
#else
		/*
			Instead of "&Text", Kanji Windows menu items
			look like:  "Text (1ET1Fkk)" where 0x1E indicates
			that next character (e.g., 'T') is the English hotkey
			and 0x1F indicates that next character (which may be
			double byte!!) is the Kanj hotkey.  Note parentheses!
			But 0x1E and 0x1F can't be output directly on NEC
			machines!
		*/
		if (no_convert)		/* Japan */
		{
			char *s;
			printf("MENU \"");
			s = idname;
			while (*s && (*s != '('))
			{
				putchar(*s);
				s++;
			}
			if (*s == '(')
			{
				putchar(*s);
				s++;
				if (*s == 0x1E)
					printf("\036");
				s++;
				putchar(*s);
				s++;
				if (*s == 0x1F)
					printf("\037");
			}
			while (*s)
			{
				putchar(*s);
				s++;
			}
			printf("\"\n");
		}
		else
#endif			
		{
			printf("MENU \"%s\"\n", AnsiToOemBuf(idname));
		}
	}
	else
		printf("MENU %u (%02Xh)\n", id, id);
	m = menu + 4;	// skip header
	while (m < (menu + length))
	{
		flags = *((WORD *) m);
		p = &m[2];
		if (flags & MF_POPUP)
		{
			printf("    POPUP \"%s\"\n", AnsiToOemBuf(p));
			m += (3 + strlen(p));
			saw_end = 0;
		}
		else
		{
			if (saw_end < 2)
			{
				if ((! flags) && (! *((WORD *) p)) && (! m[4]))
					printf("        SEPARATOR\n");
				else
					printf("        %u (%02Xh) \"%s\"\n", 
						*((WORD *) p),
						*((WORD *) p),
						AnsiToOemBuf(&m[4]));
			}
			m += (5 + strlen(&m[4]));
		}
		if (flags & MF_END)
		{
			printf("        END\n");
			saw_end++;
		}
	}
	free(menu);
}

// do 1.0-style menus

void do_menu10(DWORD offset, DWORD length, WORD id)
{
	BYTE *m, *p;
	BYTE *menu = (BYTE *) allocate(1, length);
	BYTE flags;
	int saw_end = 0;
	exe_read(f, menu, offset, length);
	printf("MENU %u (%02Xh)\n", id, id);
	m = menu;	// 1.0 menus have no header
	while (m < (menu + length))
	{
		flags = *((BYTE *) m);
		p = &m[1];
		if (flags & MF_POPUP)
		{
			printf("    POPUP \"%s\"\n", p);
			m += (2 + strlen(p));
			saw_end = 0;
		}
		else
		{
			if (saw_end < 2)
			{
				printf("        %u (%02Xh) \"%s\"\n", 
					*((BYTE *) p),
					*((BYTE *) p),
					&m[3]);
			}
			m += (4 + strlen(&m[3]));
		}
		if (flags & MF_END)
		{
			printf("        END\n");
			saw_end++;
		}
	}
	free(menu);
}

/**********************************************************************/

void do_string(DWORD offset, DWORD length, WORD id)
{
	BYTE buf[1024];	// workaround Borland problem with %*s
	BYTE *str = (BYTE *) allocate(1, length);
	BYTE *s;
	WORD ord;
	exe_read(f, str, offset, length);
	printf("STRINGTABLE %u (%02Xh)\n", id, id);
	// check string numbering scheme
	for (s=str, ord=(id-1)<<4; s<(str+length); ord++)
	{
		if (s[0])
		{
			memcpy(buf, &s[1], s[0]);
			buf[s[0]] = '\0';
			printf("    %u (%02Xh) \"%s\"\n", ord, ord, AnsiToOemBuf(buf));
		}
		s += (s[0] + 1);
	}
	free(str);
}

/**********************************************************************/

typedef struct {
	BYTE fFlags;
	WORD wEvent;
	WORD wId;
	} ACCEL;

/* lifted from WINDOWS.H */
#define VK_LBUTTON          0x01
#define VK_RBUTTON          0x02
#define VK_CANCEL           0x03
#define VK_MBUTTON          0x04
#define VK_BACK             0x08
#define VK_TAB              0x09
#define VK_CLEAR            0x0C
#define VK_RETURN           0x0D
#define VK_SHIFT            0x10
#define VK_CONTROL          0x11
#define VK_MENU             0x12
#define VK_PAUSE            0x13
#define VK_CAPITAL          0x14
#define VK_ESCAPE           0x1B
#define VK_SPACE            0x20
#define VK_PRIOR            0x21
#define VK_NEXT             0x22
#define VK_END              0x23
#define VK_HOME             0x24
#define VK_LEFT             0x25
#define VK_UP               0x26
#define VK_RIGHT            0x27
#define VK_DOWN             0x28
#define VK_SELECT           0x29
#define VK_PRINT            0x2A
#define VK_EXECUTE          0x2B
#define VK_SNAPSHOT         0x2C
#define VK_COPY				0x2C
#define VK_INSERT           0x2D
#define VK_DELETE           0x2E
#define VK_HELP             0x2F
#define VK_A				'A'
#define VK_Z				'Z'
#define VK_0				'0'
#define VK_9				'9'
#define VK_NUMPAD0          0x60
#define VK_NUMPAD9          0x69
#define VK_MULTIPLY         0x6A
#define VK_ADD              0x6B
#define VK_SEPARATOR        0x6C
#define VK_SUBTRACT         0x6D
#define VK_DECIMAL          0x6E
#define VK_DIVIDE           0x6F
#define VK_F1               0x70
#define VK_F16              0x7F
#define VK_NUMLOCK          0x90

void do_accel(DWORD offset, DWORD length, WORD id)
{
	ACCEL *accel = (ACCEL *) allocate(1, length);
	WORD count = length / sizeof(ACCEL);
	ACCEL *pa;
	WORD i;
	
	printf("ACCELERATOR %u (%02Xh)\n", id, id);
	exe_read(f, accel, offset, length);
	for (i=count, pa=accel; i--; pa++)
	{
		printf("    %d (%04xh) ", pa->wId, pa->wId);
		if (pa->fFlags & 0x04)		printf("SHIFT ");
		else if (pa->fFlags & 0x08)	printf("CTRL ");
		else if (pa->fFlags & 0x10) printf("ALT ");
		if (pa->wEvent >= 'A' && pa->wEvent <= 'Z')
			printf("VK_%c\n", pa->wEvent);
		else if (pa->wEvent >= '0' && pa->wEvent <= '9')
			printf("VK_%c\n", pa->wEvent);
		else if (pa->wEvent >= VK_NUMPAD0 && pa->wEvent <= VK_NUMPAD9)
			printf("VK_NUMPAD%d\n", pa->wEvent - VK_NUMPAD0);
		else if (pa->wEvent >= VK_F1 && pa->wEvent <= VK_F16)
			printf("VK_F%d\n", 1 + pa->wEvent - VK_F1);
		else switch (pa->wEvent)
		{
			case VK_LBUTTON: puts("VK_LBUTTON"); break;
			case VK_RBUTTON: puts("VK_RBUTTON"); break;
			case VK_CANCEL: puts("VK_CANCEL"); break;
			case VK_MBUTTON: puts("VK_MBUTTON"); break;
			case VK_BACK: puts("VK_BACK"); break;
			case VK_TAB: puts("VK_TAB"); break;
			case VK_CLEAR: puts("VK_CLEAR"); break;
			case VK_RETURN: puts("VK_RETURN"); break;
			case VK_SHIFT: puts("VK_SHIFT"); break;
			case VK_CONTROL: puts("VK_CONTROL"); break;
			case VK_MENU: puts("VK_MENU"); break;
			case VK_PAUSE: puts("VK_PAUSE"); break;
			case VK_CAPITAL: puts("VK_CAPITAL"); break;
			case VK_ESCAPE: puts("VK_ESCAPE"); break;
			case VK_SPACE: puts("VK_SPACE"); break;
			case VK_PRIOR: puts("VK_PRIOR"); break;
			case VK_NEXT: puts("VK_NEXT"); break;
			case VK_END: puts("VK_END"); break;
			case VK_HOME: puts("VK_HOME"); break;
			case VK_LEFT: puts("VK_LEFT"); break;
			case VK_UP: puts("VK_UP"); break;
			case VK_RIGHT: puts("VK_RIGHT"); break;
			case VK_DOWN: puts("VK_DOWN"); break;
			case VK_SELECT: puts("VK_SELECT"); break;
			case VK_PRINT: puts("VK_PRINT"); break;
			case VK_EXECUTE: puts("VK_EXECUTE"); break;
			case VK_SNAPSHOT: puts("VK_SNAPSHOT [VK_COPY]"); break;
			case VK_INSERT: puts("VK_INSERT"); break;
			case VK_DELETE: puts("VK_DELETE"); break;
			case VK_HELP: puts("VK_HELP"); break;
			case VK_MULTIPLY: puts("VK_MULTIPLY"); break;
			case VK_ADD: puts("VK_ADD"); break;
			case VK_SEPARATOR: puts("VK_SEPARATOR"); break;
			case VK_SUBTRACT: puts("VK_SUBTRACT"); break;
			case VK_DECIMAL: puts("VK_DECIMAL"); break;
			case VK_DIVIDE: puts("VK_DIVIDE"); break;
			case VK_NUMLOCK: puts("VK_NUMLOCK"); break;
			default:
				printf("%04x\n", pa->wEvent); break;	// ASCII char of VK_
		}
		if (pa->fFlags & 0x80)	// END
			break;
	}
	free(accel);
}

/**********************************************************************/

/* from 3.1 SDK, Vol. 4 */
typedef struct {
	WORD cbBlock;
	WORD cbValue;
#if 0	
	char szKey[];
	BYTE abValue[];
#endif	
	} INFOBLOCK;

#if 1
typedef struct tagVS_FIXEDFILEINFO
{
    DWORD   dwSignature;            /* e.g. 0xfeef04bd */
    DWORD   dwStrucVersion;         /* e.g. 0x00000042 = "0.42" */
    DWORD   dwFileVersionMS;        /* e.g. 0x00030075 = "3.75" */
    DWORD   dwFileVersionLS;        /* e.g. 0x00000031 = "0.31" */
    DWORD   dwProductVersionMS;     /* e.g. 0x00030010 = "3.10" */
    DWORD   dwProductVersionLS;     /* e.g. 0x00000031 = "0.31" */
    DWORD   dwFileFlagsMask;        /* = 0x3F for version "0.42" */
    DWORD   dwFileFlags;            /* e.g. VFF_DEBUG | VFF_PRERELEASE */
    DWORD   dwFileOS;               /* e.g. VOS_DOS_WINDOWS16 */
    DWORD   dwFileType;             /* e.g. VFT_DRIVER */
    DWORD   dwFileSubtype;          /* e.g. VFT2_DRV_KEYBOARD */
    DWORD   dwFileDateMS;           /* e.g. 0 */
    DWORD   dwFileDateLS;           /* e.g. 0 */
} VS_FIXEDFILEINFO;

/* ----- VS_VERSION.dwFileFlags ----- */
#define	VS_FF_DEBUG		0x00000001L
#define	VS_FF_PRERELEASE	0x00000002L
#define	VS_FF_PATCHED		0x00000004L
#define	VS_FF_PRIVATEBUILD	0x00000008L
#define	VS_FF_INFOINFERRED	0x00000010L
#define	VS_FF_SPECIALBUILD	0x00000020L

/* ----- VS_VERSION.dwFileOS ----- */
#define	VOS_UNKNOWN		0x00000000L
#define	VOS_DOS			0x00010000L
#define	VOS_OS216		0x00020000L
#define	VOS_OS232		0x00030000L
#define	VOS_NT			0x00040000L

#define	VOS__BASE		0x00000000L
#define	VOS__WINDOWS16		0x00000001L
#define	VOS__PM16		0x00000002L
#define	VOS__PM32		0x00000003L
#define	VOS__WINDOWS32		0x00000004L

#define	VOS_DOS_WINDOWS16	0x00010001L
#define	VOS_DOS_WINDOWS32	0x00010004L
#define	VOS_OS216_PM16		0x00020002L
#define	VOS_OS232_PM32		0x00030003L
#define	VOS_NT_WINDOWS32	0x00040004L

/* ----- VS_VERSION.dwFileType ----- */
#define	VFT_UNKNOWN		0x00000000L
#define	VFT_APP			0x00000001L
#define	VFT_DLL			0x00000002L
#define	VFT_DRV			0x00000003L
#define	VFT_FONT		0x00000004L
#define	VFT_VXD			0x00000005L
#define	VFT_STATIC_LIB		0x00000007L

/* ----- VS_VERSION.dwFileSubtype for VFT_WINDOWS_DRV ----- */
#define	VFT2_UNKNOWN		0x00000000L
#define VFT2_DRV_PRINTER	0x00000001L
#define	VFT2_DRV_KEYBOARD	0x00000002L
#define	VFT2_DRV_LANGUAGE	0x00000003L
#define	VFT2_DRV_DISPLAY	0x00000004L
#define	VFT2_DRV_MOUSE		0x00000005L
#define	VFT2_DRV_NETWORK	0x00000006L
#define	VFT2_DRV_SYSTEM		0x00000007L
#define	VFT2_DRV_INSTALLABLE	0x00000008L
#define	VFT2_DRV_SOUND		0x00000009L
#define	VFT2_DRV_COMM		0x0000000AL

/* ----- VS_VERSION.dwFileSubtype for VFT_WINDOWS_FONT ----- */
#define VFT2_FONT_RASTER	0x00000001L
#define	VFT2_FONT_VECTOR	0x00000002L
#define	VFT2_FONT_TRUETYPE	0x00000003L
#else
#include "ver.h"	/* from 3.1 SDK */
#endif

#define LOWORD(l)           ((WORD)(l))
#define HIWORD(l)           ((WORD)((DWORD)(l) >> 16))

void do_version(DWORD offset, DWORD length, WORD id)
{
	BYTE buf[1024];	// workaround Borland problem with %*s
	BYTE *vers = (BYTE *) allocate(1, length);
	INFOBLOCK *infoblock;
	VS_FIXEDFILEINFO *info;
	int ofs;
	
	if (no_vers)
		return;
	
	exe_read(f, vers, offset, length);
	printf("VERSIONINFO %u (%02Xh)\n", id, id);
	printf("    %s\n", &vers[4]);	// szKey
	ofs = sizeof(INFOBLOCK) + strlen(&vers[4]);
	while (ofs % 4) ofs++;
	info = (VS_FIXEDFILEINFO *) &vers[ofs];
	infoblock = (INFOBLOCK *) vers;
	ofs += infoblock->cbValue;
	if (info->dwSignature != 0xfeef04bd)
		printf("    ERROR!\n");
	else
	{
		printf("    File Version %u.%u.%lx\n",
			HIWORD(info->dwFileVersionMS),
			LOWORD(info->dwFileVersionMS),
			info->dwFileVersionLS);
		printf("    Product Version %u.%u.%lx\n",
			HIWORD(info->dwProductVersionMS),
			LOWORD(info->dwProductVersionMS),
			info->dwFileVersionLS);
		if (info->dwFileFlags & info->dwFileFlagsMask)
		{
			printf("File Type: ");
			switch(info->dwFileFlags & info->dwFileFlagsMask)
			{
				case VS_FF_DEBUG: printf("Debug"); break;
				case VS_FF_PRERELEASE: printf("Pre-Release"); break;
				case VS_FF_PATCHED: printf("Patched"); break;
				case VS_FF_PRIVATEBUILD: printf("Private Build"); break;
				case VS_FF_INFOINFERRED: printf("(Info Inferred)"); break;
				case VS_FF_SPECIALBUILD: printf("Special Build"); break;
				default: printf("??? (%lxh)", 
					info->dwFileFlags & info->dwFileFlagsMask); break;
			}
			printf("\n");
		}
		printf("    Operating System: ");
		switch (info->dwFileOS)
		{
			case VOS_UNKNOWN: printf("Unknown"); break;
			case VOS_DOS: printf("DOS"); break;
			case VOS_OS216: printf("OS/2 16-bit"); break;
			case VOS_OS232: printf("OS/2 32-bit"); break;
			case VOS_NT: printf("NT"); break;
			case VOS_DOS_WINDOWS16: printf("Windows 16-bit"); break;
			case VOS_DOS_WINDOWS32: printf("Windows 32-bit"); break;
			case VOS_OS216_PM16: printf("OS/2 PM 16-bit"); break;
			case VOS_OS232_PM32: printf("OS/2 PM 32-bit"); break;
			case VOS_NT_WINDOWS32: printf("NT Windows 32-bit"); break;
			default: printf("??? (%lxh)", info->dwFileOS); break;
		}
		printf("\n\tType: ");
		switch (info->dwFileType)
		{
			case VFT_UNKNOWN: printf("Unknown"); break;
			case VFT_APP: printf("Application"); break;
			case VFT_DLL: printf("Dynamic Link Library (DLL)"); break;
			case VFT_DRV: printf("Device Driver (DRV): ");
				switch (info->dwFileSubtype)
				{
					case VFT2_UNKNOWN: printf("Unknown"); break;
					case VFT2_DRV_PRINTER: printf("Printer"); break;
					case VFT2_DRV_KEYBOARD: printf("Keyboard"); break;
					case VFT2_DRV_LANGUAGE: printf("Language"); break;
					case VFT2_DRV_DISPLAY: printf("Display"); break;
					case VFT2_DRV_MOUSE: printf("Mouse"); break;
					case VFT2_DRV_NETWORK: printf("Network"); break;
					case VFT2_DRV_SYSTEM: printf("System"); break;
					case VFT2_DRV_INSTALLABLE: printf("Installable"); break;
					case VFT2_DRV_SOUND: printf("Sound"); break;
					case VFT2_DRV_COMM: printf("Comm"); break;
					default: printf("??? (%lxh)", info->dwFileSubtype); break;
				}
				break;
			case VFT_FONT: printf("Font: ");
				switch (info->dwFileSubtype)
				{
					case VFT2_FONT_RASTER: printf("Raster"); break;
					case VFT2_FONT_VECTOR: printf("Vector"); break;
					case VFT2_FONT_TRUETYPE: printf("TrueType"); break;
					default: printf("??? (%lxh)", info->dwFileSubtype);
				}
				break;
			case VFT_VXD: printf("Virtual Device Driver (VxD)"); break;
			case VFT_STATIC_LIB: printf("Static Link Library"); break;
			default: printf("??? (%lxh)", info->dwFileType); break;
		}
		printf("\n");
		// deal with other blocks later
	}
	free(vers);
}

void dump_generic(DWORD offset, DWORD length)
{
	BYTE *res = (BYTE *) allocate(1, length);
	exe_read(f, res, offset, length);
	dump(res, length, "%08lx", 0);
	free(res);
}

/**********************************************************************/

/* read at offset from NE */
int ne_read(int f, void *p, DWORD ne_offset, WORD bytes)
{
	int retval;
    if (lseek(f, (long) start_newexe + ne_offset, SEEK_SET) == -1)
		fail_perror();
	if ((retval = read(f, p, bytes)) == -1)
		fail_perror();
	return retval;
}

/* read at offset from beginning of file */
int exe_read(int f, void *p, DWORD exe_offset, WORD bytes)
{
	int retval;
    if (lseek(f, exe_offset, SEEK_SET) == -1)
		fail_perror();
	if ((retval = read(f, p, bytes)) == -1)
		fail_perror();
	return retval;
}

void *allocate(size_t x, size_t y)
{
	void *p;
	if (! x) x = 1;
	if (! y) y = 1;
	if (((long) x * (long) y) > 0xFFFFUL)
	{
		printf("\n\nx=%02X y=%02X\n", x, y);
		fail("request > 64k");
	}
	if ((p = calloc(x, y)) == 0)
	{
		printf("\n\nx=%02X y=%02X\n", x, y);
		fail("insufficient memory");
	}
	return p;
}

#define WIDTH		16

void dump(BYTE far *fp, WORD bytes, 
	char *mask, DWORD addr)
{
	BYTE far *p;
	WORD i, j, c;
	
    for (i=0; i<bytes; i += WIDTH)
    {
		c = ((bytes-i) > WIDTH) ? WIDTH : bytes-i;
		printf(mask, addr+i);
		putchar(' '); putchar('|'); putchar(' ');
        for (j=c, p=fp+i; j--; p++)
            printf("%02X ", *p);
		for (j=WIDTH-c; j--; )	// pad out on last line
			printf("   ");
		putchar('|'); putchar(' ');
        for (j=c, p=fp+i; j--; p++)
			putchar( isprint(*p) ? *p : '.' );
		putchar('\n');
    }
}

/**********************************************************************/

/*
	TO DO:
		figure out what wrong with CURSORDIRENTRY in doc
		use (wid*ht / 8) dimensions to display actual icon, curs later
*/

typedef struct {
	WORD wReserved, wType, cwCount;
	} CURSICONDIR;
	
typedef struct {
	BYTE bWidth, bHeight, bColorCount, bReserved;
	WORD wPlanes, wBitCount;
	DWORD dwBytesInRes;
	WORD pNameOrdinal;
	} ICONDIRENTRY;
	
typedef struct {
	WORD bWidth, bHeight;
	WORD wPlanes, wBitCount;
	DWORD dwBytesInRes;
	WORD pNameOrdinal;
	} CURSDIRENTRY;

void do_icondir(DWORD offset, DWORD length, WORD id)
{
	CURSICONDIR *cidir = (CURSICONDIR *) allocate(1, length);
	ICONDIRENTRY *icon = 
		(ICONDIRENTRY *) (((BYTE *) cidir) + sizeof(CURSICONDIR));
	int i;
	exe_read(f, cidir, offset, length);
	if (id == 0xFFFF)
		printf("ICON DIRECTORY \"%s\" -- %u icon",
			idname, cidir->cwCount);
	else
		printf("ICON DIRECTORY %u (%02Xh) -- %u icon", 
			id, id, cidir->cwCount);
	if (cidir->cwCount > 1) putchar('s');
	putchar('\n');
	for (i=0; i<cidir->cwCount; i++)
	{
		printf("    %d    %ux%u, %u color, %u plane, %u bit (%lu bytes)\n",
			icon->pNameOrdinal,
			icon->bWidth, icon->bHeight, 
			icon->bColorCount, 
			icon->wPlanes, icon->wBitCount, 
			icon->dwBytesInRes);
		icon++;
	}
	free(cidir);
}

void do_cursdir(DWORD offset, DWORD length, WORD id)
{
	CURSICONDIR *cidir = (CURSICONDIR *) allocate(1, length);
	CURSDIRENTRY *curs = 
		(CURSDIRENTRY *) (((BYTE *) cidir) + sizeof(CURSICONDIR));
	int i;
	exe_read(f, cidir, offset, length);
	if (id == 0xFFFF)
		printf("CURSOR DIRECTORY \"%s\" -- %u cursor", 
			idname, cidir->cwCount);
	else
		printf("CURSOR DIRECTORY %u (%02Xh) -- %u cursor", 
			id, id, cidir->cwCount);
	if (cidir->cwCount > 1) putchar('s');
	putchar('\n');
	for (i=0; i<cidir->cwCount; i++)
	{
		printf("    %d    %ux%u, %u plane, %u bit (%lu bytes)\n",
			curs->pNameOrdinal, 
			curs->bWidth, curs->bHeight, 
			curs->wPlanes, curs->wBitCount, 
			curs->dwBytesInRes);
		curs++;
	}
	free(cidir);
}

/**********************************************************************/

typedef struct {
	WORD fontOrdinal, dfVersion;
	DWORD dfSize;
	char copyright[60];
	WORD dfType, dfPoints, dfVertRes, dfHorizRes, dfAscent;
	WORD dfInternalLeading, dfExternalLeading;
	BYTE dfItalic, dfUnderline, dfStrikeOut;
	WORD dfWeight;
	BYTE dfCharSet;
	WORD dfPixWidth, dfPixHeight;
	BYTE dfPitchAndFamily;
	WORD dfAvgWidth, dfMaxWidth;
	BYTE dfFirstChar, dfLastChar, dfDefaultChar, dfBreakChar;
	WORD dfWidthBytes;
	DWORD dfDevice, dfFace, dfReserved;
	// char szDeviceName[];
	// char szFaceName[];	
	} FONTDIRENTRY;
	
static char *wtname[] = {
	"DontCare", "Thin", "Extralight", "Light", "Regular", 
	"Medium", "Demibold", "Bold", "Extrabold", "Black"
	} ;

void do_fontdir(DWORD offset, DWORD length, WORD id)
{
	BYTE *fontdir = (BYTE *) allocate(1, length);
	BYTE *devname, *facename, *orig;
	FONTDIRENTRY *fentry;
	unsigned num_font;
	int i;
	exe_read(f, &num_font, offset, 2);
	exe_read(f, fontdir, offset+2, length);
	printf("FONT DIRECTORY %u (%02Xh) -- %u font",
		id, id, num_font);
	if (num_font > 1) putchar('s');
	putchar('\n');
	orig = fontdir;
	for (i=0; i<num_font; i++)
	{
		fentry = (FONTDIRENTRY *) fontdir;
		fontdir += sizeof(FONTDIRENTRY);
		devname = fontdir;
		fontdir += (strlen(devname) + 1);
		facename = fontdir;
		printf("   %u   %u pt. ", 
			fentry->fontOrdinal, fentry->dfPoints);
		if (*facename)
		{
			printf("%s ", facename);
			if (fentry->dfWeight != 400)
				printf("%s ", wtname[fentry->dfWeight / 100]);
			if (fentry->dfItalic)		printf("Italic ");
			if (fentry->dfUnderline)	printf("Underline ");
			if (fentry->dfStrikeOut)	printf("StrikeOut ");
		}
		printf("\n");
		if (*devname)
			printf("        Device \"%s\"\n", devname);
		if (* fentry->copyright)
			printf("        \"%s\"\n", fentry->copyright);
		fontdir += (strlen(facename) + 1);
	}
	free(orig);
}

/**********************************************************************/

typedef struct {
	DWORD biSize;
	LONG biWidth, biHeight;
	WORD biPlanes, biBitCount;
	DWORD biCompression, biSizeImage;
	LONG biXPelsPerMeter, biYPelsPerMeter;
	DWORD biClrUsed, biClrImportant;
	} BITMAPINFOHEADER;
	
typedef struct {
	DWORD bcSize;
	WORD bcWidth, bcHeight, bcPlanes, bcBitCount;
	} BITMAPCOREHEADER;
	
char *bicomp_str[] = { "RGB", "RLE8", "RLE4" } ;

void do_bitmap(DWORD offset, DWORD length, WORD id)
{
	BITMAPINFOHEADER *bi = 
		(BITMAPINFOHEADER *) allocate(1, sizeof(BITMAPINFOHEADER));
	exe_read(f, bi, offset, sizeof(BITMAPINFOHEADER));
	if (id == 0xFFFF)
		printf("BITMAP \"%s\"", idname);
	else
		printf("BITMAP %d (%02Xh)", id, id);
	if (verbose) printf(" (offset=%08lXh, length=%lu)", 
		offset, length);
	printf("\n");
	if (bi->biSize == sizeof(BITMAPCOREHEADER))
	{
		BITMAPCOREHEADER *bc = (BITMAPCOREHEADER *) bi;
		printf("    %u x %u, %u plane, %u bit/pixel\n",
			bc->bcWidth, bc->bcHeight, bc->bcPlanes, bc->bcBitCount);
	}
	else if (bi->biSize == sizeof(BITMAPINFOHEADER))
	{
		printf("    %ld x %ld, %u plane, %u bit/pixel, %s\n",
			bi->biWidth, bi->biHeight, bi->biPlanes, bi->biBitCount,
			(bi->biCompression < 3) ? bicomp_str[bi->biCompression] : "??");
	}
	free(bi);
}

// primarily to check for Visual Basic .FRM, etc.
void do_userdata(DWORD offset, DWORD length, WORD id)	// > 64k??
{
	BYTE *data = (BYTE *) allocate(1, length);
	exe_read(f, data, offset, length);
	printf("USERDATA %u (%02Xh)\n", id, id);
	if (memcmp(data, "\x23\xcc\x24\x00", 4) == 0)
		do_vb(data, length);
	free(data);
}

void do_nametab(DWORD offset, DWORD length, WORD id)
{
	BYTE *data = (BYTE *) allocate(1, length);
	char *orig = data;
	unsigned ofs = 0;
	unsigned restype;
	exe_read(f, data, offset, length);
	printf("NAMETABLE %u (%02Xh)\n", id, id);
	while (ofs < length)
	{
		if (! *((WORD *) &data[0]))	// length zero marks end of table??
			break;
		printf("    ");
		// printf("%04X ", *((WORD *) &data[0]));	// length: don't show
		restype = *((WORD *) &data[2]);
		if (restype >= 1 && restype <= 16)
			printf("%s ", restypename[restype]);
		else
			printf("#%u ", restype);
		printf("#%u ", *((WORD *) &data[4]) & ~0x8000);	// strip hi bit
		ofs += 6; data += 6;
		if (*data)
		{
			printf("\"%s\" ", data);
			ofs += (strlen(data) + 1); data += (strlen(data) + 1);
		}
		else
		{
			ofs++; data++;
		}
		if (*data)
		{
			printf("\"%s\" ", data);
			ofs += (strlen(data) + 1); data += (strlen(data) + 1);
		}
		else
		{
			ofs++; data++;
		}
		putchar('\n');
	}
	free(orig);
}

// see "Dateien entschlusselt", WinDOS, March 1992
// should probably check if modref[0]=="VBRUN100"
void do_vb(BYTE *data, DWORD length)
{
#if 1
	printf("Visual Basic FRM file\n");
#else
	WORD offset;
	int i;
	
	printf("Next free object: %u\n", *((BYTE *) &data[4]));
	printf("Bytes: %lu\n", *((DWORD *) &data[5]));
	
	for (offset = 9; offset < length; )
	{
		// check for FF -- global
		WORD ofs = offset;
		WORD slen;
		signed long len = *((signed long *) &data[offset]);
		offset += 4;
		if (len < 0)
		{
			len = -len;
			offset += 2;
		}
		printf("\nObject #%u, length %lu\n", *((BYTE *) &data[offset]), len);
		offset++;
		slen = *((BYTE *) &data[offset]);
		offset++;
		printf("Name: <%.*s>\n", slen, &data[offset]);
		offset += slen;
		printf("Type: %u\n", *((BYTE *) &data[offset]));
		offset = ofs + len;
	}
#endif	
}

// code borrowed from Rainer G. Haselier and Klaus Fahnenstich

static BYTE oem_tab[] = {
	'?', '?', '?', '?', '?', '?', '?', '?', 
	'?', '?', '?', '?', '?', '?', '?', '?', 
	'?', '?', '?', '?', '?', '?', '?', '?', 
	'?', '?', '?', '?', '?', '?', '?', '?', 
	32, 173, 155, 156, '?', 157, 124, 21,
	'?', '?', 166, 174, 170, '?', '?', '?',
	148, 241, 253, '?', 39, 230, 20, 250,
	'?', '?', 167, 175, 172, 171, '?', 168,
	'?', '?', '?', '?', 142, 143, 146, 128,
	144, '?', '?', '?', '?', '?', '?', '?',
	'?', 156, '?', '?', '?', '?', 153, '?',
	237, '?', '?', '?', 154, '?', '?', 225,
	133, 160, 131, '?', 132, 134, 145, 135,
	138, 130, 136, 137, 141, 161, 140, 139,
	'?', 164, 149, 162, 147, '?', 148, '?',
	237, 151, 163, 150, 129, '?', '?', 152, 
	} ;

BYTE *AnsiToOemBuf(BYTE *buf)
{
	BYTE *s = buf;
	BYTE c;
	if (no_convert)
	{
		static char *buf2 = (char *) malloc(2048);
		char *s2 = buf2;
		if (! buf2) 
			fail("insufficient memory");
		while ((c = *s) != 0)		// conversion for Japan
		{
			if (c == 0x1E)
			{
				*s2++ = '\\'; *s2++ = '0'; *s2++ = '3'; *s2 = '6'; 
			}
			else if (c == 0x1F)
			{
				*s2++ = '\\'; *s2++ = '0'; *s2++ = '3'; *s2 = '7'; 
			}
			else *s2 = c;
			s2++; s++;
		}
		*s2 = '\0';
		return buf2;			// unchanged except for 1E -> \036 etc.
	}
	else
	{
		while ((c = *s) != 0)
		{
			if (c >= 128)			// no translation for lower 127 chars
				*s = oem_tab[c & 0x7F];
			s++;
		}
		return buf;
	}
}

