/* 
TEST2F16.C -- Show that Windows doesn't reflect INT 2Fh func 16h to V86 mode
Andrew Schulman, 1994
bcc -2 -P- test2f16.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

typedef unsigned short WORD;
typedef unsigned long DWORD;

#pragma pack(1)

typedef struct {
#ifdef __TURBOC__
    WORD bp,di,si,ds,es,dx,cx,bx,ax;
#else
    WORD es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    WORD ip,cs,flags;
    } REG_PARAMS;

void interrupt far int2f(REG_PARAMS r);
void (interrupt far *old_2F)();

static DWORD int2f16_calls[0x100] = {0};
static DWORD total_int2f16_calls = 0;
static DWORD int2f_calls = 0;
static DWORD int2f1200_calls = 0;

int win3e(void) // is Windows Enhanced mode running?
{
    int maj = 0;
    _asm mov ax, 1600h
    _asm int 2fh
    _asm mov byte ptr maj, al
    return (maj && (maj != 0x80));
}

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
	DWORD i;
	DWORD num_iter = (argc < 2) ? 100 : atol(argv[1]);
	
    /* hook INT INT 2Fh */
    old_2F = _dos_getvect(0x2F);
    _dos_setvect(0x2f, int2f);
	
	/* issue lots of 2F/168x calls */
	for (i=0; i<num_iter; i++)
	{
		_asm mov ax, 1200h		// sanity check
		_asm int 2fh
		_asm mov ax, 1680h
		_asm int 2fh
		_asm mov ax, 1681h
		_asm int 2fh
		_asm mov ax, 1682h
		_asm int 2fh
		_asm mov ax, 1683h
		_asm int 2fh
	}

    /* unhook INT 2Fh */
    _dos_setvect(0x2F, old_2F);
	
	/* sanity check */
	if (int2f1200_calls < num_iter)
		fail("Something wrong!");
	
	/* display results */
	printf("TEST2F16 issued %lu calls to 2F/1200\n", num_iter);
	printf("TEST2F16 issued %lu calls each to 2F/1680-2F/1683\n", num_iter);
	printf("TEST2F16 detected %lu INT 2Fh calls\n", int2f_calls);
	if (total_int2f16_calls)
	{
		printf("TEST2F16 detected the following 2F/16 calls:\n");
		for (i=0; i<0x100; i++)
			if (int2f16_calls[i])
				printf("%02lX:%lu\t", i, int2f16_calls[i]);
		printf("\n");
	}
	else
		printf("TEST2F16 detected 0 calls to INT 2Fh function 16h!\n");
	
	if (total_int2f16_calls < num_iter)
		printf("\nINT 2Fh function 16h isn't reflected to V86 mode!\n");
	
    return 0;
}

void interrupt far int2f(REG_PARAMS r)
{
	int2f_calls++;
	if (r.ax == 0x1200)
		int2f1200_calls++;
	if ((r.ax >> 8) == 0x16)
	{
		total_int2f16_calls++;
		int2f16_calls[r.ax & 0xff]++;
	}
	_chain_intr(old_2F);
}
