/* 
USEDPMI.C -- very simple DPMI client
bcc usedpmi.c
Andrew Schulman, April 1994

This program *MUST* be compiled in small model because, once in
protected mode, it will call printf from the real-mode C run-time
library.  In large model, the C run-time would use far pointers,
but USEDPMI's segment registers are going to change out from under
it when it switches to protected mode.
*/

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

void _dos_exit(int retval)
{
    _asm mov ah, 04ch
    _asm mov al, byte ptr retval
    _asm int 21h
}

void fail(const char *s) { puts(s); exit(1); }

// Call the DPMI "Obtain Real to Protected Mode Switch Entry
// Point" function (INT 2Fh AX=1687h)
int dpmi_init(void)
{
    void (far *switch_proc)();
    unsigned hostdata_seg, hostdata_para, dpmi_flags;
    _asm push si
    _asm push di
    _asm mov ax, 1687h           /* test for DPMI presence */
    _asm int 2Fh
    _asm and ax, ax
    _asm jz got_dpmi             /* if (AX == 0) DPMI is present */
    _asm jmp no_dpmi
got_dpmi:        
    _asm mov dpmi_flags, bx
    _asm mov hostdata_para, si   /* paras for DPMI host private data */
    _asm mov word ptr switch_proc, di
    _asm mov word ptr switch_proc+2, es
    _asm pop di
    _asm pop si

    if (_dos_allocmem(hostdata_para, &hostdata_seg) != 0)
        fail("can't allocate memory");
    
    dpmi_flags &= ~1;   /* this is a 16-bit protected-mode program */
    
    /* enter protected mode */
    _asm mov ax, hostdata_seg
    _asm mov es, ax
    _asm mov ax, dpmi_flags
    (*switch_proc)();
    /* I don't think we're in Kansas anymore, Toto! */
    return 1;
no_dpmi:
	_asm pop di
	_asm pop si
    return 0;
}

void print_regs(void)
{
	unsigned short ds_reg, cs_reg;
	_asm mov cs_reg, cs
	_asm mov ds_reg, ds
	printf("CS=%04Xh DS=%04Xh\n", cs_reg, ds_reg);
}

main()
{
	print_regs();
    if (! dpmi_init())      // switch into protected mode, via DPMI
        fail("This program requires DPMI");   
	else
	{
		print_regs();
        printf("Greetings from protected mode!\n");
	}
	
    // to exit from protected mode, must use 21/4C!
	// and since bypassing C exit cleanup, must do flushall ourselves
	flushall();	
    _dos_exit(0);
}

