/* 
VMWALK.C -- Enumerates VM handles, using generic VxD
Andrew Schulman, February 1993
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

CHANGED FROM VERSION IN UNDOCDOS 2E:  VMWALK GIVES APP PTR TO
ACTUAL VM CB, RATHER THAN COPY!

Changed again, adding thread walk for Chicago:
Get_Cur_Thread_Handle -> EDI
Get_Sys_Thread_Handle -> EDI
Get_Initial_Thread_Handle (EBX = VM) -> EDI					VM[0]
Get_VM_Handle_For_Thread (EDI = thread) -> EBX				thread[14h]
Get_Next_Thread_Handle (EDI = thread) -> EDI				thread[30h]

DPMI version:
bcc -DTESTING -DDPMI_APP -DCHICAGO -2 vmwalk.c vxdcalls.c dpmi_app.lib 

bcc -DDPMI_APP -2 -c vmwalk.c vxdcalls.c
tlib dpmi_app-+vmwalk-+vxdcalls

Windows version:
bcc -W -DTESTING -DWINDOWS -2 vmwalk.c vxdcalls.c prot.c 
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#ifdef WINDOWS
#include "windows.h"
#endif
#include "prot.h"
#include "vxdcalls.h"
#include "vmwalk.h"

#define CHICAGO

// changed from version in UndocDOS 2e:  give app ptr to actual VM CB,
// not a copy.
int vmwalk(WALKFUNC walkfunc)
{
    DWORD sys_vm = GetSysVMHandle();
    VM_CB far *vm_cb;
    DWORD vm;
    int num_vm;
    for (vm=sys_vm, num_vm=0;; num_vm++)
    {
        set_base(0);  // in case app fiddled with it
        if (! (vm_cb = (VM_CB far *) map_linear(vm, sizeof(VM_CB)+0x2000)))
			return 0;
        if (! (*walkfunc)(vm, vm_cb))  // give app actual VM_CB
		{
			free_mapped_linear(vm_cb);
			goto done;
		}
        free_mapped_linear(vm_cb);
        if ((vm = GetNextVMHandle(vm)) == sys_vm)
            break;  // circular list
    }
done:
    return num_vm;
}

#define Get_Next_VM_Handle	0x01003BL
#define Get_Cur_VM_Handle	0x010001L
#define Get_Sys_VM_Handle	0x010003L

DWORD GetSysVMHandle(void)
{
    VxDParams p;
    p.CallNum = Get_Sys_VM_Handle;
    return (VxDCall(&p)) ? p.OutEBX : 0;
}

DWORD GetCurVMHandle(void)
{
    VxDParams p;
    p.CallNum = Get_Cur_VM_Handle;
    return (VxDCall(&p)) ? p.OutEBX : 0;
}

DWORD GetNextVMHandle(DWORD vm)
{
    VxDParams p;
    p.CallNum = Get_Next_VM_Handle;
    p.InEBX = vm;
    return (VxDCall(&p)) ? p.OutEBX : 0;
}

#ifdef CHICAGO
#define Get_Cur_Thread_Handle		0x010108L
#define Get_Sys_Thread_Handle		0x01010AL
#define Get_Initial_Thread_Handle	0x01010DL
#define Get_VM_Handle_For_Thread	0x010111L
#define Get_Next_Thread_Handle		0x010113L

DWORD GetSysThreadHandle(void)
{
    VxDParams p;
    p.CallNum = Get_Sys_Thread_Handle;
    return (VxDCall(&p)) ? p.OutEDI : 0;
}

DWORD GetCurThreadHandle(void)
{
    VxDParams p;
    p.CallNum = Get_Cur_Thread_Handle;
    return (VxDCall(&p)) ? p.OutEDI : 0;
}

DWORD GetNextThreadHandle(DWORD thread)
{
    VxDParams p;
    p.CallNum = Get_Next_Thread_Handle;
    p.InEDI = thread;
    return (VxDCall(&p)) ? p.OutEDI : 0;
}

DWORD GetInitialThreadHandle(DWORD vm)
{
    VxDParams p;
    p.CallNum = Get_Initial_Thread_Handle;
    p.InEBX = vm;
    return (VxDCall(&p)) ? p.OutEDI : 0;
}

DWORD GetVMHandleForThread(DWORD thread)
{
    VxDParams p;
    p.CallNum = Get_VM_Handle_For_Thread;
    p.InEDI = thread;
    return (VxDCall(&p)) ? p.OutEBX : 0;
}
#endif

// The next three functions do nothing but add vm_cb->CB_High_Linear
// onto a linear address, but they make it a lot easier to work with
// multiple VMs.

DWORD get_vm_base(DWORD vm)
{
    VM_CB far *vm_cb;
    DWORD base;
    if (vm == 0) return 0;  // stay in current VM
    vm_cb = (VM_CB far *) map_linear(vm, sizeof(VM_CB));
    base = vm_cb->CB_High_Linear;
    free_mapped_linear(vm_cb);
    return base;
}

void set_vm(DWORD vm)   // sets base address for map_linear
{
    set_base(get_vm_base(vm));
}

void far *map_real_vm(DWORD vm, void far *fp, WORD bytes)
{
    return map_linear(get_vm_base(vm) + MK_LIN(fp), bytes);
}

void far *map_linear_vm(DWORD vm, DWORD lin_addr, WORD bytes)
{
    return map_linear(get_vm_base(vm) + lin_addr, bytes);
}

#ifdef TESTING
#include <stdio.h>

WORD get_thread_id(DWORD thread)
{
	WORD id;
	BYTE far *thcb = (BYTE far *) map_linear(thread, 0x20);
	if (! thcb) return -1;
	id = *((WORD far *) &thcb[0x18]);
	free_mapped_linear(thcb);
	return id;
}

void print_thread(DWORD thread)
{
	static DWORD cur_thread = 0;
	static DWORD sys_thread = 0;
	if (! (cur_thread && sys_thread))
	{
		cur_thread = GetCurThreadHandle();
		sys_thread = GetSysThreadHandle();
	}
	printf("   #%-4d -- Thread %08lX", 
		get_thread_id(thread), thread);
	if (thread == cur_thread) printf(" (Current)");
	if (thread == sys_thread) printf(" (System)");
	printf("\n");
}

WORD get_win_vers(void)
{
	_asm mov ax, 1600h
	_asm int 2fh
	_asm xor ah, ah
}

BOOL walkfunc(DWORD vm, VM_CB far *vm_cb)
{
	DWORD thread;
    printf("%08lX   VM #%lu  lin=%08lXh  crs=%08lXh\n",
		vm,
        vm_cb->CB_VMID,
        vm_cb->CB_High_Linear,
        vm_cb->CB_Client_Pointer);
#ifdef CHICAGO	
	if (get_win_vers() >= 4)
	if ((thread = GetInitialThreadHandle(vm)) != 0)
	{
		DWORD first_thread = thread;
		print_thread(thread);
		while ((thread = GetNextThreadHandle(thread)) != first_thread)
			if (GetVMHandleForThread(thread) == vm)		// not working????
				print_thread(thread);
	}
#endif		
    return 1;
}

void fail(const char *s, ...) { printf("%s\n", s); exit(1); }
void panic(const char *s) { fail(s); }

#ifdef DPMI_APP
#pragma argsused
int real_main(int argc, char *argv[])   { return 0; }
#pragma argsused
int pmode_main(int argc, char *argv[])  { return vmwalk(walkfunc); }
#else
int main()								{ return vmwalk(walkfunc); }
#endif
#endif
