/* 
W3MAP.C -- WIN386 mapper

bcc -I..\include -L..\lib w3map.c unauthwr.lib 

note: embedded LE have no nonresident names table, so no description
string or DDB name

CHANGES:
	1.32 (5/11/92):  
		added new 3.1 functions to vxdcalls.h for VMM, VDD, VPICD, 
			V86MMGR, VMCPD (new table)
			
	1.33 (7/27/92):
		fixed le2.c so that Control_Proc recognized even when addr==0
			
	1.34 (8/12/92):
		changed Service_Table from dd, r ?? to dd, osn, r ?? (le2.c)
			
	1.35 (9/11/92):
		print out Start CS:EIP too
		Added -VXD switch to just display individual VxD (or multiple
		VxDs whose names all start with same prefix)
			
	1.36 (11/3/92):
		changed le2.c more to check for garbage ptrs -- check that
		file offset < filelength(f)
			
	1.37 (1/16/93):
		printing out init order in le2.c
			
	2.02 (5/4/94):
		changed DDB printout in le2.c, MAYBE_API for 0
		Added new VMM services, new VxDs for WfW 3.11
			
	2.10 (5/9/94):
		Doing real-mode init
		More new VxDs
			
	2.20 (8/1/94):
		improved multiple object support, put ClientXXX into def file
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <share.h>
#include <dos.h>
#include "filelib.h"

/* lifted from MSoft newexe.h */
#define ENEWHDR     0x003CL         /* offset of new EXE header */
#define EMAGIC      0x5A4D          /* old EXE magic id:  'MZ'  */
#define NEMAGIC     0x454E          /* new EXE magic id:  'NE'  */
#define LEMAGIC     0x454C          /* linear executable */
#define LXMAGIC     0x584C          /* IBM OS/2 2.0 linear executable */
#define PEMAGIC     0x4550          /* NT portable executable */
#define W3MAGIC     0x3357          /* WIN386.EXE signature */
#define SZMAGIC		0x5A53			/* SZ -- compressed file */

#pragma pack(1)

typedef struct {
    unsigned char sign[8];
    unsigned long offset;
    unsigned char pad[4];
    } W3REC;

void fail(char *s) { fputs("w3map fail: ", stdout); puts(s); exit(1); }

void fail_perror(void)	{ perror("w3map fail"); exit(1); }

void usage()
{
	puts("W3MAP displays information about the Virtual Device Drivers");
	puts("(VxDs) in a Windows 3.x Enhanced mode WIN386 (W3) file.");
	puts("");
	puts("To display a list of all VxDs in the WIN386 file:");
	puts("         W3MAP win386_file");
	puts("example: w3map \\windows\\system\\win386.exe");
	puts("");
	puts("To display information on an individual VxD in the WIN386 file:");
	puts("         W3MAP -VXD vxd_name win386_file");
	puts("example: w3map -vxd dosmgr \\windows\\system\\win386.exe");
	puts("");
	puts("To display detailed information on each VxD in the WIN386 file:");
	puts("         W3MAP -VERBOSE win386_file");
	puts("example: w3map -verbose \\windows\\system\\win386.exe");
    exit(1);
}
    
void le2(unsigned long start_le);
void w3(char *name);

unsigned long start_newexe = 0;

static unsigned long align;
static int f;
static int paranoid = 0;
static int verbose = 0;
static char *one_vxd = 0;

main(int argc, char *argv[])
{
	char *arg;
    int i;
    
    fputs("WIN386 (W3) Dumper version 2.20\n", stderr);
	fputs("from \"Unauthorized Windows\" by Andrew Schulman "
		  "(IDG Books, 1994)\n", stderr);
    fputs("Copyright (c) 1992-94 Andrew Schulman.  All rights reserved.\n\n", 
        stderr);
    
    if (argc < 2)
        usage();
    
    for (i=1; i<argc; i++)
        if ((argv[i][0] == '-') || (argv[i][0] == '/'))
        {
            if (argv[i][1] == '?')
                usage();
			
			arg = strupr(&argv[i][1]);
			
            if (strncmp(arg, "VXD", 3) == 0)
            {
				char buf[128];
				one_vxd = strupr(argv[i+1]);
				if (strcmp(one_vxd, "VMM") == 0)	// hack!!
					one_vxd = "WIN386";
				else if (strcmp(one_vxd, "VMM32") == 0)
					one_vxd = "VMM";
                i ++;
            }
            else if (strcmp(arg, "PARANOID") == 0)  // UNDOC
                paranoid = 1;
            else if (strncmp(arg, "VERBOSE", 4) == 0)
                verbose = 1;
            else
                usage();
        }
        else
		{
			char buf[128];
			strcpy(buf, argv[i]);
			if (! strchr(buf, '.'))
				strcat(buf, ".exe");
			strupr(buf);
			if (strstr(buf, "VMM32.VXD"))	// hack!!
				if (one_vxd && (strcmp(one_vxd, "WIN386") == 0))
					strcpy(one_vxd, "VMM");
            w3(buf);
		}
    return 0;
}

void w3(char *name)
{
    W3REC rec;
    unsigned char le[2];
    unsigned long pos;
    unsigned short w;
    unsigned i;
    
	if ((f = sopen(name, O_RDONLY | O_BINARY, SH_DENYNO)) == -1)
		fail_perror();
	exe_read(f, &w, 0, 2);	// read first two bytes of file
    if (w != EMAGIC)
	{
		if (w == SZMAGIC)
			fail("compressed file (SZ) -- use EXPAND.EXE");
		else
			fail("not an executable");
	}
	exe_read(f, &w, 0x18, 2);
	if (w < 0x40)
		fail("not a W3 (WIN386) executable");
	exe_read(f, &start_newexe, ENEWHDR, 4);
	exe_read(f, &w, start_newexe, 2);
	if (w == LEMAGIC)
		fail("use LEDUMP to examine this VxD");
    if (w != W3MAGIC)
        fail("not a W3 (WIN386) executable");
	if (! one_vxd)
	{
		printf("W3      \t%08lx\n", start_newexe);
		if (verbose)
			putchar('\n');
	}
    pos = start_newexe + sizeof(W3REC);
    for (;;)
    {
        exe_read(f, &rec, pos, sizeof(W3REC));
        pos = tell(f);
        if (rec.sign[0] == 0)
            break;

		if (one_vxd)
		{
			if (strncmp(strupr(rec.sign), one_vxd, 
				strlen(one_vxd)) == 0)
			{
				printf("File offset %08lx\n", rec.offset);
				le2(rec.offset);
				putchar('\n');
			}
		}
		else
			printf("%-8s\t%08lx\n", rec.sign, rec.offset);
        
        if (verbose)
        {
            le2(rec.offset);
            putchar('\n');
        }
        else if (paranoid)
        {
            // ensure there is an 'LE' file at designated offset
            exe_read(f, &w, rec.offset, sizeof(w));
            if (w != LEMAGIC)
                fail("WIN386 file corrupt?");
        }
        
    }
    close(f);
}

/**********************************************************************/

#ifdef DO_NRESTAB
#undef DO_NRESTAB
#endif
#include "le2.c"

/**********************************************************************/

