/* 
WINBP.C -- Display Windows Breakpoints
from Schulman, Unauthorized Windows, 1994

bcc -DDPMI_APP -2 -I..\include -c prot.c dpmish.c ctrl_c.asm
tlib ..\lib\unauthw+prot+dpmish+ctrl_c
bcc -DDPMI_APP -2 -B -I..\include -L..\lib winbp.c unauthwd.lib 
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <dos.h>
#include "dpmish.h"
#include "prot.h"

typedef void far *FP;

FP GetVxDAPI(WORD vxd_id);
DWORD GetGDTSelectorBase(WORD seg);
DWORD GetGDTSelectorLimit(WORD seg);

#define PUT(s)				  { fputs(s, stderr); fputs("\n", stderr); }

void fail(const char *s, ...) { PUT(s); _dos_exit(1); }

static DWORD v86bp_lin, pmbp_base, bp_tab_lin;
static WORD pmbp_size, max_bps;
static int win_ver = 0;

#pragma pack(1)

typedef struct {	// V86/PM callback/breakpoint
	union {
		struct {
			DWORD callback, refdata;
			} CALLBACK;
		struct {
			DWORD brk_addr;
			union {
				struct {	// better packing in Chicago
					WORD bp_num;
					BYTE replaced;
					} WIN4;
				struct {
					BYTE replaced;
					WORD bp_num;
					} WIN3;
				} REPLBP;
			BYTE ff;
			} V86;
		} u;
		// extra DWORD in debug?
	} CB;
	
#define ARPL		0x63
#define INT30		0x30CD

int real_main(int argc, char *argv[])   
{
	FP v86_api;
	BYTE far *fp;
	
	PUT("WINBP -- Examine Windows Breakpoints and Callbacks");
	PUT("Copyright (c) 1994 Andrew Schulman. All rights reserved.");
	PUT("From \"Unauthorized Windows\" (IDG Books, 1994)\n");
	
	if (! (v86_api = GetVxDAPI(5)))	// get any V86 callback
		fail("This program requires Windows Enhanced mode");
	fp = (BYTE far *) v86_api;
	if (*fp != ARPL)
		fail("Something wrong! 2F/1684/5 V86 should return ptr to ARPL");
	
	v86bp_lin = MK_LIN(v86_api);
	printf("V86 breakpoints @ %08lXh ('", v86bp_lin);
	while (isprint(*fp)) { putchar(*fp); fp++; }
	printf("')\n");

	_asm mov ax, 1600h
	_asm int 2fh
	_asm mov byte ptr win_ver+1, al

	return 0;
}

int pmode_main(int argc, char *argv[])
{
	CB far *cb_tab, far *cb;
	FP pm_api;
	DWORD api_cb_addr;
	WORD avail_bps;
	WORD v86_seg, v86_ofs, pm_seg, pm_ofs;
	WORD first_v86_bp_seg, first_v86_bp_ofs;
	WORD last_v86_bp_seg, last_v86_bp_ofs;
	int verbose = 0;
	int i;
	
	char *s = argv[1];
	if ((argc > 1) && 
		((s[0] == '-') || (s[0] == '/')) &&
		((s[1] == 'v') || (s[1] == 'V')))
		verbose++;
	
	if (! (pm_api = GetVxDAPI(5)))		// get any PM callback
		fail("This program requires Windows Enhanced mode");
	if (*((WORD far *) pm_api) != INT30)
		fail("Something wrong! 2F/1684/5 PM should return ptr to INT 30h");
	
	pm_seg = FP_SEG(pm_api);
	pmbp_base = GetGDTSelectorBase(pm_seg);
	printf("PM breakpoints @ %08lXh\n", pmbp_base);
	
	pmbp_size = GetGDTSelectorLimit(pm_seg) + 1;
	max_bps = (pmbp_size >> 1) - 0x100;	
	printf("MaxBPs = %u (%04Xh)\n", max_bps, max_bps);

	last_v86_bp_seg = v86bp_lin >> 4;
	last_v86_bp_ofs = v86bp_lin & 0x0F;
	first_v86_bp_seg = (v86bp_lin >> 4) - max_bps;
	first_v86_bp_ofs = v86bp_lin - ((DWORD) first_v86_bp_seg << 4);
	printf("First V86 BP = %04X:%04X\n", first_v86_bp_seg, first_v86_bp_ofs);
	printf("Last  V86 BP = %04X:%04X\n", last_v86_bp_seg, last_v86_bp_ofs);
	
	bp_tab_lin = pmbp_base - (max_bps << 3);
	if (bp_tab_lin & 0xFFF)	// not on page boundary!
		fail( (bp_tab_lin & 0xFFF) == (max_bps * 4) ?
			"Looks like Debug Windows has larger BPs?" :
			"Can't get BP table address!");
	
	printf("BP table @ %08lXh\n", bp_tab_lin);
	
	if (! verbose)
		fail("\nRun WINBP -VERBOSE to list all breakpoints and callbacks");
	
	cb_tab = (CB far *) map_linear(bp_tab_lin, max_bps * sizeof(CB));
	if (! cb_tab) fail("Couldn't map callback table!");
	
	printf("\nV86 Breakpoints:\n");
	printf("BP #    V86 addr   Repl   CB #    Callback    Ref Data\n");
	printf("----    --------   ----   ----    --------    --------\n");
	
	for (i=0, cb=cb_tab; i<max_bps; i++, cb++)
	{
		if (cb->u.V86.ff == 0xff)
		{
			WORD bp_num;
			BYTE replaced;
			if (cb->u.V86.brk_addr > 0x1010FFEFL)	// only V86 mode mem
				continue;	// skip bogus entries
			if (win_ver < 4)
			{
				bp_num = cb->u.V86.REPLBP.WIN3.bp_num;
				replaced = cb->u.V86.REPLBP.WIN3.replaced;
			}
			else
			{
				bp_num = cb->u.V86.REPLBP.WIN4.bp_num;
				replaced = cb->u.V86.REPLBP.WIN4.replaced;
			}
			printf("%04X    %08lX    %02X    %04X    ",
				i,
				cb->u.V86.brk_addr,
				replaced,
				bp_num);
			if (bp_num <= max_bps)
				printf("%08lX    %08lX",
					cb_tab[bp_num].u.CALLBACK.callback,
					cb_tab[bp_num].u.CALLBACK.refdata);
			printf("\n");
					
#ifdef	SANITY_CHECK
			{
				BYTE far *fp = (BYTE far *) map_linear(cb->u.V86.brk_addr, 1);
				if (! fp) fail("Couldn't map breakpoint");
				if (*fp != ARPL) 
					printf("Something wrong!!\n");
				free_mapped_linear(fp);
			}
#endif			
		}
		if (ctrl_c_hit)
			goto done;
	}
	
	/* locate V86, PM API 2f/1684 callbacks */
	api_cb_addr = 
		cb_tab[((FP_OFF(pm_api) << 2) - 0x800) >> 3].u.CALLBACK.callback;
	
	printf("\nV86 and PM callbacks:\n");
	printf("BP #    Callback    Ref Data       V86          PM    \n");
	printf("----    --------    --------    ---------    ---------\n");

	v86_seg = first_v86_bp_seg;
	v86_ofs = first_v86_bp_ofs;
	pm_ofs = 0x200;
	
	for (i=0, cb=cb_tab; 
		 i<max_bps; 
		 i++, cb++, v86_seg++, v86_ofs -= 0x10, pm_ofs += 2)
	{
		if (cb->u.V86.ff != 0xff)
		{
			if (cb->u.CALLBACK.callback == 0)
				continue;	// skip blanks
			printf("%04X    %08lX    %08lX    %04X:%04X    %04X:%04X", 
				i,
				cb->u.CALLBACK.callback,
				cb->u.CALLBACK.refdata,
				v86_seg, v86_ofs,
				pm_seg, pm_ofs);
			if (cb->u.CALLBACK.callback == api_cb_addr)
			{
				DWORD far *fp = (DWORD far *) 
					map_linear(cb->u.CALLBACK.refdata, sizeof(DWORD));
				if (! fp) fail("Couldn't map API callback addr");
				printf("  ->  %08lX", *fp);	// show actual API callback addr
				free_mapped_linear(fp);
			}
			printf("\n");
		}
		if (ctrl_c_hit)
			goto done;
	}

done:
	free_mapped_linear(cb_tab);

	return 0;
}

/**********************************************************************/

// call the Windows "Get Device Entry Point Address" function
// Interrupt 2Fh Function 1684h
FP GetVxDAPI(WORD vxd_id)
{
	_asm {
		push di
		push es
		xor di, di
		mov es, di
		mov ax, 1684h
		mov bx, vxd_id
		int 2fh
		mov ax, di
		mov dx, es
		pop es
		pop di
		}
    // return value in DX:AX
}

/**********************************************************************/

#pragma pack(1)

typedef struct {
    WORD rpl : 2;
    WORD ti : 1;
    WORD index : 13;
    } SELECTOR; 

typedef struct {
    WORD limit_lo, base_lo;
    BYTE base_hi, rts_lo, limitrts_hi, base_xhi;
    } DESCRIPTOR;

typedef struct { WORD limit; DWORD base; } GDTR;

DESCRIPTOR far *get_gdt(WORD *plimit);
void sgdt(GDTR far *pgdtr);
static DWORD get_desc_base(DESCRIPTOR far *desc);
static DWORD get_desc_limit(DESCRIPTOR far *desc);

DWORD GetGDTSelectorBase(WORD seg)
{
    SELECTOR sel = *((SELECTOR *) &seg);
    DWORD base;
    if (sel.ti == 0)    // GDT
    {
        WORD gdt_limit;
        DESCRIPTOR far *gdt = get_gdt(&gdt_limit);
        if ((seg & ~8) > gdt_limit)
            fail("Invalid GDT selector");
        base = get_desc_base(&gdt[sel.index]);
        free_mapped_linear(gdt);
    }
    else                // LDT: oh what the heck, get it for them
		base = GetSelectorBase(seg);
	return base;
}

DWORD GetGDTSelectorLimit(WORD seg)
{
    SELECTOR sel = *((SELECTOR *) &seg);
    DWORD limit;
    if (sel.ti == 0)    // GDT
    {
        WORD gdt_limit;
        DESCRIPTOR far *gdt = get_gdt(&gdt_limit);
        if ((seg & ~8) > gdt_limit)
            fail("Invalid GDT selector");
        limit = get_desc_limit(&gdt[sel.index]);
        free_mapped_linear(gdt);
    }
    else                // LDT: oh what the heck, get it for them
		limit = GetSelectorLimit(seg);
	return limit;
}

// when done, free with free_mapped_linear
DESCRIPTOR far *get_gdt(WORD *plimit)
{
    DESCRIPTOR far *gdt;
    GDTR gdtr;
    WORD sel;
    
    /* get the linear base address and size (limit) of the Global
       Descriptor Table (GDT), using the Intel SGDT instruction */
    sgdt(&gdtr);
    gdt = (DESCRIPTOR far *) map_linear(gdtr.base, gdtr.limit + 1);
	if (! gdt) fail("Couldn't map GDT!");
    *plimit = gdtr.limit;
    return gdt;
}

void sgdt(GDTR far *pgdtr)
{
    _asm les bx, pgdtr
    _asm sgdt fword ptr es:[bx]
}

DWORD get_desc_base(DESCRIPTOR far *desc)
{
    return ((DWORD) desc->base_xhi << 24L) + 
	   ((DWORD) desc->base_hi << 16L) +  desc->base_lo;
}

DWORD get_desc_limit(DESCRIPTOR far *desc)
{
    DWORD limit = ((DWORD) (desc->limitrts_hi & 0x0F) << 16L) + 
        (DWORD) desc->limit_lo;
    if (desc->limitrts_hi & 0x80)   // page granularity
        limit *= 4096;
    return limit;
}
