/* 
WINPSP.C
Windows app displays PSPs for all Win16 and Win32 apps
from Schulman, Unauthorized Windows, 1994

bcc -I..\include -W -2 -P- winpsp.c prot.c winio.c wmhandlr.c argcargv.c 
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include "windows.h"
#include "prot.h"
#include "toolhelp.h"

#ifdef __cplusplus
extern "C" BOOL FAR PASCAL IsWinOldApTask(HANDLE hTask);
#else
extern BOOL FAR PASCAL IsWinOldApTask(HANDLE hTask);
#endif

#define MAP(ptr, bytes)     map_real((ptr), (bytes))
#define FREE_MAP(ptr)       free_mapped_linear(ptr)
#define GET_REAL(ptr)       get_real_addr(ptr)

#pragma pack(1)
typedef struct {
    BYTE type;      
    WORD owner, size;
    BYTE unused[3], name[8];
    } MCB;
#pragma pack()

WORD get_first_mcb(void)
{
    RMODE_CALL r;
    memset(&r, 0, sizeof(r));
    r.eax = 0x5200;
    if (dpmi_rmode_intr(0x21, 0, 0, &r))
    {
        // Extract seg of first MCB from SysVars.  Note that this is
        // at sysvars[-2]. You can't call map_real() on sysvars, and
        // then back up 2!  You must map in sysvars-2 to begin with.
        WORD far *tmp = (WORD far *) MAP(MK_FP(r.es, (WORD) r.ebx-2), 2);
        WORD first_mcb = *tmp;
        FREE_MAP(tmp);
        return first_mcb;
    }
    else
        return 0;
}

void display_dos_psp(WORD psp_seg, BYTE far *psp, MCB far *mcb)
{
    printf("%04X\t ", psp_seg);
    if (_osmajor >= 4)
    {
        char buf[9];
        _fmemcpy(buf, mcb->name, 8);
        buf[8] = '\0';
        if (*buf)
            printf("\t%-8s", buf);
		else
			printf("\t\t");
    }
	printf("\t%04X\n", mcb->size);
}

void display_win_psp(WORD prot_psp, char *szModule, WORD hTask)
{
	BYTE far *psp = (BYTE far *) MK_FP(prot_psp, 0);
    WORD real_psp = GetSelectorBase(prot_psp) >> 4;
	WORD flags;

	printf("%04X\t%04X\t%-8s\t%04X\t%0lX\t", 
		real_psp, prot_psp, szModule, hTask,
		GetSelectorLimit(prot_psp) + 1);

	// WinOldAp flag in Windows PSP
	flags = *((WORD far *) &psp[0x48]);
	if (IsWinOldApTask(hTask) && (! (flags & 1))) /* insanity check */
		fail("IsWinOldApTask flag weirdness!"); 
	if (flags & 1) printf("DOS ");
	
	// Win32 flag in TDB
	flags = *((WORD far *) MK_FP(hTask, 0x16));
	if (flags & 0x10) printf("WIN32");
	printf("\n");
}
		
// void fail(char *s) { printf("%s\n", s); exit(1); }

main()
{
    TASKENTRY te;
    BYTE far *maybe_psp;
    MCB far *mcb;
    WORD mcb_seg, mapped;
    BOOL ok;
	
	printf("DOS PSPs (from MCB chain):\n");
	printf("Real\t    \tName    \tParas\n");
    
    // walk DOS MCB chain, looking for PSPs
    if (! (mcb_seg = get_first_mcb()))
        fail("Can't get MCB chain!");
    for (;;)
    {
        mcb = (MCB far *) MAP(MK_FP(mcb_seg, 0), sizeof(MCB));
        maybe_psp = (BYTE far *) MAP(MK_FP(mcb_seg + 1, 0), 512);
        if (*((WORD far *) maybe_psp) == 0x20CD)  // does it look like a PSP?
        {
            if ((mcb_seg + 1) == mcb->owner)      // regular DOS app PSP
                display_dos_psp(mcb_seg + 1, maybe_psp, mcb);
        }
        FREE_MAP(maybe_psp);
        if (mcb->type == 'Z')
            break;                              // end of list
        mcb_seg = mcb_seg + mcb->size + 1;      // walk list
        FREE_MAP(mcb);
    }

    FREE_MAP(mcb);                              // free last one
    mapped = get_mapped();
    if (mapped != 0)
        printf("ERROR!  %u mapped selectors remaining!\n", mapped);

	printf("\nWindows PSPs (from task list):\n");
	printf("Real\tProt\tName    \tTask\tSize\n");
    
    // now walk Windows task list, and extract PSPs
    te.dwSize = sizeof(te);
    ok = TaskFirst(&te);
    while (ok)
    {
        BYTE far *tdb = (BYTE far *) MK_FP(te.hTask, 0);
        WORD prot_psp = *((WORD far *) &tdb[0x60]);
        if (*((WORD far *) MK_FP(prot_psp, 0)) == 0x20CD) /* really a PSP? */
			display_win_psp(prot_psp, te.szModule, te.hTask);
        ok = TaskNext(&te);
    }

    return 0;
}

