/*
WV86TEST.C -- take over INT 21h and INT 2Fh, count calls in V86 mode
Uses wv86test.def with STUB 'v86test.exe'

from Schulman, Unauthorized Windows, 1994

bcc -I..\include -L..\lib -W -2 -P- wv86test.c unauthw.lib 
*/

#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <process.h>
#include <time.h>
#include <dos.h>
#include "windows.h"
#include "winio.h"
#include "wmhandlr.h"
#include "prot.h"

typedef unsigned short WORD;
typedef unsigned long DWORD;

#define VM_MAX      8
#define VM_OTHER    VM_MAX

#define GET_STATS   0x0FFFF
#define SIGNATURE   "V86TEST"

#define VXD_MAX     0x100
#define VXD_OTHER   VXD_MAX

typedef enum { 
    NO_WIN, 
    WIN_INIT_BEGIN,             // got WIN_INIT_NOTIFY
    WIN_INIT_DONE,              // got WIN_INIT_COMPLETE
    WIN_FINI_BEGIN,             // got WIN_EXIT_BEGIN
    WIN_FINI_DONE,              // got WIN_EXIT_NOTIFY
    NUM_STATES 
    } STATE;

static char *state_str[NUM_STATES] = {
    "Before Windows started", 
    "During Windows init",      // got WIN_INIT_NOTIFY
    "While Windows running",    // got WIN_INIT_COMPLETE
    "During Windows exit",      // got WIN_EXIT_BEGIN
    "After Windows exited",     // got WIN_EXIT_NOTIFY
    } ;

#pragma pack(1)

typedef struct {
    char signature[8];
    DWORD calls[NUM_STATES], v86_calls[NUM_STATES];
    DWORD iopl_count[4];
    DWORD vm[VM_MAX+1];
    DWORD int21[0x100];
    DWORD int2f[0x100], int2f16[0x100], int2f1607[VXD_OTHER+1];
    time_t start, end;
    } STATS;
    
static STATE state = NO_WIN;
static BOOL display_changes = 0;
static BOOL do_auto_refresh = 0;
static UINT htimer = 0;
static STATS far *fp;
static STATS *stats;
static HMENU gmenu;

/**********************************************************************/

void display_results(STATS far *fp2)
{
    STATS *fp = stats;
    STATS *delta;
    DWORD elapsed;
    int i;
    
    if (display_changes)
    {
        if (! (delta = malloc(sizeof(STATS))))
            fail("Insufficient memory");

        // Show changes: first copy the new stats over to delta
        _fmemcpy(delta, fp2, sizeof(STATS));

        // Then subtract the previous stats to get changes
        time(&delta->end);
        elapsed = delta->end - fp->end;
        delta->calls[WIN_INIT_DONE] -= fp->calls[WIN_INIT_DONE];
        delta->v86_calls[WIN_INIT_DONE] -= fp->v86_calls[WIN_INIT_DONE];
        printf("%lu seconds elapsed\n", elapsed);
        printf("%lu calls, %lu in V86 mode\n",
            delta->calls[WIN_INIT_DONE], delta->v86_calls[WIN_INIT_DONE]);
        if (elapsed)
            printf("%lu INT 21/2F calls/second\n", 
                delta->calls[WIN_INIT_DONE] / elapsed);

        for (i=0; i<4; i++)
        {
            delta->iopl_count[i] -= fp->iopl_count[i];
            delta->vm[i] -= fp->vm[i];
        }
        delta->vm[VM_OTHER] -= fp->vm[VM_OTHER];
        for (i=0; i<0x100; i++)
        {
            delta->int21[i] -= fp->int21[i];
            delta->int2f[i] -= fp->int2f[i];
            delta->int2f16[i] -= fp->int2f16[i];
        }
        for (i=0; i<=VXD_MAX; i++)  // get VXD_OTHER too
            delta->int2f1607[i] -= fp->int2f1607[i];
        fp = delta; // have display_results work on delta
    }
    else
    {
        _fmemcpy(fp, fp2, sizeof(STATS));

        for (i=NO_WIN; i<NUM_STATES; i++)
            printf("%s:\t%lu INT 21/2F calls, %lu in V86 mode\n",
                state_str[i], fp->calls[i], fp->v86_calls[i]);

        printf("\nWhile Windows running:\n");
        time(&fp->end);
        if (! fp->start) time(&fp->start);
        if ((elapsed = fp->end - fp->start) != 0)
        {
            printf("Windows active for %lu seconds\n", elapsed);
            printf("%lu INT 21/2F calls/second\n", 
                fp->calls[WIN_INIT_DONE] / elapsed);
        }
    }
    
    for (i=0; i<4; i++)
        if (fp->iopl_count[i])
            printf("IOPL=%d -- %lu calls\n", i, fp->iopl_count[i]);

    for (i=0; i<VM_MAX; i++)
        if (fp->vm[i])
            printf("VM #%d -- %lu calls\n", i, fp->vm[i]);
    if (fp->vm[VM_OTHER])
        printf("VM > #%d -- %lu calls\n", VM_MAX, fp->vm[VM_OTHER]);

    #define PRINT_CALLS(s, fp) { \
        DWORD *p; \
        printf("\n%s:\n", (s)); \
        for (i=0, p=fp; i<0x100; i++, p++) \
            if (*p) \
                printf("%02X: %lu\t", i, *p); \
        }

    PRINT_CALLS("INT 21h calls", fp->int21);
    PRINT_CALLS("\nINT 2Fh calls", fp->int2f);
    PRINT_CALLS("\nINT 2Fh AH=16h calls", fp->int2f16);
    PRINT_CALLS("\nINT 2Fh AX=1607h calls", fp->int2f1607);

    if (fp->int2f1607[VXD_OTHER])
        printf("VxD>#%04X: %lu\n", VXD_MAX, fp->int2f1607[VXD_OTHER]);

    printf("\n\n");
    
    _fmemcpy(stats, fp2, sizeof(STATS));
    time(&stats->end);
    if (display_changes) free(delta);
}

/**********************************************************************/

STATS far *get_stats(void)  // call resident copy of V86TEST
{
    union REGS r;
    struct SREGS s;
    STATS far *fp;
    r.x.ax = GET_STATS;
    memset(&s, 0, sizeof(s));
    real_int86x(0x2f, &r, &r, &s);      // real_int86x in PROT.C
    if (s.es == 0)
        return 0;
    fp = MK_FP(s.es, r.x.bx);
    if ((fp = map_real(fp, sizeof(STATS))) == 0)   // map_real in PROT.C
        return 0;
    return (_fstrcmp(fp->signature, SIGNATURE) == 0) ? fp : 0;
}

void free_stats(HWND hwnd) { free_mapped_linear(fp); }

/**********************************************************************/

/* MENU HANDLERS (see MK_MENU in main) */
    
void refresh(HWND hwnd, int wID)
{
    winio_clear(hwnd);
    display_results(fp); 
}

void clear(HWND hwnd, int wID) { _fmemset(fp, 0, sizeof(STATS)); }

void show_changes(HWND hwnd, int wID)
{
    CheckMenuItem(gmenu, 3, display_changes ? MF_UNCHECKED : MF_CHECKED);
    display_changes = ! display_changes;
    refresh(hwnd, wID);
}

long on_time(HWND hwnd, unsigned message, WORD wParam, LONG lParam)
{
    winio_clear(hwnd);
    display_results(fp); 
}

void auto_refresh(HWND hwnd, int wID)
{
    if (do_auto_refresh)    // on, turn off
    {
        do_auto_refresh = 0;
        KillTimer(hwnd, htimer);
        CheckMenuItem(gmenu, 4, MF_UNCHECKED);
    }
    else    // off, turn on
    {
        do_auto_refresh = 1;
        wmhandler_set(hwnd, WM_TIMER, on_time);
        if (! (htimer = SetTimer(hwnd, 1, 5000, NULL))) // every 5 seconds
            fail("Can't create timer");
        CheckMenuItem(gmenu, 4, MF_CHECKED);
    }
}

/**********************************************************************/

main(int argc, char *argv[])
{
    HWND hwnd;
    int i;

    // only one instance: should put INT handler in DLL
    if (__hPrevInst)
        fail("WV86TEST already running");
    
    hwnd = winio_current();
    gmenu = CreateMenu();   // global
    #define MK_MENU(id, str, func) { \
        AppendMenu(gmenu, MF_STRING | MF_ENABLED, id, str); \
        winio_setmenufunc(hwnd, id, func); \
        }
    MK_MENU(1, "&Refresh", refresh);
    MK_MENU(2, "&Clear", clear);
    MK_MENU(3, "&Show Changes", show_changes);
    MK_MENU(4, "&Auto-Refresh", auto_refresh);
    InsertMenu(winio_hmenumain(hwnd), 1, 
        MF_STRING | MF_POPUP | MF_BYPOSITION, gmenu, "&Test");
    DrawMenuBar(hwnd);
    
    if (! (fp = get_stats()))
    {
        // Instead of failing, could do ExitWindowsExec and
        // ExitWindows(EW_RESTARTWINDOWS) here.
        fail("Can't get V86TEST statistics");
    }
    winio_onclose(hwnd, free_stats);
    
    if (! (stats = malloc(sizeof(STATS))))
        fail("Insufficient memory");
    display_results(fp); 
    
    return 0;
}
